/*	$Kolabore: cfg.cpp,v 1.26 2006/10/24 22:10:27 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#include <netdb.h>
#include <unistd.h>
}

#include <iostream>
#include <list>
#include <string>
#include <vector>

#include <cstdlib>

#include "cfg.h"
#include "compat/xsicompat.h"

#ifndef CHROOT_DIR
#define CHROOT_DIR	"/var/empty"
#endif
#ifndef PRIVDROP_USER
#define PRIVDROP_USER	"nobody"
#endif
#ifndef PRIVDROP_GROUP
#define PRIVDROP_GROUP	"nobody"
#endif

extern char    *optarg;
extern int	opterr;
extern int	optind;
extern int	optopt;
extern int	optreset;

using namespace std;

Config::Config(int argc, char **argv)
{
	int		ch, error;
	const char     *errStr;
	string		lAddr;
	struct addrinfo hints;
	string listenService = DEFAULT_PORT;

	update_interval = 2;

	one_client_mode = false;
	compress_xml = false;
	dont_detach = false;
	format_output = false;
	one_shot = false;
	show_raid = false;
	show_pf = false;

	while ((ch = getopt(argc, argv, "1a:C:cdFfG:i:op:r:U:")) != -1) {
		switch (ch) {
		case '1':
			one_client_mode = true;
			xsi_options.push_back("one_client_mode");
			break;
		case 'a':
			lAddr = optarg;
			break;
		case 'C':
			chrootDir = optarg;
			if (chrootDir == "/") {
				warnx("=== Chroot to `%s'? No way. ===",
				      chrootDir.c_str());
				chrootDir.clear();
			}
			break;
		case 'c':
			compress_xml = true;
			xsi_options.push_back("compress_output");
			break;
		case 'd':
			dont_detach = true;
			xsi_options.push_back("dont_detach");
			break;
		case 'F':
			show_pf = true;
			xsi_options.push_back("show_packet_filter");
			break;
		case 'f':
			format_output = true;
			xsi_options.push_back("format_output");
			break;
		case 'G':
			unprivGroup = optarg;
			break;
		case 'i':
			update_interval = (long)strtonum(optarg, 1, 3600, &errStr);
			if (errStr)
				errx(1, "Update interval is %s (min 1, max 3600): %s", errStr, optarg);
			break;
		case 'o':
			one_shot = true;
			xsi_options.push_back("one_shot");
			break;
		case 'p':
			listenService = optarg;
			break;
		case 'r':
			show_raid = true;
			{
				string tmp = optarg;
				raid_controllers.push_back(tmp);
				raid_controllers.sort();
				raid_controllers.unique();
			}
			xsi_options.push_back("show_raid");
			break;
		case 'U':
			unprivUser = optarg;
			break;
		case '?':
			usage();
			/* NOTREACHED */
		default:
			break;
		}
	}
	argc -= optind;
	argv += optind;

	if (xsi_options.empty())
		xsi_options.push_back("defaults");
	else {
		xsi_options.sort();
		xsi_options.unique();
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	error = getaddrinfo(lAddr.empty() ? NULL : lAddr.c_str(),
			    listenService.c_str(), &hints, &listen_addr);
	if (error)
		errx(1, "address %s, service/port %s: %s",
		     lAddr.empty() ? "any" : lAddr.c_str(),
		     listenService.c_str(), gai_strerror(error));
}

Config::~Config(void)
{
	freeaddrinfo(listen_addr);
}

void
Config::usage(void)
{
	extern char *__progname;

	cerr << "usage: "
		<< __progname
		<< " "
		<< "[-1cdFfo] [-C chrootdir] [-a address] [-G group] [-i interval]\n"
		<< "           [-p port] [-r device] [-U user]"
		<< endl;

	exit(1);
}

bool
Config::oneClientMode(void)
{
	return (one_client_mode);
}

bool
Config::compressXml(void)
{
	return (compress_xml);
}

bool
Config::dontDetach(void)
{
	return (dont_detach);
}

bool
Config::formatOutput(void)
{
	return (format_output);
}

bool
Config::oneShot(void)
{
	return (one_shot);
}

bool
Config::showRaid(void)
{
	return (show_raid);
}

bool
Config::showPacketFilter(void)
{
	return (show_pf);
}

long
Config::getUpdateInterval(void)
{
	return (update_interval);
}

const char *
Config::getUpdateIntervalString(void)
{
	return (makeIntStr((uint32_t)update_interval));
}

struct addrinfo *
Config::getListenAddr(void)
{
	return (listen_addr);
}

list<string>
Config::getRaidControllers(void)
{
	return (raid_controllers);
}

list<string>
Config::getXSIOptionsList(void)
{
	return (xsi_options);
}

const char *
Config::getChrootDir(void)
{
	if (chrootDir.empty())
		return (CHROOT_DIR);
	else
		return (chrootDir.c_str());
}

const char *
Config::getPrivdropUser(void)
{
	if (unprivUser.empty())
		return (PRIVDROP_USER);
	else
		return (unprivUser.c_str());
}

const char *
Config::getPrivdropGroup(void)
{
	if (unprivGroup.empty()) {
		if (unprivUser.empty())
			return (PRIVDROP_GROUP);
		else
			return (NULL);	 // use passwd.pw_gid
	} else
		return (unprivGroup.c_str());
}
