/*	$Kolabore: freebsd.cpp,v 1.44 2006/10/29 18:44:55 maxx Exp $	*/
/*
 * Copyright (c) 2006 Joshua Elsasser <josh@elsasser.org>
 *                    Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * freebsd.cpp:
 * Gathering information and stats on FreeBSD 6.0 or newer.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/resource.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/socket.h>
#include <vm/vm_param.h>

#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>
#include <net/pfvar.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/ip_carp.h>

#include <devstat.h>
#include <err.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <stdint.h>
}

#include <string>
#include <vector>

#include <cctype>
#include <cstdarg>
#include <ctime>

#include <iostream>

#include "freebsd.h"

#define SMALLBUFSIZ	128

using namespace std;

struct freebsd_private {
	long	lastTimes[CPUSTATES];
	time_t	boot;
};

struct freebsd_interrupt {
	unsigned long	count;
	string		source;
};

static const struct ifmedia_description ifm_type_descriptions[] =
	IFM_TYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_ethernet_descriptions[] =
	IFM_SUBTYPE_ETHERNET_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_tokenring_descriptions[] =
	IFM_SUBTYPE_TOKENRING_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_fddi_descriptions[] =
	IFM_SUBTYPE_FDDI_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_ieee80211_descriptions[] =
	IFM_SUBTYPE_IEEE80211_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_atm_descriptions[] =
	IFM_SUBTYPE_ATM_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_shared_descriptions[] =
	IFM_SUBTYPE_SHARED_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_ieee80211_mode_descriptions[] =
	IFM_SUBTYPE_IEEE80211_MODE_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_tokenring_option_descriptions[] =
	IFM_SUBTYPE_TOKENRING_OPTION_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_fddi_option_descriptions[] =
	IFM_SUBTYPE_FDDI_OPTION_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_ieee80211_option_descriptions[] =
	IFM_SUBTYPE_IEEE80211_OPTION_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_atm_option_descriptions[] =
	IFM_SUBTYPE_ATM_OPTION_DESCRIPTIONS;
static const struct ifmedia_description ifm_shared_option_descriptions[] =
	IFM_SHARED_OPTION_DESCRIPTIONS;
#ifdef IFM_STATUS_VALID_LIST
static const int ifm_status_valid_list[] = IFM_STATUS_VALID_LIST;
static const struct ifmedia_status_description ifm_status_descriptions[] =
	IFM_STATUS_DESCRIPTIONS;
#endif
static const char *carp_states[] = { CARP_STATES };

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	pf_fd = -1;
	priv = new struct freebsd_private;
	memset(priv->lastTimes, 0, sizeof(long) * CPUSTATES);
}

XSI::~XSI(void)
{
	delete priv;
	if (pf_fd != -1)
		close(pf_fd);
}

/*
 * Devices to be opened before dropping privileges.
 * Silently returns 0 on success. Warns and returns >0 on error.
 *
 * NOTE: Valid XML must not rely on information gathered through privileged
 *       devices! Only optional, user-requested features may depend on
 *       things that may need super-user privileges.
 */
int
XSI::tryOpenPrivilegedDevices(void)
{
	if (configuration->showRaid()) {
		warnx("Sorry, RAID monitoring support is not (yet?) available on FreeBSD.");
		return (1);
	}

	if (configuration->showPacketFilter()) {
		if ((pf_fd = open("/dev/pf", O_RDONLY, 0)) == -1) {
			if (errno == ENOENT)
				warnx("/dev/pf does not exist, is the pf(4) kernel module loaded?");
			else
				warn("Unable to open /dev/pf to gather pf(4) stats");
			return (1);
		}
	}

	return (0);
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	char	       *str;
	int		num;
	struct timeval	when;

	if(sysctlStr(&str, CTL_HW, HW_MACHINE, -1)) {
		system.setArch(str);
		delete[] str;
	}

	if(sysctlStr(&str, CTL_HW, HW_MODEL, -1)) {
		system.setModel(str);
		delete[] str;
	}

	/* XXX this might not be good enough for all arches */
	if(sysctlFixed(&num, sizeof(num), "dev.cpu.0.freq") ||
	   sysctlFixed(&num, sizeof(num), "hw.clockrate"))
		system.setCPUFreq(num);

	if(sysctlFixed(&num, sizeof(num), CTL_HW, HW_NCPU, -1))
		system.setNumCPUs(num);

	setVers();

	if(sysctlFixed(&when, sizeof(when), CTL_KERN, KERN_BOOTTIME, -1))
		priv->boot = when.tv_sec;

	/* XXX raids */
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	int	num;
	time_t	when;
	double	load[3];

	if(sysctlFixed(&num, sizeof(num), "dev.cpu.0.freq"))
		system.setCPUFreq(num);

	setCPUTimes();

	setMem();

	if(0 < time(&when))
		system.setOSUptime(when - priv->boot);

	if(0 < getloadavg(load, 3))
		system.setLoadAvg(load);

	setFilesystems();
	setDisks();
	setInterrupts();
	setNetStats();
	setPFStats();
	setSensors();
}

void
XSI::setCPUTimes(void)
{
	long info[CPUSTATES];
	long total;
	long diff[CPUSTATES];
	float percent;
	int i;
	vector<struct cpu> cpus;

	if(!sysctlFixed(info, sizeof(info), "kern.cp_time"))
		return;

	cpus.resize(1);
	cpus[0].num = -1;

	total = 0;
	for (i = 0; i < CPUSTATES; i++) {
		diff[i] = info[i] - priv->lastTimes[i];
		priv->lastTimes[i] = info[i];
		total += diff[i];
	}

	if (total > 0)
		percent = 100 / (float)total;
	else
		percent = 0;

	cpus[0].system    = diff[CP_SYS] * percent;
	cpus[0].user      = diff[CP_USER] * percent;
	cpus[0].nice      = diff[CP_NICE] * percent;
	cpus[0].interrupt = diff[CP_INTR] * percent;
	cpus[0].idle      = diff[CP_IDLE] * percent;

	system.setCPUs(cpus);
}

void
XSI::setMem(void)
{
	struct vm_info	mem;
	u_int		phymem, phyfree;
	long		pagesize;
	int		mib[2];
	size_t		miblen;
	struct xswdev	swap;
	int		ii;

	pagesize = getpagesize();

	if(!sysctlFixed(&phymem, sizeof(phymem), "vm.stats.vm.v_page_count"))
		return;
	if(!sysctlFixed(&phyfree, sizeof(phyfree), "vm.stats.vm.v_free_count"))
		return;

	mem.ram_total  = phymem  * pagesize;
	mem.ram_free   = phyfree * pagesize;
	mem.swap_total = 0;
	mem.swap_free  = 0;

	miblen = 2;
	if(0 > sysctlnametomib("vm.swap_info", mib, &miblen))
		return;

	for(ii = 0; ; ii++) {
		if(!sysctlFixed(&swap, sizeof(swap), mib[0], mib[1], ii, -1))
			break;
		if(XSWDEV_VERSION != swap.xsw_version)
			break;
		mem.swap_total += swap.xsw_nblks * pagesize;
		mem.swap_free  += (swap.xsw_nblks - swap.xsw_used) * pagesize;
	}

	system.setVMInfo(mem);
}

void
XSI::setVers(void)
{
	char *type, *rel, *status, *fullvers;
	string vers, kern;

	if(!sysctlStr(&type, CTL_KERN, KERN_OSTYPE, -1))
		return;

	if(!sysctlStr(&rel, CTL_KERN, KERN_OSRELEASE, -1)) {
		delete[] type;
		return;
	}

	if(NULL != (status = strchr(rel, '-')))
		*(status++) = '\0';

	if(sysctlStr(&fullvers, CTL_KERN, KERN_VERSION, -1)) {
		char *sharp, *colon, *ident;

		if(NULL != (sharp = strchr(fullvers, '#')) &&
		   NULL != (colon = strchr(fullvers, ':')) &&
		   NULL != (ident = strrchr(fullvers, '/'))) {
			char *ii, *date;
			char buf[BUFSIZ];
			char wkday[5], month[5];
			int day, hour, minute, second, year;
			struct tm tm;

			/* snip entire version string in two at the newline */
			for (ii = fullvers; *ii && *ii != '\n'; ii++) {}
			*ii = '\0';

			/* learn the kernel name from the end of the 2nd half */
			ident++;
			for (ii = ident; '\0' != *ii && !isspace(*ii); ii++) {}
			*ii = '\0';

			for (date = colon + 1; isspace(*date); date++) {}

			/* reformat and parse date(1) string; %+ doesn't do it */
			if (sscanf(date, "%4s %4s %d %d:%d:%d %*s %d",
				   (char *)&wkday, (char *)&month, &day,
				   &hour, &minute, &second, &year) != 7)
				errx(1, "%s: sscanf", __func__);
			if (snprintf(buf, sizeof(buf), "%s %s %2d %0d:%0d:%0d %d",
				     wkday, month, day, hour, minute, second, year) == -1)
				errx(1, "%s: snprintf", __func__);
			if (strptime(buf, "%a %b %e %H:%M:%S %Y", &tm) == NULL)
				errx(1, "%s: strptime", __func__);

			/* assemble version string as in /etc/motd */
			vers = string(fullvers, sharp - fullvers) +
				"(" + ident + ") " + sharp;

			/* assemble normalized kernel version */
			*colon = '\0';
			if(strftime(buf, 64, "%Y%m%d%H%M", &tm) == 0)
				errx(1, "%s: strftime", __func__);
			kern = string(ident) + sharp + " " + buf;
		}
		delete[] fullvers;
	}

	system.setOSInfo(type, rel, status, kern.c_str(), vers.c_str());

	delete[] type;
	delete[] rel;
}

void
XSI::setFilesystems(void)
{
	struct statfs	*stats;
	int		 len, ii;
	vector<struct fs> fslist;

	if(0 > (len = getfsstat(NULL, 0, MNT_NOWAIT)))
		return;
	stats = new struct statfs[len];
	if(0 > (len = getfsstat(stats, len * sizeof(*stats), MNT_NOWAIT))) {
		delete[] stats;
		return;
	}

	for(ii = 0; ii < len; ii++) {
		struct fs fs;

		if(STATFS_VERSION != stats[ii].f_version)
			continue;

		/*
		 * Ignore filesystems that don't hold any interesting
		 * information (for us.)
		 */
		fs.fs_type     = stats[ii].f_fstypename;
		if (fs.fs_type == "devfs" || fs.fs_type == "linprocfs")
			continue;

		fs.mountpoint  = stats[ii].f_mntonname;
		fs.fs_name     = stats[ii].f_mntfromname;
		fs.size_total  = stats[ii].f_blocks * stats[ii].f_bsize;
		fs.size_free   = stats[ii].f_bfree  * stats[ii].f_bsize;
		fs.size_avail  = stats[ii].f_bavail * stats[ii].f_bsize;
		fs.nodes_total = stats[ii].f_files;
		fs.nodes_free  = stats[ii].f_ffree;

		fslist.push_back(fs);
	}

	delete[] stats;

	storage.setFilesystems(fslist);
}

void
XSI::setDisks(void)
{
	int			numdevs, selected = 0, selections = 0, i;
	long			generation;
	struct statinfo 	si;
	struct device_selection *ds = NULL;
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	bool			drop_old_stats = false, disks_changed;

	/* Make sure that we're linked against the proper libdevstat */
	{
		static bool first = true;

		if (first) {
			if (devstat_checkversion(NULL) == -1)
				errx(1, "%s", devstat_errbuf);
			first = false;
		}
	}

	bzero(&si, sizeof(si));
	si.dinfo = new struct devinfo;
	bzero(si.dinfo, sizeof(struct devinfo));

	if (devstat_getdevs(NULL, &si) == -1)
		errx(1, "%s", devstat_errbuf);

	numdevs = si.dinfo->numdevs;
	if (devstat_selectdevs(&ds, &selected, &selections,
			       &generation, si.dinfo->generation,
			       si.dinfo->devices, numdevs, NULL, 0,
			       NULL, 0, DS_SELECT_ONLY, 0,
			       1) == -1)
		errx(1, "%s", devstat_errbuf);

	if (numdevs != (int)old_disks.size())
		drop_old_stats = true;

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (i = 0; i < numdevs; i++) {
				struct devstat *dev =
					&si.dinfo->devices[ds[i].position];
				XSI_DiskIO disk;
				char buf[BUFSIZ];

				snprintf(buf, sizeof(buf), "%s%d",
					 dev->device_name, dev->unit_number);
				disk.setDiskName(buf);
				disk.setBytesWritten(dev->bytes[DEVSTAT_WRITE]);
				disk.setBytesRead(dev->bytes[DEVSTAT_READ]);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			for (i = 0; i < numdevs; i++) {
				struct devstat *dev =
					&si.dinfo->devices[ds[i].position];
				char buf[BUFSIZ];

				snprintf(buf, sizeof(buf), "%s%d",
					 dev->device_name, dev->unit_number);
				if (strcmp(old_disks[i].getDiskName(), buf)
				    != 0) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				old_disks[i].setBytesWritten(dev->bytes[DEVSTAT_WRITE]);
				old_disks[i].setBytesRead(dev->bytes[DEVSTAT_READ]);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);

	delete ds;
	delete si.dinfo;
}

void
XSI::setInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints = system.getInterrupts();
	bool			drop_old_stats = false, ints_changed;
	vector<struct freebsd_interrupt> ints;
	vector<struct freebsd_interrupt>::iterator ii;

	{
		size_t siz;
		unsigned long nintr, i;
		unsigned long *intrcnt;
		char *intrnamebuf, *cp;

		if (sysctlbyname("hw.intrcnt", NULL, &siz, NULL, 0) == -1)
			err(1, "%s: sysctlbyname", __func__);

		if ((nintr = siz / sizeof(u_long)) == 0)
			return;

		if ((intrcnt = new unsigned long[nintr]) == NULL)
			errx(1, "%s: Out of memory", __func__);

		intrnamebuf = sysctlDynamic("hw.intrnames", NULL);
		if (sysctlbyname("hw.intrcnt", intrcnt, &siz, NULL, 0) == -1)
			err(1, "%s: sysctlbyname", __func__);

		for (cp = intrnamebuf, i = 0; i < nintr; i++) {
			if (intrcnt[i]) {
				struct freebsd_interrupt intr;

				intr.source = split(cp)[1];
				intr.count = intrcnt[i];
				ints.push_back(intr);
			}
			cp += strlen(cp) + 1;
		}

		delete[] intrcnt;
		delete[] intrnamebuf;
	}

	if (ints.size() != old_ints.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = ints.begin(); ii != ints.end(); ii++) {
				XSI_Interrupt intr(ii->source.c_str());

				intr.setIntsCount((uint64_t)ii->count);

				new_ints.push_back(intr);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < ints.size(); i++) {
				if (ints[i].source != old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount((uint64_t)ints[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

void
XSI::setNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		err(1, "%s: getifaddrs", __func__);

	/* Convenience; make a vector of pointers to link-layer interfaces */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_LINK)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);
				struct if_data *ifd;

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(netIfStat, (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(netIfStat, (*it), (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfmediaInfo(netIfStat);
				getNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

				netIfStat.setPacketsIn(ifd->ifi_ipackets);
				netIfStat.setPacketsOut(ifd->ifi_opackets);
				netIfStat.setErrorsIn(ifd->ifi_ierrors);
				netIfStat.setErrorsOut(ifd->ifi_oerrors);
				netIfStat.setBytesIn(ifd->ifi_ibytes);
				netIfStat.setBytesOut(ifd->ifi_obytes);
				netIfStat.setPacketsDropped(ifd->ifi_iqdrops);
				netIfStat.setCollisions(ifd->ifi_collisions);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				struct if_data	       *ifd;

				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(old_nstats[i], (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(old_nstats[i], (*it), (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfmediaInfo(old_nstats[i]);
				getNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

				old_nstats[i].setPacketsIn(ifd->ifi_ipackets);
				old_nstats[i].setPacketsOut(ifd->ifi_opackets);
				old_nstats[i].setErrorsIn(ifd->ifi_ierrors);
				old_nstats[i].setErrorsOut(ifd->ifi_oerrors);
				old_nstats[i].setBytesIn(ifd->ifi_ibytes);
				old_nstats[i].setBytesOut(ifd->ifi_obytes);
				old_nstats[i].setPacketsDropped(ifd->ifi_iqdrops);
				old_nstats[i].setCollisions(ifd->ifi_collisions);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

void
XSI::setPFStats(void)
{
	struct pf_status	status;
	struct pfioc_iface	iface;
	struct pfi_kif		ifs[256];	// Eats >50 kB of stack memory
	list<int>		ifs_with_traffic;
	list<int>::iterator	iwt_it;
	int			i;
	vector<FilteredIfStat>	old_fIfStats = packetFilter.getFilteredIfStats();
	bool			drop_old_stats = false, filteredIfs_changed;

	if (pf_fd == -1)
		return;

	bzero(&status, sizeof(status));
	if (ioctl(pf_fd, DIOCGETSTATUS, &status) == -1)
		err(1, "%s: Unable to get status from pf(4)", __func__);

	if (!status.running) {
		packetFilter.setEnabled(false);
		return;
	} else
		packetFilter.setEnabled(true);
	packetFilter.setRuntime(time(NULL) - status.since);
	packetFilter.setRuleMatches(status.counters[PFRES_MATCH]);
	packetFilter.setStates(status.states);
	packetFilter.setStateSearches(status.fcounters[FCNT_STATE_SEARCH]);
	packetFilter.setStateInserts(status.fcounters[FCNT_STATE_INSERT]);
	packetFilter.setStateRemovals(status.fcounters[FCNT_STATE_REMOVALS]);
	packetFilter.setStateMismatches(status.counters[PFRES_BADSTATE]);

	bzero(&iface, sizeof(iface));
	iface.pfiio_buffer = ifs;
	iface.pfiio_esize = sizeof(ifs[0]);
	iface.pfiio_size = sizeof(ifs) / sizeof(ifs[0]);
	if (ioctl(pf_fd, DIOCIGETIFACES, &iface) == -1) {
		/*
		 * The ioctl is there, but it fails. Probably due to a too-old
		 * pf(4) implementation? Who knows. Let's accept this as a soft
		 * error and continue regardless. Once it works, it'll work in
		 * XSI as well.
		 */
		if (errno != ENODEV)
			err(1, "%s: Unable to get interface stats from pf(4)", __func__);
		return;
	}

	/* Find interfaces that have seen traffic */
	for (i = 0; i < iface.pfiio_size; i++) {
		int j;

		for (j = 0; j < 8; j++) {
			if (ifs[i].pfik_packets[j&1?0:1][j&2?0:1][j&4?0:1] != 0)
				ifs_with_traffic.push_back(i);
		}
	}
	ifs_with_traffic.sort();
	ifs_with_traffic.unique();

	if (ifs_with_traffic.size() != old_fIfStats.size())
		drop_old_stats = true;

	/* Get counters from ifs_with_traffic: bytes/packets[AF][direction][op] */
	do {
		filteredIfs_changed = false;
		if (drop_old_stats) {
			vector<FilteredIfStat>	fIfStats;

			for (iwt_it = ifs_with_traffic.begin();
			     iwt_it != ifs_with_traffic.end();
			     iwt_it++) {
				FilteredIfStat iwt(ifs[*iwt_it].pfik_name);

				iwt.setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				iwt.setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				iwt.setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				iwt.setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				iwt.setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				iwt.setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				iwt.setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				iwt.setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				iwt.setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				iwt.setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				iwt.setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				iwt.setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				iwt.setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				iwt.setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				iwt.setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				iwt.setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);

				fIfStats.push_back(iwt);
			}
			packetFilter.setFilteredIfStats(fIfStats);
		} else {
			vector<FilteredIfStat>::iterator it;

			for (it = old_fIfStats.begin(), iwt_it = ifs_with_traffic.begin();
			     it != old_fIfStats.end(); it++, iwt_it++) {
				if (strcmp(ifs[*iwt_it].pfik_name, it->getIfName()) != 0) {
					filteredIfs_changed = true;
					drop_old_stats = true;
					break;
				}
				it->setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				it->setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				it->setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				it->setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				it->setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				it->setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				it->setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				it->setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				it->setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				it->setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				it->setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				it->setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				it->setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				it->setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				it->setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				it->setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);
			}
			if (!filteredIfs_changed)
				packetFilter.setFilteredIfStats(old_fIfStats);
		}
	} while (filteredIfs_changed);
}

/*
 * Rudimentary sensors "support", which boils down to getting all ACPI
 * temperature sensors. There isn't anything else that is usable.
 */
/* Quoting /usr/src/sys/dev/acpica/acpi_thermal.c:
 * #define TZ_THFLAG_NONE	0
 * #define TZ_THFLAG_PSV	(1<<0)
 * #define TZ_THFLAG_HOT	(1<<2)
 * #define TZ_THFLAG_CRT	(1<<3)
 * ... which makes me hope that the following is good to use here as well.
 * I wish this were properly documented. :-(
 */
#define ACPI_TEMP_NONE	0
#define ACPI_TEMP_OK	(1<<0)
#define ACPI_TEMP_WARN	(1<<2)
#define ACPI_TEMP_CRIT	(1<<3)
void
XSI::setSensors(void)
{
	int	i;

	if (!sensors.empty()) {
		vector<Sensor *>::iterator it;

		for (it = sensors.begin(); it != sensors.end(); it++)
			delete *it;
		sensors.clear();
	}

	for (i = 0; i < 256; i++) {
		char	buf[BUFSIZ];
		int	temp, flags;
		size_t	siz = sizeof(temp);
		Sensor *s = NULL;
		enum SensorStatus status;

		snprintf(buf, sizeof(buf),
			 "hw.acpi.thermal.tz%d.temperature", i);
		if (sysctlbyname(buf, &temp, &siz, NULL, 0) == -1) {
			if (errno == ENOENT)
				break;
			else
				err(1, "%s: sysctlbyname", __func__);
		}
		snprintf(buf, sizeof(buf),
			 "hw.acpi.thermal.tz%d.thermal_flags", i);
		if (sysctlbyname(buf, &flags, &siz, NULL, 0) == -1)
			err(1, "%s: sysctlbyname", __func__);

		if (flags & ACPI_TEMP_CRIT)
			status = XMLSYS_SENSOR_STATUS_CRIT;
		else if (flags & ACPI_TEMP_WARN)
			status = XMLSYS_SENSOR_STATUS_WARN;
		else if (flags & ACPI_TEMP_OK)
			status = XMLSYS_SENSOR_STATUS_OK;
		else if (flags == ACPI_TEMP_NONE)
			status = XMLSYS_SENSOR_STATUS_UNSPEC;
		else
			status = XMLSYS_SENSOR_STATUS_UNKNOWN;

		snprintf(buf, sizeof(buf), "tz%d", i);
		s = new TemperatureSensor(i, "acpi", temp * 100000, status, buf);
		if (s == NULL)
			errx(1, "%s: Out of memory", __func__);
		sensors.push_back(s);
	}
}

void
XSI::getNetLLAddr(XSI_NetIfStat &netIfStat, struct sockaddr_dl *sdl)
{
	if (sdl->sdl_alen) {
		switch (sdl->sdl_type) {
		case IFT_ETHER:
		case IFT_CARP:
			netIfStat.setLladdrString(ether_ntoa((struct ether_addr *)LLADDR(sdl)));
			return;
		default:
			break;
		}
	}
}

void
XSI::getNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr,
		  struct sockaddr_dl *sdl)
{
	struct ifreq	ifr, *ifrp;
	int		sock;
	struct carpreq	carpr;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		err(1, "%s: socket", __func__);

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), ifaddr->ifa_addr->sa_len));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));

	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);

	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);

	if (sdl->sdl_alen && sdl->sdl_type == IFT_CARP) {
		/* It's fish! */
		bzero(&carpr, sizeof(carpr));
		ifr.ifr_data = (caddr_t)&carpr;

		if (ioctl(sock, SIOCGVH, (caddr_t)&ifr) != -1 &&
		    carpr.carpr_vhid > 0) {
			int		preempt, arpbalance;
			int		mib[4];
			size_t		siz = sizeof(int);
			enum xsi_carp_state state;

			if (carpr.carpr_state > CARP_MAXSTATE)
				state = CARP_UNKNOWN;
			else if (strcmp(carp_states[carpr.carpr_state], "INIT") == 0)
				state = CARP_INIT;
			else if (strcmp(carp_states[carpr.carpr_state], "BACKUP") == 0)
				state = CARP_BACKUP;
			else if (strcmp(carp_states[carpr.carpr_state], "MASTER") == 0)
				state = CARP_MASTER;
			else
				state = CARP_UNKNOWN;

			mib[0] = CTL_NET;
			mib[1] = PF_INET;
			mib[2] = IPPROTO_CARP;
			mib[3] = CARPCTL_PREEMPT;
			if (sysctl(mib, 4, &preempt, &siz, NULL, 0) == -1)
				err(1, "%s: sysctl", __func__);
			mib[3] = CARPCTL_ARPBALANCE;
			if (sysctl(mib, 4, &arpbalance, &siz, NULL, 0) == -1)
				err(1, "%s: sysctl", __func__);

#if defined(HAVE_CARPREQ_CARPR_CARPDEV)
			netIfStat.setCARPStatus(carpr.carpr_carpdev[0] ?
						carpr.carpr_carpdev :
						"none",
#else
			netIfStat.setCARPStatus(NULL,
#endif /* HAVE_CARPREQ_CARPR_CARPDEV */
						state, carpr.carpr_vhid,
						carpr.carpr_advskew,
						carpr.carpr_advbase,
						preempt == 0 ? 0 : 1,
						arpbalance == 0 ? 0 : 1);
		}
	}

	close(sock);
}

void
XSI::getNetIfmediaInfo(XSI_NetIfStat &netIfStat)
{
	struct ifmediareq	ifmr;
	int			sock, *media_list;
	const char             *ifname = netIfStat.getIfName();

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		err(1, "%s: socket", __func__);

	bzero(&ifmr, sizeof(ifmr));
	strlcpy(ifmr.ifm_name, ifname, sizeof(ifmr.ifm_name));
	if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
		// Interface doesn't support media directly
		close(sock);
		return;
	} else {
		string mediaStr;

		if (ifmr.ifm_count == 0) {
			warnx("%s: %s: no media types?", __func__, ifname);
			close(sock);
			return;
		}

		media_list = new int[ifmr.ifm_count];
		ifmr.ifm_ulist = media_list;
		if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
			close(sock);
			err(1, "%s: SIOCGIFMEDIA", __func__);
		}

		mediaStr = getMediaWordString(ifmr.ifm_current, 1);
		if (ifmr.ifm_active != ifmr.ifm_current) {
			mediaStr += " (";
			mediaStr += getMediaWordString(ifmr.ifm_active, 0);
			mediaStr += ")";
		}

		netIfStat.setMediaString(mediaStr.c_str());

#ifdef IFM_STATUS_VALID_LIST
		if (ifmr.ifm_status & IFM_AVALID) {
			const struct ifmedia_status_description *ifms;
			int bitno, found = 0;

			for (bitno = 0; ifm_status_valid_list[bitno] != 0; bitno++) {
				for (ifms = ifm_status_descriptions;
				     ifms->ifms_valid != 0; ifms++) {
					if (ifms->ifms_type !=
					    IFM_TYPE(ifmr.ifm_current) ||
					    ifms->ifms_valid !=
					    ifm_status_valid_list[bitno])
						continue;
					netIfStat.setStatusString(IFM_STATUS_DESC(ifms, ifmr.ifm_status));
					if (ifms->ifms_bit & ifmr.ifm_status)
						netIfStat.setLinkStatus(IF_LINK_ACTIVE);
					else
						netIfStat.setLinkStatus(IF_LINK_NOCARRIER);
					found = 1;
					break;
				}
			}
			if (!found) {
				netIfStat.setStatusString("unknown");
				netIfStat.setLinkStatus(IF_LINK_UNKNOWN);
			}
		}
#endif

		delete[] media_list;
	}

	close(sock);
}

void
XSI::getNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_LINK)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					(size_t)ifa->ifa_addr->sa_len,
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    ifa->ifa_dstaddr->sa_len);
			break;
		default:
			if (!have_warned) {
				warnx("Unhandled address family %d found, please report this!",
				      ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
void
XSI::getDefaultRoutes(void)
{
	int			mib[6];
	size_t			siz;
	char		       *buf, *next, *ep = NULL;
	struct rt_msghdr       *rtm;
	struct sockaddr        *sa, *rti_info[RTAX_MAX];
	char			host[MAXHOSTNAMELEN];
	int			error;
	vector<struct default_route> defRoutes;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;	// protocol number
	mib[3] = 0;	// AF
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_GATEWAY;
	if (sysctl(mib, 6, NULL, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	if (siz == 0)
		return;

	buf = new char[siz];
	if (buf == NULL)
		errx(1, "%s: Out of memory", __func__);
	if (sysctl(mib, 6, buf, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	ep = buf + siz;
	for (next = buf; next < ep; next += rtm->rtm_msglen) {
		unsigned int		i;
		struct default_route	defRt;

		rtm = (struct rt_msghdr *)next;
		sa = (struct sockaddr *)(rtm + 1);
		for (i = 0; i < RTAX_MAX; i++) {
			if (rtm->rtm_addrs & (1 << i)) {
				rti_info[i] = sa;
				sa = (struct sockaddr *)((char *)(sa) +
							 ROUNDUP(sa->sa_len));
			} else
				rti_info[i] = NULL;
		}
		if (rti_info[RTAX_DST] == NULL || rti_info[RTAX_GATEWAY] == NULL)
			continue;
		if (rti_info[RTAX_DST]->sa_family == AF_INET) {
			if (((struct sockaddr_in *)rti_info[RTAX_DST])->sin_addr.s_addr == INADDR_ANY)
				defRt.family = XSI_RTAF_IPV4;
			else
				continue;
		} else if (rti_info[RTAX_DST]->sa_family == AF_INET6) {
			struct sockaddr_in6 sin6;
			bool zero_mask = true;;

			sin6 = *((struct sockaddr_in6 *)rti_info[RTAX_DST]);
			if (rti_info[RTAX_NETMASK]) {
				struct sockaddr_in6 mask;
				unsigned int lim, final = 0, illegal = 0;
				unsigned char *p;

				mask = *((struct sockaddr_in6 *)rti_info[RTAX_NETMASK]);
				lim = mask.sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
				lim = lim < sizeof(struct in6_addr) ? lim : sizeof(struct in6_addr);
				for (p = (unsigned char *)&mask.sin6_addr, i = 0; i < lim; p++) {
					int masks[9] = {0x00, 0x80, 0xc0, 0xe0,
							0xf0, 0xf8, 0xfc, 0xfe,
							0xff};
					int j;

					if (final && *p) {
						illegal++;
						sin6.sin6_addr.s6_addr[i++] = 0x00;
						continue;
					}

					for (j = 0; j < 9; j++) {
						if (*p & 0xff == masks[j])
							break;
					}

					if (j >= 1 && j <= 8) {
						zero_mask = false;
						break;
					}
					if (j == 9)
						illegal++;
					if (j != 8)
						final++;
					if (!illegal)
						sin6.sin6_addr.s6_addr[i++] &= *p;
					else
						sin6.sin6_addr.s6_addr[i++] = 0x00;
				}
				while (i < sizeof(struct in6_addr))
					sin6.sin6_addr.s6_addr[i++] = 0x00;
			} else
				zero_mask = false;

			if (zero_mask && IN6_IS_ADDR_UNSPECIFIED(&sin6.sin6_addr))
				defRt.family = XSI_RTAF_IPV6;
			else
				continue;
		} else
			continue;
		if ((error = getnameinfo(rti_info[RTAX_GATEWAY], rti_info[RTAX_GATEWAY]->sa_len,
					 host, sizeof(host), NULL, 0, NI_NUMERICHOST)) != 0)
			errx(1, "%s: getnameinfo: %s", __func__, gai_strerror(error));
		defRt.gateway = host;
		defRoutes.push_back(defRt);
	}

	network.setDefaultRoutes(defRoutes);

	delete[] buf;
}

/* convenience functions for retrieving information via sysctl */
u_int
XSI::getmib(int *name, va_list ap)
{
	u_int len;

	for(len = 0; CTL_MAXNAME > len; len++) {
		name[len] = va_arg(ap, int);
		if(0 > name[len])
			break;
	}

	return len;
}

int
XSI::sysctlStr(char **val, ...)
{
	int name[CTL_MAXNAME];
	u_int namelen;
	va_list ap;
	size_t len;

	va_start(ap, val);
	namelen = getmib(name, ap);
	va_end(ap);

	if(0 > sysctl(name, namelen, NULL, &len, NULL, 0))
		return 0;
	*val = new char[len];

	if(0 > sysctl(name, namelen, *val, &len, NULL, 0)) {
		delete[] *val;
		return 0;
	}

	return 1;
}

int
XSI::sysctlFixed(void *val, size_t len, ...)
{
	int name[CTL_MAXNAME];
	u_int namelen;
	va_list ap;

	va_start(ap, len);
	namelen = getmib(name, ap);
	va_end(ap);

	if(0 > sysctl(name, namelen, val, &len, NULL, 0))
		return 0;
	return 1;
}

int
XSI::sysctlFixed(void *val, size_t len, const char *name)
{
	if(0 > sysctlbyname(name, val, &len, NULL, 0))
		return 0;
	return 1;
}

/*
 * sysctl_dynread(), adapted from FreeBSD 6.1:
 * $FreeBSD: src/usr.bin/systat/fetch.c,v 1.5 2001/12/12 00:13:37 markm Exp $
 *
 * Returns a pointer to a freshly allocated memory that holds the requested
 * data. If szp is not NULL, the size of the returned data is written into it.
 */
/* Number of tries */
#define SD_NTRIES	10
/* Percent of over-allocation (initial) */
#define SD_MARGIN	10
/* Increase margin by this on failure */
#define SD_FACTOR	50
/* Maximum supported MIB depth */
#define SD_MAXMIB	16
char *
XSI::sysctlDynamic(const char *name, size_t *szp)
{
	char   *rv = NULL;
	int	mib[SD_MAXMIB];
	size_t	mibsz = SD_MAXMIB;
	size_t	mrg = SD_MARGIN;
	size_t	sz;
	int	i;

	if (sysctlnametomib(name, mib, &mibsz) == -1)
		err(1, "%s: sysctlnametomib", __func__);
	for (i = 0; i < SD_NTRIES; i++) {
		/* get needed buffer size */
		if (sysctl(mib, mibsz, NULL, &sz, NULL, 0) == -1)
			err(1, "%s: sysctl", __func__);
		sz += sz * mrg / 100;
		if ((rv = new char[sz]) == NULL)
			errx(1, "%s: Out of memory", __func__);
		if (sysctl(mib, mibsz, rv, &sz, NULL, 0) == -1) {
			delete[] rv;
			rv = NULL;
			if (errno == ENOMEM) {
				mrg += mrg * SD_FACTOR / 100;
			} else
				err(1, "%s: sysctl", __func__);
		} else {
			/* success */
			if (szp != NULL)
				*szp = sz;
			break;
		}
	}

	return rv;
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:print_media_word()
 */
/* IFM_TYPE_MATCH from OpenBSD net/if_media.h */
#ifndef IFM_TYPE_MATCH
#define IFM_TYPE_MATCH(dt, t)						\
	(IFM_TYPE((dt)) == 0 || IFM_TYPE((dt)) == IFM_TYPE((t)))
#endif /* !IFM_TYPE_MATCH */
string
XSI::getMediaWordString(int ifmw, int print_type)
{
	const struct ifmedia_description *desc;
	int	seen_option = 0;
	char	tbuf[SMALLBUFSIZ];
	string	ret = "";
	enum {
		NONE,
		ETHERNET,
		TOKENRING,
		FDDI,
		IEEE80211,
		ATM,
		CARP,
		OTHER
	} media_type = NONE;

	tbuf[0] = '\0';

	{
		bool found = false;

		for (desc = ifm_type_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE(ifmw) == desc->ifmt_word) {
				switch (desc->ifmt_word) {
				case IFM_ETHER:
					media_type = ETHERNET;
					break;
				case IFM_TOKEN:
					media_type = TOKENRING;
					break;
				case IFM_FDDI:
					media_type = FDDI;
					break;
				case IFM_IEEE80211:
					media_type = IEEE80211;
					break;
				case IFM_ATM:
					media_type = ATM;
					break;
				case IFM_CARP:
					media_type = CARP;
					break;
				default:
					media_type = OTHER;
					break;
				}
				found = true;
				if (print_type)
					snprintf(tbuf, sizeof(tbuf), "%s ", desc->ifmt_string);
				break;
			}
		}
		if (!found && print_type)
			return ("unknown");
	}
	ret += tbuf;

	if (media_type == NONE)
		return (ret);

	{
		bool found = false;

		if (media_type == ETHERNET) {
			for (desc = ifm_subtype_ethernet_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		} else if (media_type == TOKENRING) {
			for (desc = ifm_subtype_tokenring_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		} else if (media_type == FDDI) {
			for (desc = ifm_subtype_fddi_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		} else if (media_type == IEEE80211) {
			for (desc = ifm_subtype_ieee80211_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		} else if (media_type == ATM) {
			for (desc = ifm_subtype_atm_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		}
		if (!found) {
			for (desc = ifm_subtype_shared_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
					break;
				}
			}
		}
		if (!found)
			strlcpy(tbuf, "unknown subtype", sizeof(tbuf));
	}
	ret += tbuf;

	/* Find mode. */
	if (IFM_MODE(ifmw) != 0) {
		bool found = false;

		if (media_type == IEEE80211) {
			for (desc = ifm_subtype_ieee80211_mode_descriptions; desc->ifmt_string != NULL;
			     desc++) {
				if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
				    IFM_MODE(ifmw) == IFM_MODE(desc->ifmt_word)) {
					found = true;
					snprintf(tbuf, sizeof(tbuf), " mode %s", desc->ifmt_string);
					break;
				}
			}
		}
		if (found)
			ret += tbuf;
	}

	/* Find options. */
	if (media_type == TOKENRING) {
		for (desc = ifm_subtype_tokenring_option_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
			    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
				if (seen_option == 0)
					ret += " ";
				snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
				ret += tbuf;
				seen_option |= IFM_OPTIONS(desc->ifmt_word);
			}
		}
	} else if (media_type == FDDI) {
		for (desc = ifm_subtype_fddi_option_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
			    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
				if (seen_option == 0)
					ret += " ";
				snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
				ret += tbuf;
				seen_option |= IFM_OPTIONS(desc->ifmt_word);
			}
		}
	} else if (media_type == IEEE80211) {
		for (desc = ifm_subtype_ieee80211_option_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
			    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
				if (seen_option == 0)
					ret += " ";
				snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
				ret += tbuf;
				seen_option |= IFM_OPTIONS(desc->ifmt_word);
			}
		}
	} else if (media_type == ATM) {
		for (desc = ifm_subtype_atm_option_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
			    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
				if (seen_option == 0)
					ret += " ";
				snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
				ret += tbuf;
				seen_option |= IFM_OPTIONS(desc->ifmt_word);
			}
		}
	} else {
		for (desc = ifm_shared_option_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
			    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
				if (seen_option == 0)
					ret += " ";
				snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
				ret += tbuf;
				seen_option |= IFM_OPTIONS(desc->ifmt_word);
			}
		}
	}
	if (IFM_INST(ifmw) != 0) {
		snprintf(tbuf, sizeof(tbuf), " instance %d", IFM_INST(ifmw));
		ret += tbuf;
	}

	return (ret);
}
