/*	$Kolabore: linux.cpp,v 1.52 2006/10/26 15:28:03 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * linux.cpp:
 * Gathering information and stats on Linux 2.6(.16?) or newer.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/sysinfo.h>

#include <inttypes.h>
#include <stdint.h>
/*
 * Note to self: The following would be required to allow inclusion of
 * <linux/ethtool.h>. We can't use it to get network interface media
 * information, because it's restricted to root. It is also sad that
 * even if we could use it, not all drivers support it. :-(
 */
/*
#ifndef u8
#  define u8 uint8_t
#endif
#ifndef u16
#  define u16 uint16_t
#endif
#ifndef u32
#  define u32 uint32_t
#endif
#ifndef u64
#  define u64 uint64_t
#endif
*/

// #include <linux/fs.h> // NTS; required for ioctl(BLKSSZGET)
// #include <linux/ethtool.h> // NTS; required for struct ethtool_cmd
#include <linux/if.h>
#include <linux/major.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
// #include <linux/sockios.h> // NTS; required for ioctl(SIOCETHTOOL)

#include <net/if_arp.h>

#include <netinet/ether.h>
#include <netinet/in.h>
#include <arpa/inet.h> // This one should go away eventually. IPv4-only == bad.

#include <dirent.h>
#include <err.h>
#include <fcntl.h>
#include <fts.h>
#include <ifaddrs.h>
#include <mntent.h>
#include <unistd.h>
}

#include <cctype>
#include <cerrno>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <fstream>
#include <list>
#include <string>
#include <vector>

#include "linux.h"
#include "compat/xsicompat.h"

#ifndef MIN
#  define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#endif /* MIN */

#ifndef SA_LEN
#  ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
#    define SA_LEN(x)	((x)->sa_len)
#  else
#    define SA_LEN(x)	((x)->sa_family == AF_INET6 ? \
			sizeof(struct sockaddr_in6) : \
			sizeof(struct sockaddr_in))
#  endif /* HAVE_STRUCT_SOCKADDR_SA_LEN */
#endif /* SA_LEN */

#define SEQNUM				0xdeadbeef
#define FILESYSTEM_DESCRIPTION_FILE	"/etc/mtab"

#define XSI_CP_USER	0
#define XSI_CP_NICE	1
#define XSI_CP_SYS	2
#define XSI_CP_IDLE	3
#define XSI_CP_IOWAIT	4
#define XSI_CP_IRQ	5
#define XSI_CP_SOFTIRQ	6
#define XSI_CPUSTATES	7

/* Most interesting sector size (mostly harddisks) */
#define XSI_SECTOR_SIZE 512

#define SENSOR_ALARM_UNSPEC	0x00
#define SENSOR_ALARM_OK 	0x01
#define SENSOR_ALARM_WARN	0x02
#define SENSOR_ALARM_CRIT	0x04

using namespace std;

/*
 * Structs to locally remember some things ...
 */
struct xsi_io_data {
	char		name[32];
	uint64_t	rbytes;
	uint64_t	wbytes;
	uint32_t	major;
};

struct xsi_intr_data {
	string		source;
	uint64_t	count;
};

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	num_cpu = -1;
	rtnl_fd = -1;
}

XSI::~XSI(void)
{
	int i;

	/* allocated in linuxInit() */
	for (i = 0; i < num_cpu; i++) {
		delete[] cur_cp_time[i];
		delete[] last_cp_time[i];
	}
	delete[] cur_cp_time;
	delete[] last_cp_time;
	if (rtnl_fd != -1)
		close(rtnl_fd);
}

/*
 * Devices to be opened before dropping privileges.
 * Silently returns 0 on success. Warns and returns >0 on error.
 *
 * NOTE: Valid XML must not rely on information gathered through privileged
 *       devices! Only optional, user-requested features may depend on
 *       things that may need super-user privileges.
 */
int
XSI::tryOpenPrivilegedDevices(void)
{
	if (configuration->showRaid()) {
		warnx("Sorry, RAID monitoring support is not (yet?) available on Linux.");
		return (1);
	}

	if (configuration->showPacketFilter()) {
		warnx("Sorry, packet filter statistics are not (yet?) available on Linux.");
		return (1);
	}

	return (0);
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	linuxInit();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateCPUStats();
	updateSystemCPUFreq();
	updateSystemSysInfo();
	updateLoadAvg();
	updateInterrupts();
	updateSensors();
	updateFSStats();
	updateDiskStats();
	updateNetStats();
	updateDefaultRoutes();
	// updateNetfilter();
}

void
XSI::linuxInit(void)
{
	static bool firstrun = true;

	if (firstrun) {
		int i, j, flags;

		firstrun = false;

		if ((rtnl_fd = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE)) == -1)
			err(1, "%s: socket", __func__);
		flags = fcntl(rtnl_fd, F_GETFL);
		if (flags < 0 || fcntl(rtnl_fd, F_SETFL, O_NONBLOCK | flags) == -1)
			err(1, "%s: fcntl", __func__);
		readSystemNumCPUs();
		cur_cp_time = new int64_t*[num_cpu];
		last_cp_time = new int64_t*[num_cpu];
		if (cur_cp_time == NULL || last_cp_time == NULL)
			errx(1, "%s: Out of memory", __func__);
		for (i = 0; i < num_cpu; i++) {
			cur_cp_time[i] = new int64_t[XSI_CPUSTATES];
			last_cp_time[i] = new int64_t[XSI_CPUSTATES];
			if (cur_cp_time[i] == NULL || last_cp_time[i] == NULL)
				errx(1, "%s: Out of memory", __func__);
			for (j = 0; j < XSI_CPUSTATES; j++) {
				cur_cp_time[i][j] = 0;
				last_cp_time[i][j] = 0;
			}
		}
	}
}

void
XSI::readSystemNumCPUs(void)
{
	DIR	       *dirp;
	FTS	       *ftsp;
	FTSENT	       *p;
	const char     *path[2] = { "/sys/devices/system/cpu/", NULL };
	int		ncpu = 0;

	if ((dirp = opendir(path[0])) == NULL)
		err(1, "%s: %s", __func__, path[0]);
	else
		closedir(dirp);

	if ((ftsp = fts_open(const_cast<char *const *>(path), 0, NULL)) == NULL)
		err(1, "%s: fts_open(\"%s\")", __func__, path[0]);
	while ((p = fts_read(ftsp)) != NULL) {
		switch (p->fts_info) {
		case FTS_D:
			if (strlen(p->fts_name) < 4
			    || strncmp(p->fts_name, "cpu", 3) != 0
			    || !isdigit(p->fts_name[3]))
				break;
			ncpu++;
			break;
		case FTS_ERR:
			errx(1, "%s: fts_read: %s", __func__,
			     strerror(p->fts_errno));
		default:
			break;
		}
	}
	fts_close(ftsp);

	if (ncpu == 0)
		errx(1, "%s: Something is very wrong, this system has no CPU?!\nPlease file a bug report.",
		     __func__);
	num_cpu = ncpu;
	system.setNumCPUs(ncpu);
}

void
XSI::readSystemOSInfo(void)
{
	struct utsname  name;
	string          tmp, os_type, os_release, os_version, os_kernel,
		        os_version_string;
	vector<string>	words;
	size_t		i;
	char		buf[BUFSIZ];
	struct tm	tm;

	if (uname(&name) == -1)
		err(1, "%s: uname", __func__);

	os_type = name.sysname;
	os_release = name.release;
	os_version = name.version;

	/*
	 * Here goes, the nasty part of trying to normalize a Linux
	 * version string: Split it up into words, take the last
	 * 7 words, throw away the timezone string, put them back
	 * together and parse the date.
	 * The first words are the actual version, which also
	 * need processing and finally put it all back together.
	 * The reason for this is that Linux kernels usually don't
	 * have a name ... except when they do. And then, the name
	 * appears as a separate word - unlike in BSD-land where it's
	 * always <name>#<number> - after the version number.
	 */

	/* split string into a vector of words */
	words = split(os_version);

	/* Let's make an assertion here ... */
	if (words.size() < 7)
		errx(1, "Sorry, you found a bug. Please report the complete error message:\n%s %s%s",
		     __func__,
		     "assertion failed, name.version is:\n",
		     name.version);

	/* reassemble a string parseable by strptime() */
	tmp = "";
	for (i = words.size() - 6; i < words.size(); i++) {
		if (i != words.size() - 2) // ignore timezone string
			tmp += words[i];
		if (i < words.size() - 2)
			tmp += " ";
	}

	/* parse and reformat, keep the result in buf */
	if (strptime(tmp.c_str(), "%a %b %e %H:%M:%S %Y", &tm) == NULL)
		errx(1, "Sorry, you found a bug. Please report the complete error message:\n%s %s%s",
		     __func__,
		     ": strptime(): assertion failed, name.version is:\n",
		     name.version);

	if (strftime(buf, sizeof(buf), "%Y%m%d%H%M", &tm) == 0)
		errx(1, "%s: strftime", __func__);

	/* concatenate all as-of-yet unprocessed words, except for the first */
	tmp = "";
	for (i = 1; i < words.size() - 6; i++) {
		tmp += words[i];
		if (i < words.size() - 7)
			tmp += "_";
	}

	/* finally, put it all back together */
	os_kernel = tmp;
	os_kernel += words[0];
	os_kernel += " ";
	os_kernel += buf;

	/* assemble an "uname -a"-like version string */
	os_version_string = os_type;
	os_version_string += " ";
	os_version_string += os_release;
	os_version_string += " ";
	os_version_string += os_version;
	os_version_string += " ";
	os_version_string += name.machine;

	system.setArch(name.machine);
	system.setOSInfo(os_type.c_str(), os_release.c_str(),
			 NULL, os_kernel.c_str(), os_version_string.c_str());
}

void
XSI::readNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr)
{
	struct ifreq	ifr, *ifrp;
	int		sock;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		err(1, "%s: socket", __func__);

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), SA_LEN(ifaddr->ifa_addr)));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));
	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);
	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);
	if (ioctl(sock, SIOCGIFHWADDR, (caddr_t)&ifr) != -1
	    && ifr.ifr_hwaddr.sa_family == ARPHRD_ETHER) {
		char	lladdr[24];

		memset(lladdr, 0, sizeof(lladdr));
		snprintf(lladdr, sizeof(lladdr), "%02x:%02x:%02x:%02x:%02x:%02x",
			 (ifr.ifr_hwaddr.sa_data[0] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[1] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[2] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[3] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[4] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[5] & 0xff));
		netIfStat.setLladdrString(lladdr);
	}

	/*
	 * The following code to get interface media information would work if
	 * we had root privileges. That, however, will never happen. Just keep
	 * it here commented out as a reminder until some day, maybe, this
	 * oh-so-confidential data becomes available to regular users.
	 */
	/*
	{
		struct ethtool_cmd ecmd;

		bzero(&ecmd, sizeof(ecmd));
		ecmd.cmd = ETHTOOL_GSET;
		ifr.ifr_data = (caddr_t)&ecmd;
		if (ioctl(sock, SIOCETHTOOL, (caddr_t)&ifr) != -1) {
			// do stuff
		}
	}
	*/

	close(sock);
}

void
XSI::readNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_PACKET)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					SA_LEN(ifa->ifa_addr),
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    SA_LEN(ifa->ifa_dstaddr));
			break;
		default:
			if (!have_warned) {
				warnx("Unhandled address family %d found, please report this!",
				      ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

void
XSI::readNetIfCounters(XSI_NetIfStat &netIfStat, vector<string> net_dev, string ifName)
{
	vector<string>::iterator i;
	string		if_counters;
	uint64_t	ibytes, obytes, ipkts, opkts, ierrs,
			oerrs, idrops, odrops, ocolls;

	for (i = net_dev.begin(); i != net_dev.end(); i++) {
		if (i->find(ifName, 0) < i->size()) {
			size_t pos = i->find(':');

			if (pos < i->size()) {
				if_counters = i->substr(pos + 1);
				break;
			}
		}
	}

	if (if_counters.empty())
		return;

	/*
	 * ibytes, ipkts, ierrs, idrops, ififo, iframe, icompr, imcast,
	 *   obytes, opkts, oerrs, odrops, ofifo, ocolls, ocarr, ocompr
	 */
	if (sscanf(if_counters.c_str(),
		   "%" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64
		   " %*s %*s %*s %*s %" PRIu64 " %" PRIu64 " %" PRIu64
		   " %" PRIu64 " %*s %" PRIu64 " %*s %*s",
		   &ibytes, &ipkts, &ierrs, &idrops,
		   &obytes, &opkts, &oerrs, &odrops,
		   &ocolls) != 9)
		errx(1, "%s: sscanf", __func__);

	netIfStat.setPacketsIn(ipkts);
	netIfStat.setPacketsOut(opkts);
	netIfStat.setErrorsIn(ierrs);
	netIfStat.setErrorsOut(oerrs);
	netIfStat.setBytesIn(ibytes);
	netIfStat.setBytesOut(obytes);
	netIfStat.setPacketsDropped(idrops + odrops);
	netIfStat.setCollisions(ocolls);
}

void
XSI::updateCPUStats(void)
{
	int64_t 	       *total;
	int64_t 	      **used;
	double		       *percent;
	vector<struct cpu>	cpus;
	int			i, j;
	static bool		load_per_cpu = true;
	vector<string>		proc_stat = getFileVector("/proc/stat");

	total = new int64_t[num_cpu];
	used = new int64_t*[num_cpu];
	percent = new double[num_cpu];
	if (total == NULL || used == NULL || percent == NULL)
		errx(1, "%s: Out of memory", __func__);
	for (i = 0; i < num_cpu; i++) {
		used[i] = new int64_t[XSI_CPUSTATES];
		if (used[i] == NULL)
			errx(1, "%s: Out of memory", __func__);
	}

	for (i = 0; i < num_cpu; i++) {
		struct cpu	cpu;
		vector<string>::iterator it;
		char		cpustr[16];
		bool		found_load_per_cpu = false;

		if (load_per_cpu) {
			snprintf(cpustr, sizeof(cpustr), "cpu%d ", i);
			for (it = proc_stat.begin(); it != proc_stat.end(); it++) {
				if (it->find(cpustr) == 0) {
					if (sscanf(it->c_str(),
						   "%*s %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %*s",
						   &cur_cp_time[i][XSI_CP_USER],
						   &cur_cp_time[i][XSI_CP_NICE],
						   &cur_cp_time[i][XSI_CP_SYS],
						   &cur_cp_time[i][XSI_CP_IDLE],
						   &cur_cp_time[i][XSI_CP_IOWAIT],
						   &cur_cp_time[i][XSI_CP_IRQ],
						   &cur_cp_time[i][XSI_CP_SOFTIRQ])
					    != 7)
						errx(1, "%s: sscanf", __func__);
					found_load_per_cpu = true;
				}
			}
			if (!found_load_per_cpu) {
				load_per_cpu = false;
			}
		}
		if (!load_per_cpu) {
			for (it = proc_stat.begin(); it != proc_stat.end(); it++) {
				if (it->find("cpu ") == 0) {
					if (sscanf(it->c_str(),
						   "%*s %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %*s",
						   &cur_cp_time[0][XSI_CP_USER],
						   &cur_cp_time[0][XSI_CP_NICE],
						   &cur_cp_time[0][XSI_CP_SYS],
						   &cur_cp_time[0][XSI_CP_IDLE],
						   &cur_cp_time[0][XSI_CP_IOWAIT],
						   &cur_cp_time[0][XSI_CP_IRQ],
						   &cur_cp_time[0][XSI_CP_SOFTIRQ])
					    != 7)
						errx(1, "%s: sscanf", __func__);
					break;
				}
			}
		}

		if (load_per_cpu) {
			total[i] = 0;
			for (j = 0; j < XSI_CPUSTATES; j++) {
				used[i][j] = cur_cp_time[i][j] - last_cp_time[i][j];
				last_cp_time[i][j] = cur_cp_time[i][j];
				total[i] += used[i][j];
			}
			if (total[i] > 0)
				percent[i] = 100 / (double)total[i];
			else
				percent[i] = 0;
			cpu.num = i;
			cpu.system = (used[i][XSI_CP_SYS] + used[i][XSI_CP_IOWAIT])
				* percent[i];
			cpu.user = used[i][XSI_CP_USER] * percent[i];
			cpu.nice = used[i][XSI_CP_NICE] * percent[i];
			cpu.interrupt = (used[i][XSI_CP_IRQ] + used[i][XSI_CP_SOFTIRQ])
				* percent[i];
			cpu.idle = used[i][XSI_CP_IDLE] * percent[i];
			cpus.push_back(cpu);
		} else {
			total[0] = 0;
			for (j = 0; j < XSI_CPUSTATES; j++) {
				used[0][j] = cur_cp_time[0][j] - last_cp_time[0][j];
				last_cp_time[0][j] = cur_cp_time[0][j];
				total[0] += used[0][j];
			}
			if (total[0] > 0)
				percent[0] = 100 / (double)total[0];
			else
				percent[0] = 0;
			cpu.num = -1;
			cpu.system = (used[0][XSI_CP_SYS] + used[0][XSI_CP_IOWAIT]) * percent[0];
			cpu.user = used[0][XSI_CP_USER] * percent[0];
			cpu.nice = used[0][XSI_CP_NICE] * percent[0];
			cpu.interrupt = (used[0][XSI_CP_IRQ] + used[0][XSI_CP_SOFTIRQ]) * percent[0];
			cpu.idle = used[0][XSI_CP_IDLE] * percent[0];
			cpus.push_back(cpu);
			break;
		}
	}

	system.setCPUs(cpus);

	for (i = 0; i < num_cpu; i++)
		delete[] used[i];
	delete[] used;
	delete[] total;
	delete[] percent;
}

/*
 * Enter the 3rd word of the 4th paragraph on page 7 of your manual to
 * find out the CPU frequency on Linux.
 */
void
XSI::updateSystemCPUFreq(void)
{
	static bool	have_cpufreq = true;
	vector<string>	cpuinfofile;
	vector<string>::iterator i;
	string		freq_line;
	double		divisor = 1.0;

	if (!have_cpufreq)
		return;

	cpuinfofile = getFileVector("/proc/cpuinfo");
	for (i = cpuinfofile.begin(); i != cpuinfofile.end(); i++) {
		size_t pos = i->size();

		if (i->find("cpu MHz", 0) < pos) {	// most systems, I hope
			freq_line = *i;
			break;
		} else if (i->find("cycle frequency [Hz]", 0) < pos) {	// alpha
			freq_line = *i;
			divisor = 1000000;
			break;
		} else if (i->find("clock", 0) < pos) {	// ppc
			freq_line = *i;
			break;
		}
	}

	if (!freq_line.empty()) {
		size_t pos = freq_line.find(':') + 2;
		double freq;
		int32_t iFreq;

		freq_line = freq_line.substr(pos);
		freq = strtod(freq_line.c_str(), NULL) / divisor;
		if (freq == 0 || freq == HUGE_VAL || freq == -HUGE_VAL) {
			warnx("%s: Error converting CPU frequency string `%s' to a number",
			      __func__, freq_line.c_str());
		} else {
			if ((iFreq = (int32_t)freq) == INT_MAX)
				warnx("%s: CPU frequency %f is unreasonably large", __func__, freq);
			else {
				double d;

				if (modf(freq, &d) >= 0.5)
					iFreq++;
				system.setCPUFreq(iFreq);
				return;
			}
		}
	}

	have_cpufreq = false;
	system.setCPUFreq(-1);
}

void
XSI::updateSystemSysInfo(void)
{
	struct sysinfo si;
	struct vm_info vm;

	if (sysinfo(&si) == -1)
		err(1, "%s: sysinfo", __func__);

	system.setOSUptime(si.uptime);
	vm.ram_total = si.totalram * si.mem_unit;
	vm.ram_free = si.freeram * si.mem_unit;
	vm.swap_total = si.totalswap * si.mem_unit;
	vm.swap_free = si.freeswap * si.mem_unit;
	system.setVMInfo(vm);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		err(1, "%s: getloadavg", __func__);

	system.setLoadAvg(v);
}

/*
 * Dear Linux kernel developers. Please stop with this insane filesystem-based
 * terror. Please fix sysctl(). PLEASE!
 * This is so sick.
 */
void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints;
	bool			drop_old_stats, ints_changed;
	vector<string>		intr_list;
	vector<string>::iterator si;
	vector<struct xsi_intr_data> interrupts;
	vector<struct xsi_intr_data>::iterator ii;

	if (num_cpu <= 0)
		errx(1, "%s: Internal error: num_cpu not set", __func__);

	{
		vector<string> tmp = getFileVector("/proc/interrupts");

		/* Keep only those lines with enough columns */
		for (si = tmp.begin(); si != tmp.end(); si++) {
			if ((int)split(*si).size() > num_cpu + 2)
				intr_list.push_back(*si);
		}
	}

	if (intr_list.empty())
		return;

	for (si = intr_list.begin(); si != intr_list.end(); si++) {
		struct xsi_intr_data intr;
		vector<string>	intrs = split(*si);
		int		i;

		/* Each CPU has a counter per interrupt source */
		intr.count = 0;
		for (i = 0; i < num_cpu; i++) {
			const char     *errstr;
			int64_t 	count;

			count = strtonum(intrs[i + 1].c_str(), 0, INT64_MAX, &errstr);
			if (errstr)
				err(1, "%s: %s is %s", __func__,
				    intrs[i + 1].c_str(), errstr);
			else if (count < 0)
				errx(1, "%s: Interrupt counter has a negative value?",
				     __func__);
			intr.count += count;
		}

		/*
		 * Interrupts can come from a list of devices. Thankfully,
		 * those lists already have a meaningful separator (', '),
		 * of which we'll only keep the comma.
		 */
		for (i = num_cpu + 2; i < (int)intrs.size(); i++)
			intr.source += intrs[i];

		interrupts.push_back(intr);
	}

	if (interrupts.empty())
		return;

	drop_old_stats = false;
	if ((old_ints = system.getInterrupts()).size() != interrupts.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = interrupts.begin(); ii != interrupts.end(); ii++) {
				XSI_Interrupt interrupt;

				interrupt.setSourceName(ii->source.c_str());
				interrupt.setIntsCount(ii->count);

				new_ints.push_back(interrupt);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < interrupts.size(); i++) {
				if (interrupts[i].source != old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount(interrupts[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

void
XSI::updateSensors(void)
{
	vector<struct linux_sensor_private> linux_sensors;
	vector<struct linux_sensor_private>::iterator si;

	{
		const char     *i2cdevs = "/sys/bus/i2c/devices";
		DIR	       *dirp;
		struct dirent  *d;
		vector<string>	dev_paths;

		if ((dirp = opendir(i2cdevs)) == NULL)
			/* No need to look any further */
			return;
		else {
			string path = i2cdevs;

			while ((d = readdir(dirp)) != NULL) {
				if (d->d_name[0] != '.')
					dev_paths.push_back(path + "/" + d->d_name);
			}
			closedir(dirp);
			linux_sensors = getSensors(dev_paths);
		}
	}

	if (!sensors.empty()) {
		vector<Sensor *>::iterator it;

		for (it = sensors.begin(); it != sensors.end(); it++)
			delete *it;
		sensors.clear();
	}

	for (si = linux_sensors.begin(); si != linux_sensors.end(); si++) {
		enum SensorStatus status;
		Sensor *s;

		if (si->has_fault)
			continue;

		if (si->alarm & SENSOR_ALARM_CRIT)
			status = XMLSYS_SENSOR_STATUS_CRIT;
		else if (si->alarm & SENSOR_ALARM_WARN)
			status = XMLSYS_SENSOR_STATUS_WARN;
		else if (si->alarm & SENSOR_ALARM_OK)
			status = XMLSYS_SENSOR_STATUS_OK;
		else
			status = XMLSYS_SENSOR_STATUS_UNSPEC;

		s = NULL;
		switch (si->type) {
		case VOLT:
			s = new VoltsDCSensor(si->id, si->driver.c_str(), si->value * 1000,
					      status, si->name.c_str());
			break;
		case TEMP:
			s = new TemperatureSensor(si->id, si->driver.c_str(), (si->value + 273150) * 1000,
						  status, si->name.c_str());
			break;
		case FAN:
			/* special case: can contain `-1' as nonsense value */
			if (si->value != -1)
				s = new FanRPMSensor(si->id, si->driver.c_str(),
						     si->value, status,
						     si->name.c_str());
			break;
		case CPUVOLT:
			s = new VoltsDCSensor(si->id, si->driver.c_str(), si->value * 1000,
					      status, si->name.c_str());
			break;
		case AMP:
			s = new AmpsSensor(si->id, si->driver.c_str(), si->value * 1000,
					   status, si->name.c_str());
			break;
		case OTHER:
			s = new GenericSensor(si->id, si->driver.c_str(), si->value,
					      status, si->name.c_str());
			break;
		default:
			errx(1, "%s: Internal error: Unknown sensor type", __func__);
		}

		if (s != NULL)
			sensors.push_back(s);
	}
}

void
XSI::updateFSStats(void)
{
	vector<struct fs>	filesystems;
	struct mntent	       *mnt;
	FILE		       *fp = NULL;

	if ((fp = fopen(FILESYSTEM_DESCRIPTION_FILE, "r")) == NULL)
		err(1, "%s: fopen", __func__);

	while ((mnt = getmntent(fp)) != NULL) {
		struct fs	fs;
		struct statvfs	f;

		fs.mountpoint = mnt->mnt_dir;
		fs.fs_name = mnt->mnt_fsname;
		fs.fs_type = mnt->mnt_type;

		if (statvfs(mnt->mnt_dir, &f) == -1) {
			if (errno != EACCES)
				err(1, "%s: statvfs", __func__);
			else
				continue;
		}

		fs.size_total = f.f_blocks * f.f_bsize;
		if (fs.size_total == 0 || fs.mountpoint.find("/dev") == 0)
			continue;	// ignore; boring stuff

		fs.size_free = f.f_bfree * f.f_bsize;
		fs.size_avail = f.f_bavail * f.f_bsize;
		fs.nodes_total = f.f_files;
		fs.nodes_free = f.f_ffree;

		filesystems.push_back(fs);
	}

	fclose(fp);

	storage.setFilesystems(filesystems);
}

void
XSI::updateDiskStats(void)
{
	vector<struct xsi_io_data> xsi_io_data;
	vector<struct xsi_io_data>::iterator ii;
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	vector<string>	        diskstats = getFileVector("/proc/diskstats");
	vector<string>::iterator si;
	bool			drop_old_stats = false, disks_changed;

	/*
	 * This is plain awful. I don't know how to figure out how many disks
	 * devices are actually in use and how they are called.
	 * This is all about making an educated guess and weeding out obviously
	 * uninteresting devices.
	 */
	for (si = diskstats.begin(); si != diskstats.end(); si++) {
		struct xsi_io_data io;
		uint64_t	sects_read, sects_written;
		// size_t		sectsiz;

		if (sscanf(si->c_str(),
			   "%u %*s %31s %*s %*s %*s %" PRIu64
			   " %*s %*s %*s %" PRIu64,
			   &io.major, io.name, &sects_read, &sects_written) != 4)
			continue;

		/* Ignore ramdisks and loopback devices */
		if (io.major == RAMDISK_MAJOR || io.major == LOOP_MAJOR)
			continue;

		/*
		 * The following commented code is similar to what we'd need to
		 * ensure that numbers are correct. Unfortunately, this requires
		 * superuser privileges and that is unacceptable.
		 */
		/*
		{
			int	fd;
			char	device[256];

			snprintf(device, sizeof(device), "/dev/%s", io.name);
			if ((fd = open(device, O_RDONLY)) == -1)
				err(1, "%s: open", __func__);
			if (ioctl(fd, BLKSSZGET, &sectsiz) == -1)
				err(1, "%s: BLKSSZGET", __func__);
			close(fd);
		}
		*/

		/* XXX: Not correct for all block devices ... */
		io.rbytes = sects_read * XSI_SECTOR_SIZE;
		io.wbytes = sects_written * XSI_SECTOR_SIZE;

		xsi_io_data.push_back(io);
	}

	if (xsi_io_data.size() != old_disks.size())
		drop_old_stats = true;

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (ii = xsi_io_data.begin(); ii != xsi_io_data.end(); ii++) {
				XSI_DiskIO disk(ii->name);
				disk.setBytesWritten(ii->wbytes);
				disk.setBytesRead(ii->rbytes);
				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			size_t i;

			for (i = 0; i < xsi_io_data.size(); i++) {
				if (strcmp(xsi_io_data[i].name,
					   old_disks[i].getDiskName()) != 0) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				old_disks[i].setBytesWritten(xsi_io_data[i].wbytes);
				old_disks[i].setBytesRead(xsi_io_data[i].rbytes);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);
}

/*
 * Looks like Linux has an undocumented, neutered getifaddrs().
 * Let's use it as much as possible, as it greatly simplifies
 * things.
 */
void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	vector<string>		net_dev = getFileVector("/proc/net/dev");
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		err(1, "%s: getifaddrs", __func__);

	/*
	 * Convenience; make a vector of pointers to link-layer (AF_PACKET on
	 * Linux) interfaces
	 */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_PACKET)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				readNetIfInfo(netIfStat, (*it));
				readNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				readNetIfCounters(netIfStat, net_dev, (*it)->ifa_name);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				readNetIfInfo(old_nstats[i], (*it));
				readNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				readNetIfCounters(old_nstats[i], net_dev, (*it)->ifa_name);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

/*
 * XXX -- Supports IPv4 default routes only
 */
void
XSI::updateDefaultRoutes(void)
{
	size_t			len;
	uint8_t 	       *buf;
	struct in_addr		addr;
	struct sockaddr_nl	snl;
	struct {
		struct nlmsghdr nlh;
		struct rtgenmsg rtg;
	} req;
	vector<struct default_route> defRoutes;

	memset(&snl, 0, sizeof(snl));
	snl.nl_family = AF_NETLINK;

	memset(&req, 0, sizeof(req));
	req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(req.rtg));
	req.nlh.nlmsg_type = RTM_GETROUTE;
	req.nlh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
	req.nlh.nlmsg_seq = SEQNUM;
	req.nlh.nlmsg_pid = 0;
	req.rtg.rtgen_family = AF_INET;

	if (sendto(rtnl_fd, &req, sizeof(req), 0,
		   (struct sockaddr *)&snl, sizeof(snl)) == -1)
		err(1, "%s: sendto", __func__);

	for (;;) {
		struct default_route	defRt;

		if ((buf = getNetlinkMessage(rtnl_fd, len)) == NULL)
			break;
		if (parseRoutes(buf, len, &addr) == 0) {
			defRt.family = XSI_RTAF_IPV4;
			defRt.gateway = inet_ntoa(addr);
			defRoutes.push_back(defRt);
		}
		delete[] buf;
	}

	network.setDefaultRoutes(defRoutes);
}

/*
 * XXX: This is just too hard to do without documentation for the netfilter
 * netlink system. Using /proc/net/ip_conntrack instead, even for the most
 * basic level of support, would be very bad.
 *
 * Here is an excerpt from a note, dated 04/17/05, by the author of a utility
 * called "IPTState", which shows netfilter's conntrack states in a top-like
 * manner:
 * ``[Harald Welte, Netfilter developer, informs users about using /proc/net
 * /ip_conntrack:] ..., it may be unreliable on SMP systems and can have a big
 * impact on the performance of your firewall.'' He also says that this proc
 * file is a ``broken interface'', which I cannot agree with any stronger.
 * Another issue with this is that the proc file may (dis-/re-)appear during
 * runtime long after we dropped root privileges (which we need to open it.)
 *
 * This means: Unless we can use NETLINK_NETFILTER, which is not documented
 * at all, this feature will not be available on Linux.
 */
void
XSI::updateNetfilter(void)
{
}

uint8_t *
XSI::getNetlinkMessage(int fd, size_t &buflen)
{
	uint8_t 	       *buf;
	size_t			len;
	ssize_t 		res;
	struct sockaddr_nl	snl;
	socklen_t		slen;

	len = BUFSIZ * 2;
	if ((buf = new uint8_t[len]) == NULL)
		err(1, "%s: new", __func__);

	do {
		memset(&snl, 0, sizeof(snl));
		slen = sizeof(snl);
		if ((res = recvfrom(fd, buf, len, 0, (struct sockaddr *)&snl, &slen)) == -1) {
			if (errno != EAGAIN)
				err(1, "%s: recvfrom", __func__);
			delete[] buf;
			buflen = 0;
			return (NULL);
		}
		if (sizeof(snl) > slen || snl.nl_family != AF_NETLINK) {
			err(1, "%s: Bad address", __func__);
			delete[] buf;
			buflen = 0;
			return (NULL);
		}
	} while (snl.nl_pid != 0);

	buflen = res;

	return (buf);
}

int
XSI::parseRoutes(uint8_t *buf, size_t len, struct in_addr *addr)
{
	struct nlmsghdr        *nlm;
	struct nlmsgerr        *nle;
	struct rtmsg	       *rtm;
	struct rtattr	       *rta;
	int			rtalen;
	struct in_addr		gw, dst;

	nlm = (struct nlmsghdr *)buf;
	while (NLMSG_OK(nlm, len)) {
		gw.s_addr = INADDR_ANY;
		dst.s_addr = INADDR_ANY;
		if (nlm->nlmsg_type == NLMSG_ERROR) {
			nle = (struct nlmsgerr *)NLMSG_DATA(nlm);
			if (NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct nlmsgerr))) >
			    nlm->nlmsg_len)
				errx(1, "%s: Netlink message truncated", __func__);
			else
				errx(1, "%s: Netlink error: %s", __func__, strerror(nle->error));
		} else if (nlm->nlmsg_type == RTM_NEWROUTE && nlm->nlmsg_seq == SEQNUM
			   && (pid_t) nlm->nlmsg_pid == getpid()
			   && nlm->nlmsg_len >= NLMSG_LENGTH(sizeof(struct rtmsg))) {
			rtm = (struct rtmsg *)NLMSG_DATA(nlm);
			rta = RTM_RTA(rtm);
			rtalen = RTM_PAYLOAD(nlm);

			while(RTA_OK(rta, rtalen)) {
				if (RTA_PAYLOAD( rta ) >= sizeof(struct in_addr)) {
					switch(rta->rta_type) {
					case RTA_GATEWAY:
						memcpy(&gw, RTA_DATA(rta), sizeof(gw));
						break;
					case RTA_DST:
						memcpy(&dst, RTA_DATA(rta), sizeof(dst));
						break;
					}
				}
				rta = RTA_NEXT(rta, rtalen);
			}
		}

		if (gw.s_addr != INADDR_NONE && gw.s_addr != INADDR_ANY
		    && dst.s_addr == INADDR_ANY) {
			*addr = gw;
			return (0);
		}

		nlm = NLMSG_NEXT(nlm, len);
	}

	return (1);
}

vector<struct linux_sensor_private>
XSI::getSensors(vector<string> dev_paths)
{
	vector<string>::iterator si;
	vector<char **> sensor_paths;
	vector<char **>::iterator cpi;
	vector<struct linux_sensor_private> linux_sensors;

	/*
	 * Create an array of arrays, to be used by fts_open(). While this
	 * could also be in one single array, this guarantees that we always
	 * get the sensor's device name without having to check it for each
	 * entry in the filesystem.
	 */
	for (si = dev_paths.begin(); si != dev_paths.end(); si++) {
		char **path;
		size_t siz = si->size() + 1;

		path = new char*[2];
		path[0] = new char[siz];
		strlcpy(path[0], si->c_str(), siz);
		path[1] = NULL;
		sensor_paths.push_back(path);
	}

	for (cpi = sensor_paths.begin(); cpi != sensor_paths.end(); cpi++) {
		FTS	       *ftsp;
		FTSENT	       *p;
		string		driver = getSysfsDeviceName(*cpi[0]);

		if ((ftsp = fts_open(const_cast<char *const *>(*cpi),
				     FTS_COMFOLLOW | FTS_NOSTAT, NULL)) == NULL)
			err(1, "%s: fts_open", __func__);
		while ((p = fts_read(ftsp)) != NULL) {
			string file;
			struct linux_sensor_private sensor;
			bool sensor_found = false;

			switch (p->fts_info) {
			case FTS_DNR:
			case FTS_ERR:
				errx(1, "%s: fts_read: %s", __func__,
				     strerror(p->fts_errno));
			case FTS_F:
				file = p->fts_name;
				/* *_input files contain sensor values */
				if (file.size() > 6
				    && file.find("_input") == file.size() - 6) {
					string path = getDirName(p->fts_path);
					string fault_file = path + file
						+ "_fault";
					string alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_alarm";
					string min_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_min_alarm";
					string max_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_max_alarm";
					string crit_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_crit_alarm";
					int fault, alarm;
					bool error;

					sensor.alarm = SENSOR_ALARM_UNSPEC;
					sensor.value = getValueFromFile(p->fts_path);
					sensor.type = getSensorType(file);

					fault = getValueFromFile(fault_file.c_str(),
								 &error, 0, 1);
					if (error)
						sensor.has_fault = false;
					else {
						if (fault == 1)
							sensor.has_fault = true;
						else
							sensor.has_fault = false;
					}

					alarm = getValueFromFile(alarm_file.c_str(),
								 &error, 0, 1);
					if (error) {
						alarm = getValueFromFile(min_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_WARN;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
						alarm = getValueFromFile(max_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_WARN;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
						alarm = getValueFromFile(crit_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_CRIT;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
					} else {
						if (alarm == 1)
							sensor.alarm |= SENSOR_ALARM_WARN;
						else
							sensor.alarm |= SENSOR_ALARM_OK;
					}

					sensor_found = true;
				} else if (file.size() > 4 && file.find("_vid") == file.size() - 4) {
					/* For unknown reasons, voltages in CPU cores get special treatments */
					sensor.value = getValueFromFile(p->fts_path);
					sensor.type = getSensorType(file);
					sensor_found = true;
				}
				break;
			default:
				break;
			}
			if (sensor_found) {
				getSensorID(file, sensor.name, sensor.id);
				sensor.driver = driver;
				linux_sensors.push_back(sensor);
			}
		}
		fts_close(ftsp);
		delete[] *cpi[0];
		delete[] *cpi;
	}

	return (linux_sensors);
}

enum linux_sensor_type
XSI::getSensorType(string name)
{
	if (name.find("in") == 0)
		return (VOLT);
	else if (name.find("temp") == 0)
		return (TEMP);
	else if (name.find("fan") == 0)
		return (FAN);
	else if (name.find("curr") == 0)
		return (AMP);
	else if (name.find("_vid") == name.size() - 4)
		return (CPUVOLT);
	else
		return (OTHER);
}

void
XSI::getSensorID(string filename, string &name, int &num)
{
	string::iterator	i;
	string			buf;
	const char	       *errstr;

	name = "";
	num = 0;
	for (i = filename.begin(); i != filename.end(), !isdigit(*i); i++)
		name += *i;
	while (isdigit(*i))
		buf += *(i++);
	num = strtonum(buf.c_str(), INT_MIN, INT_MAX, &errstr);
	if (errstr)
		err(1, "%s: %s is %s", __func__,
		    buf.c_str(), errstr);
	name += buf;
}

string
XSI::getSysfsDeviceName(const char *path)
{
	string namefile = path;

	namefile += "/name";

	return (getFileString(namefile.c_str()));
}

string
XSI::getDirName(const char *path)
{
	vector<string>	tmp = split(path, '/');
	vector<string>::iterator it;
	unsigned int	i;
	string		ret = "/";

	for (it = tmp.begin(), i = 0; it != tmp.end(), i < tmp.size() - 1; i++, it++)
		ret += *it + "/";

	return (ret);
}
