/*	$Kolabore: main.cpp,v 1.57 2006/10/29 14:19:26 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netinet/in.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#include <grp.h>
#include <netdb.h>
#include <poll.h>
#include <pwd.h>
#include <signal.h>
#include <unistd.h>

#ifdef HAVE_ZLIB_H
#  include <zlib.h>
#endif

#include <libxml/tree.h>
}

#include <cerrno>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <string>
#include <vector>

#include "cfg.h"
#include "interface.h"
#include "server.h"
#include "compat/xsicompat.h"

extern int	optind;
extern int	errno;

#ifndef HAVE___PROGNAME
char *__progname;
#endif /* !HAVE___PROGNAME */

volatile sig_atomic_t update = 0;
volatile sig_atomic_t   quit = 0;
volatile sig_atomic_t reread = 0;

static void sighdlr(int);
static unsigned char * zCompress(xmlChar *, size_t, unsigned long *);
static void zFree(unsigned char *);
static char * xsi_strdup(const char *);
static char * xsi_getProgname(char *);

using namespace std;

static void
sighdlr(int sig)
{
	switch (sig) {
	case SIGALRM:
		update = 1;
		break;
	case SIGHUP:
		reread = 1;
		break;
	default:
		quit = 1;
		break;
	}
}

/*
 * Wrapper for zlib's compress() function
 */
static unsigned char *
zCompress(xmlChar *src, size_t sLen, unsigned long *zLen)
{
	unsigned char *dest = NULL;

#ifdef HAVE_WORKING_ZLIB
	unsigned long dLen = compressBound(sLen);
	int ret;

	dest = new unsigned char[(size_t)dLen];
	if (dest == NULL)
		errx(1, "%s: Out of memory (new)", __func__);

	ret = compress((Bytef *)dest, (uLongf *)&dLen, (Bytef *)src, sLen);
	switch (ret) {
	case Z_MEM_ERROR:
		errx(1, "%s: Out of memory (compress)", __func__);
		/* NOTREACHED */
	case Z_BUF_ERROR:
		errx(1, "%s: Destination buffer too small", __func__);
		/* NOTREACHED */
	default:
		break;
	}

	*zLen = dLen;
#endif /* HAVE_WORKING_ZLIB */

	return (dest);
}

static void
zFree(unsigned char *buf)
{
	if (buf != NULL) {
		delete buf;
		buf = NULL;
	}
}

static char *
xsi_strdup(const char *s)
{
	char *c;
	size_t siz = strlen(s) + 1;

	if ((c = new char[siz]) == NULL) {
		fprintf(stderr, "%s: Out of memory.\n", __func__);
		exit(1);
	}
	strlcpy(c, s, siz);

	return(c);
}

/* borrowed from OpenNTPd-portable's compat-openbsd/bsd-misc.c */
static char *
xsi_getProgname(char *argv0)
{
#ifdef HAVE___PROGNAME
	extern char *__progname;

	return (xsi_strdup(__progname));
#else
	char *p;

	if (argv0 == NULL)
		return (const_cast<char *>("xsi"));
	p = strrchr(argv0, '/');
	if (p == NULL)
		p = argv0;
	else
		p++;

	return (xsi_strdup(p));
#endif /* HAVE___PROGNAME */
}

int
main(int argc, char **argv)
{
	Config		       *cfg;
	Server		       *server;
	XSI		       *sysinfo;
	xmlDocPtr		xmlDocument = NULL;
	xmlChar 	       *buffer = NULL;
	unsigned char	       *zBuffer = NULL;
	vector<struct client>	clients;
	struct itimerval	updateTimer;
	struct passwd	       *passwd;
	struct group	       *group;
	bool			may_talk = false;
	int			error = 0;
	unsigned long		request_counter = 0;
#if defined(INSANE_OS)
	const bool		insane = true;
#else
	const bool		insane = false;
#endif /* INSANE_OS */
	extern char	       *__progname;

	__progname = xsi_getProgname(argv[0]);

	LIBXML_TEST_VERSION;

	cfg = new Config(argc, argv);
	may_talk = cfg->dontDetach();

	sysinfo = new XSI;

	if (may_talk)
		warnx("[startup] %s", PACKAGE_STRING);
	sysinfo->XSIBase_init(cfg);
	if (sysinfo->tryOpenPrivilegedDevices() != 0)
		return (1);

	if (cfg->oneShot()) {
		sysinfo->readSysInfo();
		sysinfo->buildSysInfoDoc();
		xmlDocument = sysinfo->getSysInfoDocPtr();
		xmlSaveFormatFileEnc("-", xmlDocument, "UTF-8", 1);
		delete sysinfo;
		return (0);
	}

#if !defined(HAVE_WORKING_ZLIB)
	if (cfg->compressXml()) {
		cerr << "Sorry, your system's libz is too old or compression support was explicitly\n"
			<< "disabled. Install a recent libz that offers compressBound() and recompile\n"
			<< "this program." << endl;
		return (1);
	}
#endif /* !HAVE_WORKING_ZLIB */

	if (geteuid() == 0) {
		const char *gp = cfg->getPrivdropGroup();
		gid_t gid;

		if ((passwd = getpwnam(cfg->getPrivdropUser())) == NULL)
			errx(1, "%s: unknown user %s", __func__,
			     cfg->getPrivdropUser());
		if (passwd->pw_uid == 0)
			errx(1, "%s: %s refuses to run with super-user privileges",
			     __func__, __progname);

		if (gp != NULL) {
			if ((group = getgrnam(gp)) == NULL)
				errx(1, "%s: unknown group %s", __func__, gp);
			gid = group->gr_gid;
		} else
			gid = passwd->pw_gid;

		if (may_talk)
			warnx("[startup] looks like I got root ...");

		if (insane) {
			if (may_talk)
				warnx("[startup] CHROOT FEATURE NOT AVAILABLE");
		} else {
			if (chroot(cfg->getChrootDir()) == -1)
				err(1, "%s: chroot", __func__);
			if (chdir("/") == -1)
				err(1, "%s: chdir(\"/\")", __func__);
			if (may_talk)
				warnx("[startup] chroot to %s successful",
				      cfg->getChrootDir());
		}

		if (setgroups(1, &gid) ||
#if defined(HAVE_SETRESGID) && defined(HAVE_SETRESUID)
		    setresgid(gid, gid, gid) ||
		    setresuid(passwd->pw_uid, passwd->pw_uid, passwd->pw_uid))
#else
		    setregid(gid, gid) ||
		    setreuid(passwd->pw_uid, passwd->pw_uid))
#endif
			err(1, "%s: can't drop privileges", __func__);
		if (may_talk)
			warnx("[startup] successfully dropped privileges to UID %d, GID %d",
			      passwd->pw_uid, gid);
	} else {
		if (may_talk)
			warnx("[startup] running as regular user");
	}

	server = new Server(cfg);
	sysinfo->setListenAddrs(server->getListenAddrStrings());
	sysinfo->setListenService(server->getListenServiceString());

	if (!cfg->dontDetach()) {
		if (daemon(1, 0) == -1)
			err(1, "%s: daemon", __func__);
	}

	if (may_talk) {
		vector<string> listen_addrs = server->getListenAddrStrings();
		vector<string>::iterator it;
		string service = server->getListenServiceString();

		warnx("[config] update interval: %ld seconds",
		      cfg->getUpdateInterval());
		for (it = listen_addrs.begin(); it != listen_addrs.end(); it++) {
			warnx("[config] listening on [%s]:%s", it->c_str(),
			      service.c_str());
		}
		if (cfg->oneClientMode())
			warnx("[config] running in One-Client mode!");
	}

	signal(SIGTERM, sighdlr);
	signal(SIGINT, sighdlr);
	signal(SIGHUP, sighdlr);
	if (!cfg->oneClientMode()) {
		memset(&updateTimer, 0, sizeof(updateTimer));
		updateTimer.it_value.tv_sec = cfg->getUpdateInterval();
		updateTimer.it_interval.tv_sec = cfg->getUpdateInterval();
		setitimer(ITIMER_REAL, &updateTimer, NULL);
	}

	sysinfo->readSysInfo();

	while (!quit) {
		signal(SIGALRM, sighdlr);
		clients = server->poll();
		if (clients.size() > 0) {
			vector<struct client>::iterator it;
			int format = 0;

			if (cfg->oneClientMode())
				sysinfo->updateSysInfo();
			if (cfg->formatOutput() && !cfg->compressXml())
				format = 1;
			sysinfo->buildSysInfoDoc();
			xmlDocument = sysinfo->getSysInfoDocPtr();

			for (it = clients.begin(); it != clients.end(); it++) {
				ssize_t sent = 0;
				int	len;

				if (xmlDocument == NULL) {
					if (may_talk)
						warnx("[urggh] some kajiggerbadoo seriously screwed up");
					quit = 1;
					break;
				}

				xmlDocDumpFormatMemoryEnc(xmlDocument, &buffer,
							  &len, "UTF-8",
							  format);

				if (cfg->compressXml()) {
					unsigned long zLen;
					ssize_t bytes;

					zBuffer = zCompress(buffer, (size_t)len, &zLen);
					if ((bytes = write(it->fd, zBuffer, (size_t)zLen)) == -1) {
						if (may_talk)
							warn("[write] error");
					} else
						sent += bytes;
				} else {
					ssize_t bytes;

					if ((bytes = write(it->fd, buffer, (size_t)len)) == -1) {
						if (may_talk)
							warn("[write] error");
					} else
						sent += bytes;
				}

				close(it->fd);
				request_counter++;

				if (may_talk) {
					char hbuf[MAXHOSTNAMELEN], sbuf[NI_MAXSERV];
					int ni_error;

					ni_error = getnameinfo((struct sockaddr *)&it->c_addr,
							       it->c_addr_len, hbuf, sizeof(hbuf),
							       sbuf, sizeof(sbuf),
							       NI_NUMERICHOST | NI_NUMERICSERV);
					if (ni_error) {
						warnx("%s: getnameinfo: %s",
						      __func__, gai_strerror(ni_error));
						quit = 1;
						error = 1;
					} else {
						if (sent > 0)
							warnx("[write] sent %ld bytes to [%s]:%s",
							      (long int)sent, hbuf, sbuf);
					}
				}

				xmlFree(buffer);
				buffer = NULL;
				zFree(zBuffer);
			}
		}

		if (reread) {
			if (may_talk)
				warnx("[update] rereading all system information");
			sysinfo->readSysInfo();
			reread = 0;
		} else if (update) {
			sysinfo->updateSysInfo();
			update = 0;
		}

		if (quit && may_talk)
			warnx("[shutdown] %ld customer%s served", request_counter,
			      request_counter == 1 ? "" : "s");
	}

	delete sysinfo;
	delete server;
	delete[] __progname;

	return (error);
}
