/*	$Kolabore: netbsd.cpp,v 1.31 2006/10/28 00:33:57 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * netbsd.cpp:
 * Gathering information and stats on NetBSD 3.0 or newer.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/param.h>
#include <sys/disk.h>
#include <sys/socket.h>
#include <sys/statvfs.h>
#include <sys/swap.h>
#include <sys/sysctl.h>
#include <sys/wait.h> // XXX for the ugly hack

#include <machine/cpu.h>

#include <netinet/in.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_ether.h>
#include <net/if_media.h>
#include <net/if_types.h>
#include <net/pfvar.h>
#include <net/route.h>

#include <err.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <libgen.h> // XXX for the ugly hack
#include <limits.h> // XXX for the ugly hack
#include <netdb.h>
#include <signal.h> // XXX for the ugly hack
#include <stdint.h>
#include <unistd.h>
}

#include <cerrno>
#include <cstdio>
#include <ctime>

#include <string>
#include <vector>

#include <iostream>

#include "netbsd.h"

#define SMALLBUFSIZ	128

using namespace std;

struct netbsd_private {
	uint64_t        last_cp_time[CPUSTATES];
	time_t		boottime;
};

struct netbsd_interrupt {
	uint64_t	count;
	string		source;
};

static const struct ifmedia_description ifm_type_descriptions[] =
	IFM_TYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_descriptions[] =
	IFM_SUBTYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_mode_descriptions[] =
	IFM_MODE_DESCRIPTIONS;
static const struct ifmedia_description ifm_option_descriptions[] =
	IFM_OPTION_DESCRIPTIONS;
static const int ifm_status_valid_list[] = IFM_STATUS_VALID_LIST;
static const struct ifmedia_status_description ifm_status_descriptions[] =
	IFM_STATUS_DESCRIPTIONS;

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	if ((priv = new struct netbsd_private) == NULL)
		errx(1, "%s: Out of memory", __func__);
	pf_fd = -1;
}

XSI::~XSI(void)
{
	delete priv;
	if (pf_fd != -1)
		close(pf_fd);
}

/*
 * Devices to be opened before dropping privileges.
 * Silently returns 0 on success. Warns and returns >0 on error.
 *
 * NOTE: Valid XML must not rely on information gathered through privileged
 *       devices! Only optional, user-requested features may depend on
 *       things that may need super-user privileges.
 */
int
XSI::tryOpenPrivilegedDevices(void)
{
	if (configuration->showRaid()) {
		warnx("Sorry, RAID monitoring support is not (yet?) available on NetBSD.");
		return (1);
	}

	if (configuration->showPacketFilter()) {
		if ((pf_fd = open("/dev/pf", O_RDONLY, 0)) == -1) {
			warn("Unable to open /dev/pf to gather pf(4) stats");
			return (1);
		}
	}

	return (0);
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	netbsdInit();

	readSystemArch();
	readSystemModel();
	readSystemNumCPUs();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateCPUFreq();
	updateCPUStats();
	updateVMStats();
	updateFSStats();
	updateDiskStats();
	updateOSUptime();
	updateLoadAvg();
	updateInterrupts();
	updateSensors();
	updateNetStats();
	updatePFStats();
}

void
XSI::netbsdInit(void)
{
	int		mib[] = { CTL_KERN, KERN_BOOTTIME };
	struct timeval	tv;
	size_t		size = sizeof(tv);
	static bool	first = true;

	if (first)
		first = false;
	else
		return;

	if (sysctl(mib, 2, &tv, &size, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);
	priv->boottime = tv.tv_sec;
}

string
XSI::getSysctlString(int mib0, int mib1, int mib2, unsigned int len,
		     bool must_exist, const char *caller)
{
	int	mib[] = { mib0, mib1, mib2 };
	size_t	siz;
	string	ret;
	char   *buf;

	if (sysctl(mib, len, NULL, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP || must_exist)
			err(1, "%s: sysctl", caller);
		return ("");
	}

	buf = new char[siz];
	if (buf == NULL)
		errx(1, "%s: Out of memory", __func__);
	if (sysctl(mib, len, buf, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);
	ret = buf;
	delete[] buf;
	return (ret);
}

void
XSI::readSystemArch(void)
{
	string arch = getSysctlString(CTL_HW, HW_MACHINE, 0, 2, true, __func__);
	system.setArch(arch.c_str());
}

void
XSI::readSystemModel(void)
{
	string buf;

	buf = getSysctlString(CTL_HW, HW_MODEL, 0, 2, false, __func__);
	if (!buf.empty())
		system.setModel(buf.c_str());
}

void
XSI::readSystemNumCPUs(void)
{
	int mib[] = { CTL_HW, HW_NCPU };
	int ncpu;
	size_t siz = sizeof(ncpu);

	if (sysctl(mib, 2, &ncpu, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);
	system.setNumCPUs(ncpu);
}

void
XSI::readSystemOSInfo(void)
{
	char *type, *rel, *fullvers;
	string vers, kern;

	if(!sysctlStr(&type, CTL_KERN, KERN_OSTYPE, -1))
		err(1, "%s: sysctl", __func__);

	if(!sysctlStr(&rel, CTL_KERN, KERN_OSRELEASE, -1)) {
		delete[] type;
		err(1, "%s: sysctl", __func__);
	}

	if(sysctlStr(&fullvers, CTL_KERN, KERN_VERSION, -1)) {
		char *sharp, *colon, *ident;

		if(NULL != (sharp = strchr(fullvers, '#')) &&
		   NULL != (colon = strchr(fullvers, ':')) &&
		   NULL != (ident = strrchr(fullvers, '/'))) {
			char *ii, *date;
			char buf[BUFSIZ];
			char wkday[5], month[5];
			int day, hour, minute, second, year;
			struct tm tm;

			/* snip entire version string in two at the newline */
			for (ii = fullvers; *ii && *ii != '\n'; ii++) {}
			*ii = '\0';

			/* learn the kernel name from the end of the 2nd half */
			ident++;
			for (ii = ident; '\0' != *ii && !isspace(*ii); ii++) {}
			*ii = '\0';

			for (date = colon + 1; isspace(*date); date++) {}

			/* reformat and parse date(1) string; %+ doesn't do it */
			if (sscanf(date, "%4s %4s %d %d:%d:%d %*s %d",
				   (char *)&wkday, (char *)&month, &day,
				   &hour, &minute, &second, &year) != 7)
				errx(1, "%s: sscanf", __func__);
			if (snprintf(buf, sizeof(buf), "%s %s %2d %0d:%0d:%0d %d",
				     wkday, month, day, hour, minute, second, year) == -1)
				errx(1, "%s: snprintf", __func__);
			if (strptime(buf, "%a %b %e %H:%M:%S %Y", &tm) == NULL)
				errx(1, "%s: strptime", __func__);

			/* assemble version string as in /etc/motd */
			vers = string(fullvers, sharp - fullvers) + sharp;

			/* assemble normalized kernel version */
			*colon = '\0';
			if(strftime(buf, 64, "%Y%m%d%H%M", &tm) == 0)
				errx(1, "%s: strftime", __func__);
			kern = string(ident) + sharp + " " + buf;
		}

		delete[] fullvers;
	}

	system.setOSInfo(type, rel, NULL, kern.c_str(), vers.c_str());

	delete[] type;
	delete[] rel;
}

void
XSI::updateCPUFreq(void)
{
	int32_t freq;
	size_t siz = sizeof(freq);

	if (sysctlbyname("machdep.powernow.frequency.current", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}

	if (sysctlbyname("machdep.est.frequency.current", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}

	/* XXX - Is this in Hz or MHz? */
	if (sysctlbyname("machdep.tm_longrun_frequency", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}
}

void
XSI::updateCPUStats(void)
{
	uint64_t info[CPUSTATES];
	uint64_t total;
	uint64_t diff[CPUSTATES];
	double percent;
	int i;
	vector<struct cpu> cpus;
	int mib[] = { CTL_KERN, KERN_CP_TIME };
	size_t siz = sizeof(info);

	if (sysctl(mib, 2, info, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	cpus.resize(1);
	cpus[0].num = -1;

	total = 0;
	for (i = 0; i < CPUSTATES; i++) {
		diff[i] = info[i] - priv->last_cp_time[i];
		priv->last_cp_time[i] = info[i];
		total += diff[i];
	}

	if (total > 0)
		percent = 100 / (double)total;
	else
		percent = 0;

	cpus[0].system    = diff[CP_SYS] * percent;
	cpus[0].user      = diff[CP_USER] * percent;
	cpus[0].nice      = diff[CP_NICE] * percent;
	cpus[0].interrupt = diff[CP_INTR] * percent;
	cpus[0].idle      = diff[CP_IDLE] * percent;

	system.setCPUs(cpus);
}

#define dbtoqb(b)	dbtob((int64_t)(b))
void
XSI::updateVMStats(void)
{
	struct vm_info	vm_info;
	struct uvmexp	uvmexp;
	struct swapent *swap;
	size_t		size;
	int		mib[2], num_swapdevs, i;
	uint64_t	swap_used = 0;

	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	size = sizeof(uvmexp);
	if (sysctl(mib, 2, &uvmexp, &size, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	vm_info.ram_total = uvmexp.npages * uvmexp.pagesize;
	vm_info.ram_free = (uvmexp.free + uvmexp.inactive) * uvmexp.pagesize;

	vm_info.swap_total = vm_info.swap_free = 0;

	if ((num_swapdevs = swapctl(SWAP_NSWAP, NULL, 0)) < 1) {
		system.setVMInfo(vm_info);
		return;
	}

	swap = new struct swapent[num_swapdevs];
	if (swap == NULL)
		errx(1, "%s: Out of memory", __func__);
	if (swapctl(SWAP_STATS, swap, num_swapdevs) < num_swapdevs) {
		delete[] swap;
		err(1, "%s: swapctl", __func__);
	}

	for (i = 0; i < num_swapdevs; i++) {
		vm_info.swap_total += dbtoqb(swap[i].se_nblks);
		swap_used += dbtoqb(swap[i].se_inuse);
	}
	vm_info.swap_free = vm_info.swap_total - swap_used;

	system.setVMInfo(vm_info);

	delete[] swap;
}

/*
 * Reason #1 for not chrooting on NetBSD:
 * Outside filesystems are being hidden by the kernel from chroot()ed
 * processes.
 */
void
XSI::updateFSStats(void)
{
	vector<struct fs> filesystems;
	struct statvfs *fsp = NULL;
	int i, num;

	if ((num = getmntinfo(&fsp, ST_NOWAIT)) == 0) {
		if (fsp == NULL)
			err(1, "%s: getmntinfo", __func__);
		storage.setFilesystems(filesystems);
	}

	for (i = 0; i < num; i++) {
		struct fs fs;

		fs.fs_type = fsp[i].f_fstypename;
		if (fs.fs_type == "kernfs" || fs.fs_type == "procfs")
			continue;

		fs.mountpoint = fsp[i].f_mntonname;
		fs.fs_name = fsp[i].f_mntfromname;
		fs.size_total = fsp[i].f_blocks * fsp[i].f_bsize;
		fs.size_free = fsp[i].f_bfree * fsp[i].f_bsize;
		fs.size_avail = fsp[i].f_bavail * fsp[i].f_bsize;
		fs.nodes_total = fsp[i].f_files;
		fs.nodes_free = fsp[i].f_ffree;

		filesystems.push_back(fs);
	}

	storage.setFilesystems(filesystems);
}

void
XSI::updateDiskStats(void)
{
	int			mib[3], ndrives, i;
	size_t			size;
	struct disk_sysctl     *dstats;
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	bool			drop_old_stats = false, disks_changed;

	{
		string disknames = getSysctlString(CTL_HW, HW_DISKNAMES, 0, 2,
						   true, __func__);
		ndrives = (int)split(disknames.c_str()).size();
	}

	if (ndrives == 0) {
		old_disks.clear();
		storage.setDisks(old_disks);
		return;
	}

	if (ndrives != (int)old_disks.size())
		drop_old_stats = true;

	mib[0] = CTL_HW;
	mib[1] = HW_DISKSTATS;
	mib[2] = (int)sizeof(struct disk_sysctl);
	size = ndrives * sizeof(struct disk_sysctl);
	dstats = new struct disk_sysctl[ndrives];
	if (dstats == NULL)
		errx(1, "%s: Out of memory", __func__);
	if (sysctl(mib, 3, dstats, &size, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (i = 0; i < ndrives; i++) {
				XSI_DiskIO disk(dstats[i].dk_name);

				disk.setBytesWritten(dstats[i].dk_wbytes);
				disk.setBytesRead(dstats[i].dk_rbytes);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			for (i = 0; i < ndrives; i++) {
				if (strcmp(old_disks[i].getDiskName(), dstats[i].dk_name) != 0) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				old_disks[i].setBytesWritten(dstats[i].dk_wbytes);
				old_disks[i].setBytesRead(dstats[i].dk_rbytes);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);

	delete[] dstats;
}

void
XSI::updateOSUptime(void)
{
	time_t	now;

	time(&now);
	system.setOSUptime(now - priv->boottime);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		err(1, "%s: getloadavg", __func__);

	system.setLoadAvg(v);
}

// BEGIN UGLY HACK

#ifdef HAVE_VMSTAT_UTILITY

static pid_t spawnOfCthulhu(int);
static void chldHandler(int);
static int checkChild(pid_t, const char *);
volatile sig_atomic_t sigChld;
volatile sig_atomic_t error;

const char vmstat_utility[] = HAVE_VMSTAT_UTILITY;

/*
 * XXXXXXXXXX!
 *
 * This is a very nasty hack to get interrupt names and -counters.
 *
 * Parse the output from spawnOfCthulhu() (which is the output of
 * `/usr/bin/vmstat -i`.) Vmstat is sgid kmem and therefore able to use KVM and
 * event counters. I don't want XSI to ever need to run with even slightly
 * elevated privileges, so this avenue is closed to XSI itself.
 *
 * There may be better alternatives, such as somehow extracting that
 * information from /dev/ksyms(4), via nlist(3) and/or ioctl(2). There,
 * interesting symbol names would be "[e]intrnames" and "[e]intrcnt". So far,
 * I tried hard to do that without any success ... I don't understand that API
 * at all and need help with it from a NetBSD guru. A working patch for this
 * would go in immediately and there's so much glory and fame to get from this!
 * Please submit a patch. :-)
 *
 * This is currently worse than all of the Linux parsing-of-files crap. :<
 */
void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints = system.getInterrupts();
	bool			drop_old_stats = false, ints_changed;

	int	pipe_fds[2];
	int	pRead, pWrite;
	FILE   *pipef;
	pid_t	pid;
	int	c;
	string	buf;
	vector<netbsd_interrupt> intrbuf;
	vector<netbsd_interrupt>::iterator ii;

	if (pipe(pipe_fds) == -1)
		err(1, "%s: pipe", __func__);

	pRead = pipe_fds[0];
	pWrite = pipe_fds[1];

	pid = spawnOfCthulhu(pWrite);

	// parent process

	close(pWrite);
	if ((pipef = fdopen(pRead, "r")) == NULL)
		err(1, "%s: fdopen", __func__);

	signal(SIGCHLD, chldHandler);

	while ((c = getc(pipef)) != EOF)
		buf.push_back((char)c);

	fclose(pipef);
	signal(SIGCHLD, SIG_DFL);

	if (sigChld) {
		if (checkChild(pid, vmstat_utility) == 0) {
			kill(pid, SIGTERM);
			error = 1;
		}
		sigChld = 0;
	}

	if (error)
		errx(1, "%s exited with an error. See error message above, if any.",
		     vmstat_utility);

	{
		vector<string> bufvect = split(buf, '\n');
		vector<string>::iterator si;

		for (si = bufvect.begin(); si != bufvect.end(); si++) {
			vector<string> tmp = split(*si);
			string numstr = tmp[tmp.size() - 2];
			struct netbsd_interrupt intr;
			char *ep;
			unsigned int i;

			if (tmp.size() <= 3 || tmp[0] == "interrupt" ||
			    tmp[0] == "Total")
				continue;

			intr.count = strtoull(numstr.c_str(), &ep, 10);
			if ((intr.count == ULLONG_MAX && errno == ERANGE)
			    || numstr.c_str() == ep || *ep != '\0')
				errx(1, "%s: Error reading %s", __func__,
				     numstr.c_str());

			intr.source = "";
			for (i = 0; i < tmp.size() - 2; i++) {
				intr.source += tmp[i];
				if (i == 0)
					intr.source += ":";
				else if (i < tmp.size() - 3)
					intr.source += "_";
			}

			intrbuf.push_back(intr);
		}
	}

	if (intrbuf.size() != old_ints.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = intrbuf.begin(); ii != intrbuf.end(); ii++) {
				XSI_Interrupt intr(ii->source.c_str());

				intr.setIntsCount(ii->count);

				new_ints.push_back(intr);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < intrbuf.size(); i++) {
				if (intrbuf[i].source != old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount(intrbuf[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

/*
 * XXX ... Uargh. :-(
 */
static pid_t
spawnOfCthulhu(int pWrite)
{
	pid_t pid;

	switch (pid = fork()) {
	case -1:
		err(1, "%s: fork", __func__);
	case 0:
		setproctitle("vmstat(8) launcher");
		break;
	default:
		return (pid);
	}

	// child process

	close(STDOUT_FILENO);
	if (dup2(pWrite, STDOUT_FILENO) == -1)
		err(1, "%s: dup2", __func__);
	close(pWrite);

	if (execl(vmstat_utility, basename(const_cast<char *>(vmstat_utility)),
		  "-i", NULL) == -1) {
		if (errno == ENOENT)
			exit(0);
		else
			err(1, "%s: execl", __func__);
	}

	/* NOTREACHED */
	exit(0); // STFU, stupid gcc. Yes. This cannot be reached.
}

static void
chldHandler(int sig)
{
	if (sig == SIGCHLD)
		sigChld = 1;
}

static int
checkChild(pid_t pid, const char *name)
{
	int status, ret = 0;

	if (waitpid(pid, &status, WNOHANG) > 0) {
		if (WIFEXITED(status)) {
			if (WEXITSTATUS(status)) {
				warnx("%s exited with error code %d", name, WEXITSTATUS(status));
				error = 1;
			}
			ret = 1;
		}
		if (WIFSIGNALED(status)) {
			warnx("%s exited due to signal %d", name, WTERMSIG(status));
			error = 1;
			ret = 1;
		}
	}

	return (ret);
}

#else

void XSI::updateInterrupts(void) {}

#endif /* HAVE_VMSTAT_UTILITY */

// END UGLY HACK

void
XSI::updateSensors(void)
{
	// XXX: Missing. Find out how envstat(8) does it and how /dev/sysmon
	//      usage would work in XSI.
}

void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		err(1, "%s: getifaddrs", __func__);

	/* Convenience; make a vector of pointers to link-layer interfaces */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_LINK)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);
				struct if_data *ifd;

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(netIfStat, (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(netIfStat, (*it));
				getNetIfmediaInfo(netIfStat);
				getNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

				netIfStat.setPacketsIn(ifd->ifi_ipackets);
				netIfStat.setPacketsOut(ifd->ifi_opackets);
				netIfStat.setErrorsIn(ifd->ifi_ierrors);
				netIfStat.setErrorsOut(ifd->ifi_oerrors);
				netIfStat.setBytesIn(ifd->ifi_ibytes);
				netIfStat.setBytesOut(ifd->ifi_obytes);
				netIfStat.setPacketsDropped(ifd->ifi_iqdrops);
				netIfStat.setCollisions(ifd->ifi_collisions);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				struct if_data	       *ifd;

				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(old_nstats[i], (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(old_nstats[i], (*it));
				getNetIfmediaInfo(old_nstats[i]);
				getNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

				old_nstats[i].setPacketsIn(ifd->ifi_ipackets);
				old_nstats[i].setPacketsOut(ifd->ifi_opackets);
				old_nstats[i].setErrorsIn(ifd->ifi_ierrors);
				old_nstats[i].setErrorsOut(ifd->ifi_oerrors);
				old_nstats[i].setBytesIn(ifd->ifi_ibytes);
				old_nstats[i].setBytesOut(ifd->ifi_obytes);
				old_nstats[i].setPacketsDropped(ifd->ifi_iqdrops);
				old_nstats[i].setCollisions(ifd->ifi_collisions);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

/*
 * updatePFStats() reads counters and other information from pf(4),
 * inspired by src/sbin/pfctl/pfctl*.c and Daniel Hartmeier's pfstat.
 */
void
XSI::updatePFStats(void)
{
	struct pf_status	status;
	struct pfioc_iface	iface;
	struct pfi_kif		ifs[256];	// Eats >50 kB of stack memory
	list<int>		ifs_with_traffic;
	list<int>::iterator	iwt_it;
	int			i;
	vector<FilteredIfStat>	old_fIfStats = packetFilter.getFilteredIfStats();
	bool			drop_old_stats = false, filteredIfs_changed;

	if (pf_fd == -1)
		return;

	bzero(&status, sizeof(status));
	if (ioctl(pf_fd, DIOCGETSTATUS, &status) == -1)
		err(1, "%s: Unable to get status from pf(4)", __func__);

	if (!status.running) {
		packetFilter.setEnabled(false);
		return;
	} else
		packetFilter.setEnabled(true);
	packetFilter.setRuntime(time(NULL) - status.since);
	packetFilter.setRuleMatches(status.counters[PFRES_MATCH]);
	packetFilter.setStates(status.states);
	packetFilter.setStateSearches(status.fcounters[FCNT_STATE_SEARCH]);
	packetFilter.setStateInserts(status.fcounters[FCNT_STATE_INSERT]);
	packetFilter.setStateRemovals(status.fcounters[FCNT_STATE_REMOVALS]);
#ifdef PFRES_BADSTATE
	packetFilter.setStateMismatches(status.counters[PFRES_BADSTATE]);
#endif

	bzero(&iface, sizeof(iface));
	iface.pfiio_buffer = ifs;
	iface.pfiio_esize = sizeof(ifs[0]);
	iface.pfiio_size = sizeof(ifs) / sizeof(ifs[0]);
	if (ioctl(pf_fd, DIOCIGETIFACES, &iface) == -1) {
		/*
		 * The ioctl is there, but it fails. Probably due to a too-old
		 * pf(4) implementation? Who knows. Let's accept this as a soft
		 * error and continue regardless. Once it works, it'll work in
		 * XSI as well.
		 */
		if (errno != ENODEV)
			err(1, "%s: Unable to get interface stats from pf(4)", __func__);
		return;
	}

	/* Find interfaces that have seen traffic */
	for (i = 0; i < iface.pfiio_size; i++) {
		int j;

		for (j = 0; j < 8; j++) {
			if (ifs[i].pfik_packets[j&1?0:1][j&2?0:1][j&4?0:1] != 0)
				ifs_with_traffic.push_back(i);
		}
	}
	ifs_with_traffic.sort();
	ifs_with_traffic.unique();

	if (ifs_with_traffic.size() != old_fIfStats.size())
		drop_old_stats = true;

	/* Get counters from ifs_with_traffic: bytes/packets[AF][direction][op] */
	do {
		filteredIfs_changed = false;
		if (drop_old_stats) {
			vector<FilteredIfStat>	fIfStats;

			for (iwt_it = ifs_with_traffic.begin();
			     iwt_it != ifs_with_traffic.end();
			     iwt_it++) {
				FilteredIfStat iwt(ifs[*iwt_it].pfik_name);

				iwt.setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				iwt.setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				iwt.setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				iwt.setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				iwt.setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				iwt.setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				iwt.setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				iwt.setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				iwt.setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				iwt.setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				iwt.setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				iwt.setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				iwt.setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				iwt.setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				iwt.setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				iwt.setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);

				fIfStats.push_back(iwt);
			}
			packetFilter.setFilteredIfStats(fIfStats);
		} else {
			vector<FilteredIfStat>::iterator it;

			for (it = old_fIfStats.begin(), iwt_it = ifs_with_traffic.begin();
			     it != old_fIfStats.end(); it++, iwt_it++) {
				if (strcmp(ifs[*iwt_it].pfik_name, it->getIfName()) != 0) {
					filteredIfs_changed = true;
					drop_old_stats = true;
					break;
				}
				it->setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				it->setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				it->setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				it->setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				it->setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				it->setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				it->setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				it->setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				it->setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				it->setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				it->setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				it->setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				it->setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				it->setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				it->setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				it->setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);
			}
			if (!filteredIfs_changed)
				packetFilter.setFilteredIfStats(old_fIfStats);
		}
	} while (filteredIfs_changed);
}

void
XSI::getNetLLAddr(XSI_NetIfStat &netIfStat, struct sockaddr_dl *sdl)
{
	if (sdl->sdl_alen) {
		switch (sdl->sdl_type) {
		case IFT_ETHER:
#if defined(IFT_CARP)
		case IFT_CARP:
#endif /* IFT_CARP */
			netIfStat.setLladdrString(ether_ntoa((struct ether_addr *)LLADDR(sdl)));
			return;
		default:
			break;
		}
	}
}

void
XSI::getNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr)
{
	struct ifreq	ifr, *ifrp;
	int		sock;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		err(1, "%s: socket", __func__);

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), ifaddr->ifa_addr->sa_len));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));
	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);
	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);

	close(sock);
}

void
XSI::getNetIfmediaInfo(XSI_NetIfStat &netIfStat)
{
	struct ifmediareq	ifmr;
	int			sock, *media_list;
	const char             *ifname = netIfStat.getIfName();

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		err(1, "%s: socket", __func__);

	bzero(&ifmr, sizeof(ifmr));
	strlcpy(ifmr.ifm_name, ifname, sizeof(ifmr.ifm_name));
	if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
		// Interface doesn't support media directly
		close(sock);
		return;
	} else {
		string mediaStr;

		if (ifmr.ifm_count == 0) {
			warnx("%s: %s: no media types?", __func__, ifname);
			close(sock);
			return;
		}

		media_list = new int[ifmr.ifm_count];
		ifmr.ifm_ulist = media_list;
		if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
			close(sock);
			err(1, "%s: SIOCGIFMEDIA", __func__);
		}

		mediaStr = getMediaWordString(ifmr.ifm_current, 1);
		if (ifmr.ifm_active != ifmr.ifm_current) {
			mediaStr += " (";
			mediaStr += getMediaWordString(ifmr.ifm_active, 0);
			mediaStr += ")";
		}

		netIfStat.setMediaString(mediaStr.c_str());

		if (ifmr.ifm_status & IFM_AVALID) {
			const struct ifmedia_status_description *ifms;
			int bitno, found = 0;

			for (bitno = 0; ifm_status_valid_list[bitno] != 0; bitno++) {
				for (ifms = ifm_status_descriptions;
				     ifms->ifms_valid != 0; ifms++) {
					if (ifms->ifms_type !=
					    IFM_TYPE(ifmr.ifm_current) ||
					    ifms->ifms_valid !=
					    ifm_status_valid_list[bitno])
						continue;
					netIfStat.setStatusString(IFM_STATUS_DESC(ifms, ifmr.ifm_status));
					if (ifms->ifms_bit & ifmr.ifm_status)
						netIfStat.setLinkStatus(IF_LINK_ACTIVE);
					else
						netIfStat.setLinkStatus(IF_LINK_NOCARRIER);
					found = 1;
					break;
				}
			}
			if (!found) {
				netIfStat.setStatusString("unknown");
				netIfStat.setLinkStatus(IF_LINK_UNKNOWN);
			}
		}

		delete[] media_list;
	}

	close(sock);
}

void
XSI::getNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_LINK)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					(size_t)ifa->ifa_addr->sa_len,
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    ifa->ifa_dstaddr->sa_len);
			break;
		default:
			if (!have_warned) {
				warnx("Unhandled address family %d found, please report this!",
				      ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
void
XSI::getDefaultRoutes(void)
{
	int			mib[6];
	size_t			siz;
	char		       *buf, *next, *ep = NULL;
	struct rt_msghdr       *rtm;
	struct sockaddr        *sa, *rti_info[RTAX_MAX];
	char			host[MAXHOSTNAMELEN];
	int			error;
	vector<struct default_route> defRoutes;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;	// protocol number
	mib[3] = 0;	// AF
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_GATEWAY;
	if (sysctl(mib, 6, NULL, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	if (siz == 0)
		return;

	buf = new char[siz];
	if (buf == NULL)
		errx(1, "%s: Out of memory", __func__);
	if (sysctl(mib, 6, buf, &siz, NULL, 0) == -1)
		err(1, "%s: sysctl", __func__);

	ep = buf + siz;
	for (next = buf; next < ep; next += rtm->rtm_msglen) {
		unsigned int		i;
		struct default_route	defRt;

		rtm = (struct rt_msghdr *)next;
		sa = (struct sockaddr *)(rtm + 1);
		for (i = 0; i < RTAX_MAX; i++) {
			if (rtm->rtm_addrs & (1 << i)) {
				rti_info[i] = sa;
				sa = (struct sockaddr *)((char *)(sa) +
							 ROUNDUP(sa->sa_len));
			} else
				rti_info[i] = NULL;
		}
		if (rti_info[RTAX_DST] == NULL || rti_info[RTAX_GATEWAY] == NULL)
			continue;
		if (rti_info[RTAX_DST]->sa_family == AF_INET) {
			if (((struct sockaddr_in *)rti_info[RTAX_DST])->sin_addr.s_addr == INADDR_ANY)
				defRt.family = XSI_RTAF_IPV4;
			else
				continue;
		} else if (rti_info[RTAX_DST]->sa_family == AF_INET6) {
			struct sockaddr_in6 sin6;
			bool zero_mask = true;;

			sin6 = *((struct sockaddr_in6 *)rti_info[RTAX_DST]);
			if (rti_info[RTAX_NETMASK]) {
				struct sockaddr_in6 mask;
				unsigned int lim, final = 0, illegal = 0;
				unsigned char *p;

				mask = *((struct sockaddr_in6 *)rti_info[RTAX_NETMASK]);
				lim = mask.sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
				lim = lim < sizeof(struct in6_addr) ? lim : sizeof(struct in6_addr);
				for (p = (unsigned char *)&mask.sin6_addr, i = 0; i < lim; p++) {
					int masks[9] = {0x00, 0x80, 0xc0, 0xe0,
							0xf0, 0xf8, 0xfc, 0xfe,
							0xff};
					int j;

					if (final && *p) {
						illegal++;
						sin6.sin6_addr.s6_addr[i++] = 0x00;
						continue;
					}

					for (j = 0; j < 9; j++) {
						if (*p & 0xff == masks[j])
							break;
					}

					if (j >= 1 && j <= 8) {
						zero_mask = false;
						break;
					}
					if (j == 9)
						illegal++;
					if (j != 8)
						final++;
					if (!illegal)
						sin6.sin6_addr.s6_addr[i++] &= *p;
					else
						sin6.sin6_addr.s6_addr[i++] = 0x00;
				}
				while (i < sizeof(struct in6_addr))
					sin6.sin6_addr.s6_addr[i++] = 0x00;
			} else
				zero_mask = false;

			if (zero_mask && IN6_IS_ADDR_UNSPECIFIED(&sin6.sin6_addr))
				defRt.family = XSI_RTAF_IPV6;
			else
				continue;
		} else
			continue;
		if ((error = getnameinfo(rti_info[RTAX_GATEWAY], rti_info[RTAX_GATEWAY]->sa_len,
					 host, sizeof(host), NULL, 0, NI_NUMERICHOST)) != 0)
			errx(1, "%s: getnameinfo: %s", __func__, gai_strerror(error));
		defRt.gateway = host;
		defRoutes.push_back(defRt);
	}

	network.setDefaultRoutes(defRoutes);

	delete[] buf;
}

/* convenience functions for retrieving information via sysctl */
u_int
XSI::getmib(int *name, va_list ap)
{
	u_int len;

	for(len = 0; CTL_MAXNAME > len; len++) {
		name[len] = va_arg(ap, int);
		if(0 > name[len])
			break;
	}

	return len;
}

int
XSI::sysctlStr(char **val, ...)
{
	int name[CTL_MAXNAME];
	u_int namelen;
	va_list ap;
	size_t len;

	va_start(ap, val);
	namelen = getmib(name, ap);
	va_end(ap);

	if(0 > sysctl(name, namelen, NULL, &len, NULL, 0))
		return 0;
	*val = new char[len];

	if(0 > sysctl(name, namelen, *val, &len, NULL, 0)) {
		delete[] *val;
		return 0;
	}

	return 1;
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:print_media_word()
 */
string
XSI::getMediaWordString(int ifmw, int print_type)
{
	const struct ifmedia_description *desc;
	int	seen_option = 0;
	char	tbuf[SMALLBUFSIZ];
	string	ret = "";

	tbuf[0] = '\0';

	if (print_type) {
		bool found = false;

		for (desc = ifm_type_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE(ifmw) == desc->ifmt_word) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s ", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			return ("unknown");
	}
	ret += tbuf;

	{
		bool found = false;

		for (desc = ifm_subtype_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			strlcpy(tbuf, "unknown subtype", sizeof(tbuf));
	}
	ret += tbuf;

	/* Find mode. */
	if (IFM_MODE(ifmw) != 0) {
		bool found = false;

		for (desc = ifm_mode_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_MODE(ifmw) == IFM_MODE(desc->ifmt_word)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), " mode %s", desc->ifmt_string);
				break;
			}
		}
		if (found)
			ret += tbuf;
	}

	/* Find options. */
	for (desc = ifm_option_descriptions; desc->ifmt_string != NULL;
	     desc++) {
		if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
		    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
		    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
			if (seen_option == 0)
				ret += " ";
			snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
			ret += tbuf;
			seen_option |= IFM_OPTIONS(desc->ifmt_word);
		}
	}
	if (IFM_INST(ifmw) != 0) {
		snprintf(tbuf, sizeof(tbuf), " instance %d", IFM_INST(ifmw));
		ret += tbuf;
	}

	return (ret);
}
