/*	$Kolabore: netifstat.cpp,v 1.24 2006/10/29 18:42:16 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/socket.h>

#include <netinet/in.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
#include <netdb.h>
}

#include <cstring>

#include <string>
#include <vector>

#include "netifstat.h"
#include "timekeeper.h"
#include "xmlsysinfo.h"
#include "compat/xsicompat.h"

#define SMALLBUFSIZ	64

using namespace std;

XSI_IfAddress::XSI_IfAddress(void)
{
	family = -1;
	prefixlen = -1;
}

void
XSI_IfAddress::setAddress(struct sockaddr *addr, size_t addrsiz,
			  struct sockaddr *nmask, int af)
{
	char buf[MAXHOSTNAMELEN];
	struct sockaddr_in *sin;
	struct sockaddr_in6 *sin6;

	if (af == AF_INET || af == AF_INET6) {
		int error;

		error = getnameinfo(addr, addrsiz, buf, sizeof(buf), NULL, 0,
				    NI_NUMERICHOST);
		if (error)
			errx(1, "%s: getnameinfo, AF %d: %s", __func__,
			     af, gai_strerror(error));
		address = buf;
	} else
		errx(1, "%s: Address family not supported, use setAddressString()",
		     __func__);

	switch (af) {
	case AF_INET:
		sin = (sockaddr_in *)nmask;
		prefixlen = getPrefixLen((void *)&sin->sin_addr,
					 (int)sizeof(struct in_addr));
		break;
	case AF_INET6:
		sin6 = (struct sockaddr_in6 *)nmask;
		prefixlen = getPrefixLen((void *)&sin6->sin6_addr,
					 (int)sizeof(struct in6_addr));
		break;
	default:
		/* Only Mr. Lobo can come here. */
		errx(1, "%s: No Sir, I don't like it.", __func__);
	}
	family = af;
}

void
XSI_IfAddress::setAddressString(const char *addr, int af)
{
	address = addr;
	family = af;
}

bool
XSI_IfAddress::hasPrefixLen(void)
{
	return (prefixlen >= 0);
}

const char *
XSI_IfAddress::getAddressString(void)
{
	return (address.c_str());
}

const char *
XSI_IfAddress::getAddressFamilyString(void)
{
	switch (family) {
	case AF_UNSPEC:
		return ("unspecified");
#if defined(AF_LOCAL)
	case AF_LOCAL:
		return ("LOCAL");
#elif defined(AF_UNIX)
	case AF_UNIX:
		return ("LOCAL");
#endif
	case AF_INET:
		return ("INET");
	case AF_INET6:
		return ("INET6");
#if defined(AF_IPX)
	case AF_IPX:
		return ("IPX");
#endif /* AF_IPX */
	default:
		break;
	}

	if (family < AF_MAX)
		return ("other");
	else
		return ("unknown");
}

const char *
XSI_IfAddress::getPrefixLenString(void)
{
	if (prefixlen == -1)
		return (NULL);
	else
		return (makeIntStr(prefixlen));
}

/*
 * From OpenBSD src/sbin/ifconfig.c:prefix(), which seems faster than a simple
 * while() loop.
 */
int
XSI_IfAddress::getPrefixLen(void *val, int size)
{
	unsigned char *name = (unsigned char *)val;
	int byte, bit, plen = 0;

	for (byte = 0; byte < size; byte++, plen += 8)
		if (name[byte] != 0xff)
			break;
	if (byte == size)
		return (plen);
	for (bit = 7; bit != 0; bit--, plen++)
		if (!(name[byte] & (1 << bit)))
			break;
	for (; bit != 0; bit--)
		if (name[byte] & (1 << bit))
			return (0);
	byte++;
	for (; byte < size; byte++)
		if (name[byte])
			return (0);
	return (plen);
}


XSI_NetIfStat::XSI_NetIfStat(const char *ifName, bool up)
{
	name = ifName;
	is_up = up;

	is_running = is_promiscuous = is_loopback = is_pointtopoint = false;
	metric = mtu = 0;
	link_status = IF_LINK_UNKNOWN;
	packets_in = last_packets_in = 0;
	packets_out = last_packets_out = 0;
	packets_dropped = last_packets_dropped = 0;
	errors_in = 0;
	errors_out = 0;
	bytes_in = last_bytes_in = 0;
	bytes_out = last_bytes_out = 0;
	collisions = 0;
	is_carpdev = false;
	carp_preempt = -1;
	carp_arpbalance = -1;
	carp_status = CARP_UNSPEC;
}

void
XSI_NetIfStat::setIsUp(bool up)
{
	is_up = up;
}

void
XSI_NetIfStat::setIsRunning(bool running)
{
	is_running = running;
}

void
XSI_NetIfStat::setIsPromiscuous(bool promisc)
{
	is_promiscuous = promisc;
}

void
XSI_NetIfStat::setIsLoopback(bool loopback)
{
	is_loopback = loopback;
}

void
XSI_NetIfStat::setDescription(const char *desc)
{
	description = desc;
}

void
XSI_NetIfStat::setMediaString(const char *md)
{
	media = md;
}

void
XSI_NetIfStat::setLladdrString(const char *ll)
{
	lladdr = ll;
}

void
XSI_NetIfStat::setStatusString(const char *s)
{
	status = s;
}

void
XSI_NetIfStat::setMetric(int n)
{
	metric = n;
}

void
XSI_NetIfStat::setMTU(int n)
{
	mtu = n;
}

void
XSI_NetIfStat::setLinkStatus(enum if_link_status stat)
{
	link_status = stat;
}

void
XSI_NetIfStat::setPacketsIn(uint64_t n)
{
	last_packets_in = packets_in;
	packets_in = n;
}

void
XSI_NetIfStat::setPacketsOut(uint64_t n)
{
	last_packets_out = packets_out;
	packets_out = n;
}

void
XSI_NetIfStat::setPacketsDropped(uint64_t n)
{
	last_packets_dropped = packets_dropped;
	packets_dropped = n;
}

void
XSI_NetIfStat::setErrorsIn(uint64_t n)
{
	errors_in = n;
}

void
XSI_NetIfStat::setErrorsOut(uint64_t n)
{
	errors_out = n;
}

void
XSI_NetIfStat::setBytesIn(uint64_t n)
{
	last_bytes_in = bytes_in;
	bytes_in = n;
}

void
XSI_NetIfStat::setBytesOut(uint64_t n)
{
	last_bytes_out = bytes_out;
	bytes_out = n;
}

void
XSI_NetIfStat::setCollisions(uint64_t n)
{
	collisions = n;
}

void
XSI_NetIfStat::setIfAddresses(vector<XSI_IfAddress> addrs)
{
	addresses = addrs;
}

void
XSI_NetIfStat::setPTPDestAddress(struct sockaddr *addr, size_t siz)
{
	char buf[MAXHOSTNAMELEN];
	int error = getnameinfo(addr, siz, buf, sizeof(buf), NULL, 0,
			    NI_NUMERICHOST);
	if (error)
		errx(1, "%s: getnameinfo: %s", __func__, gai_strerror(error));
	ptp_dest_addr = buf;
	is_pointtopoint = true;
}

void
XSI_NetIfStat::setCARPStatus(const char *carpdev, enum xsi_carp_state status,
			     int vhid, int advskew, int advbase, bool preempt,
			     bool arpbalance)
{
	is_carpdev = true;
	if (carpdev != NULL)
		carp_carpdev = carpdev;
	else
		carp_carpdev.clear();
	carp_status = status;
	carp_vhid = vhid;
	carp_advskew = advskew;
	carp_advbase = advbase;
	carp_preempt = preempt;
	carp_arpbalance = arpbalance;
}

bool
XSI_NetIfStat::isUp(void)
{
	return (is_up);
}

bool
XSI_NetIfStat::isRunning(void)
{
	return (is_running);
}

bool
XSI_NetIfStat::isPromiscuous(void)
{
	return (is_promiscuous);
}

bool
XSI_NetIfStat::isLoopback(void)
{
	return (is_loopback);
}

bool
XSI_NetIfStat::isPointToPoint(void)
{
	return (is_pointtopoint);
}

bool
XSI_NetIfStat::isCARPDev(void)
{
	return (is_carpdev);
}

bool
XSI_NetIfStat::hasMetric(void)
{
	return (metric > 0);
}

bool
XSI_NetIfStat::hasMTU(void)
{
	return (mtu > 0);
}

bool
XSI_NetIfStat::hasAddresses(void)
{
	return (!addresses.empty() || !lladdr.empty());
}

bool
XSI_NetIfStat::hasDroppedPkgs(void)
{
	return (packets_dropped > 0);
}

bool
XSI_NetIfStat::hasErrors(void)
{
	return (errors_in > 0 || errors_out > 0);
}

bool
XSI_NetIfStat::hasCollisions(void)
{
	return (collisions > 0);
}

bool
XSI_NetIfStat::hasTraffic(void)
{
	return (packets_in > 0 ||
		packets_out > 0 ||
		packets_dropped > 0 ||
		errors_in > 0 ||
		errors_out > 0 ||
		bytes_in > 0 ||
		bytes_out > 0 ||
		collisions > 0);
}

const char *
XSI_NetIfStat::getIfName(void)
{
	return (name.c_str());
}

const char *
XSI_NetIfStat::getDescriptionString(void)
{
	if (description.empty())
		return (NULL);
	else
		return (description.c_str());
}

const char *
XSI_NetIfStat::getMediaString(void)
{
	if (media.empty())
		return (NULL);
	else
		return (media.c_str());
}

const char *
XSI_NetIfStat::getLladdrString(void)
{
	if (lladdr.empty())
		return (NULL);
	else
		return (lladdr.c_str());
}

const char *
XSI_NetIfStat::getStatusString(void)
{
	if (status.empty())
		return (NULL);
	else
		return (status.c_str());
}

const char *
XSI_NetIfStat::getMetricString(void)
{
	return (makeIntStr(metric));
}

const char *
XSI_NetIfStat::getMTUString(void)
{
	return (makeIntStr(mtu));
}

const char *
XSI_NetIfStat::getLinkStatusString(void)
{
	switch (link_status) {
	case IF_LINK_ACTIVE:
		return ("ACTIVE");
	case IF_LINK_NOCARRIER:
		return ("INACTIVE");
	case IF_LINK_UNKNOWN:
	default:
		return ("unknown");
	}
}

const char *
XSI_NetIfStat::getPacketsInString(void)
{
	return (makeIntStr(packets_in));
}

const char *
XSI_NetIfStat::getPacketsOutString(void)
{
	return (makeIntStr(packets_out));
}

const char *
XSI_NetIfStat::getPacketsDroppedString(void)
{
	return (makeIntStr(packets_dropped));
}

const char *
XSI_NetIfStat::getErrorsInString(void)
{
	return (makeIntStr(errors_in));
}

const char *
XSI_NetIfStat::getErrorsOutString(void)
{
	return (makeIntStr(errors_out));
}

const char *
XSI_NetIfStat::getBytesInString(void)
{
	return (makeIntStr(bytes_in));
}

const char *
XSI_NetIfStat::getBytesOutString(void)
{
	return (makeIntStr(bytes_out));
}

const char *
XSI_NetIfStat::getCollisionsString(void)
{
	return (makeIntStr(collisions));
}

const char *
XSI_NetIfStat::getPacketsInPerSecString(void)
{
	if (!packets_in || XSI_TimeKeeper::get()->timeDelta() < 0.01
	    || !last_packets_in || packets_in < last_packets_in
	    || packets_in - last_packets_in == 0)
		return (NULL);
	else
		return (makeFloatStr(packets_in - last_packets_in,
				     XSI_TimeKeeper::get()->timeDelta()));
}

const char *
XSI_NetIfStat::getPacketsOutPerSecString(void)
{
	if (!packets_out || XSI_TimeKeeper::get()->timeDelta() < 0.01
	    || !last_packets_out || packets_out < last_packets_out
	    || packets_out - last_packets_out == 0)
		return (NULL);
	else
		return (makeFloatStr(packets_out - last_packets_out,
				     XSI_TimeKeeper::get()->timeDelta()));
}

const char *
XSI_NetIfStat::getPacketsDroppedPerSecString(void)
{
	if (!packets_dropped || XSI_TimeKeeper::get()->timeDelta() < 0.01
	    || !last_packets_dropped || packets_dropped < last_packets_dropped
	    || packets_dropped - last_packets_dropped == 0)
		return (NULL);
	else
		return (makeFloatStr(packets_dropped - last_packets_dropped,
				     XSI_TimeKeeper::get()->timeDelta()));
}

const char *
XSI_NetIfStat::getBytesInPerSecString(void)
{
	if (!bytes_in || XSI_TimeKeeper::get()->timeDelta() < 0.01
	    || !last_bytes_in || bytes_in < last_bytes_in
	    || bytes_in - last_bytes_in == 0)
		return (NULL);
	else
		return (makeFloatStr(bytes_in - last_bytes_in,
				     XSI_TimeKeeper::get()->timeDelta()));
}

const char *
XSI_NetIfStat::getBytesOutPerSecString(void)
{
	if (!bytes_out || XSI_TimeKeeper::get()->timeDelta() < 0.01
	    || !last_bytes_out || bytes_out < last_bytes_out
	    || bytes_out - last_bytes_out == 0)
		return (NULL);
	else
		return (makeFloatStr(bytes_out - last_bytes_out,
				     XSI_TimeKeeper::get()->timeDelta()));
}

const char *
XSI_NetIfStat::getPTPDestAddressString(void)
{
	return (ptp_dest_addr.c_str());
}

vector<XSI_IfAddress>
XSI_NetIfStat::getIfAddresses(void)
{
	return (addresses);
}

const char *
XSI_NetIfStat::getCARPDeviceString(void)
{
	if (carp_carpdev.empty())
		return (NULL);
	else
		return (carp_carpdev.c_str());
}

const char *
XSI_NetIfStat::getCARPStatusString(void)
{
	switch (carp_status) {
	case CARP_INIT:
		return ("INIT");
	case CARP_BACKUP:
		return ("BACKUP");
	case CARP_MASTER:
		return ("MASTER");
	case CARP_UNKNOWN:
		return ("unknown");
	default:
		/* CARP_UNSPEC */
		err(1, "%s: Internal error: No CARP state set!", __func__);
	}
}

const char *
XSI_NetIfStat::getCARPVHIDString(void)
{
	return (makeIntStr((int32_t)carp_vhid));
}

const char *
XSI_NetIfStat::getCARPAdvskewString(void)
{
	return (makeIntStr((int32_t)carp_advskew));
}

const char *
XSI_NetIfStat::getCARPAdvbaseString(void)
{
	return (makeIntStr((int32_t)carp_advbase));
}

const char *
XSI_NetIfStat::getCARPPreemptString(void)
{
	if (carp_preempt < 0)
		return (NULL);
	return (carp_preempt == 0 ? "false" : "true");
}

const char *
XSI_NetIfStat::getCARPArpbalanceString(void)
{
	if (carp_arpbalance < 0)
		return (NULL);
	return (carp_arpbalance == 0 ? "false" : "true");
}
