/*	$Kolabore: raid.cpp,v 1.13 2006/10/03 09:34:31 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>

#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
}

#include <string>
#include <vector>

#include "raid.h"

#define SMALLBUFSIZ	64

using namespace std;

RDisk::RDisk(void)
{
	bus = target = lun = -1;
	status = RAID_DSK_INVALID;
	size = 0;
}

void
RDisk::setDiskID(int i)
{
	id = i;
}

void
RDisk::setDiskLocation(int b, int t)
{
	bus = b;
	target = t;
}

void
RDisk::setDiskLUN(int l)
{
	lun = l;
}

void
RDisk::setDiskStatus(enum RaidDiskStatus s)
{
	status = s;
}

void
RDisk::setDiskSize(uint64_t siz)
{
	size = siz;
}

void
RDisk::setDiskEnclosureDevName(const char *encldev)
{
	enclosure = encldev;
}

void
RDisk::setDiskVendorName(const char *v)
{
	vendor = v;
}

void
RDisk::setDiskSerial(const char *sn)
{
	serial = sn;
}

const char *
RDisk::getDiskIDString(void)
{
	return (makeIntStr((int32_t)id));
}

const char *
RDisk::getDiskBusString(void)
{
	return (makeIntStr((int32_t)bus));
}

const char *
RDisk::getDiskTargetString(void)
{
	return (makeIntStr((int32_t)target));
}

const char *
RDisk::getDiskLUNString(void)
{
	if (lun == -1)
		return (NULL);
	else
		return (makeIntStr((int32_t)lun));
}

const char *
RDisk::getDiskStatusString(void)
{
	switch (status) {
	case RAID_DSK_ONLINE:
		return ("online");
	case RAID_DSK_OFFLINE:
		return ("offline");
	case RAID_DSK_FAILED:
		return ("failed");
	case RAID_DSK_REBUILD:
		return ("rebuild");
	case RAID_DSK_HOTSPARE:
		return ("hot_spare");
	case RAID_DSK_UNUSED:
		return ("unused");
	case RAID_DSK_SCRUBBING:
		return ("scrubbing");
	case RAID_DSK_INVALID:
		return ("invalid");
	case RAID_DSK_UNKNOWN:
	default:
		return ("unknown");
	}
}

const char *
RDisk::getDiskSizeString(void)
{
	return (makeIntStr(size));
}

const char *
RDisk::getDiskEnclosureDevName(void)
{
	if (enclosure.empty())
		return (NULL);
	else
		return (enclosure.c_str());
}

const char *
RDisk::getDiskVendorName(void)
{
	if (vendor.empty())
		return (NULL);
	else
		return (vendor.c_str());
}

const char *
RDisk::getDiskSerial(void)
{
	if (serial.empty())
		return (NULL);
	else
		return (serial.c_str());
}


RVolume::RVolume(void)
{
	id = -1;
	level = -1;
	status = RAID_VOL_INVALID;
	progress = -1;
	elapsed = -1;
	size = 0;
}

void
RVolume::setVolumeDevName(const char *dev)
{
	device = dev;
}

void
RVolume::setVolumeID(int i)
{
	id = i;
}

void
RVolume::setRaidLevel(int lvl)
{
	level = lvl;
}

void
RVolume::setVolumeStatus(enum RaidVolumeStatus s)
{
	status = s;
}

void
RVolume::setVolumeOpProgress(int p)
{
	progress = p;
}

void
RVolume::setVolumeOpElapsed(int seconds)
{
	elapsed = seconds;
}

void
RVolume::setVolumeSize(uint64_t siz)
{
	size = siz;
}

void
RVolume::setVolumeDisks(vector<RDisk> dsks)
{
	disks = dsks;
}

const char *
RVolume::getVolumeDevName(void)
{
	return (device.c_str());
}

const char *
RVolume::getVolumeIDString(void)
{
	return (makeIntStr((int32_t)id));
}

const char *
RVolume::getRaidLevelString(void)
{
	return (makeIntStr((int32_t)level));
}

const char *
RVolume::getVolumeStatusString(void)
{
	switch (status) {
	case RAID_VOL_ONLINE:
		return ("online");
	case RAID_VOL_OFFLINE:
		return ("offline");
	case RAID_VOL_DEGRADED:
		return ("degraded");
	case RAID_VOL_BUILDING:
		return ("building");
	case RAID_VOL_SCRUBBING:
		return ("scrubbing");
	case RAID_VOL_REBUILD:
		return ("rebuild");
	case RAID_VOL_INVALID:
		return ("invalid");
	case RAID_VOL_UNKNOWN:
	default:
		return ("unknown");
	}
}

const char *
RVolume::getVolumeOpProgressString(void)
{
	if (progress < 0)
		return (NULL);
	else
		return (makeIntStr((int32_t)progress));
}

const char *
RVolume::getVolumeOpElapsedString(void)
{
	if (elapsed <= 0)
		return (NULL);
	else
		return (makeIntStr((int32_t)elapsed));
}

const char *
RVolume::getVolumeSizeString(void)
{
	return (makeIntStr(size));
}

vector<RDisk>
RVolume::getVolumeDisks(void)
{
	return (disks);
}

Raid::Raid(const char *ctrl)
{
	controller = ctrl;
}

void
Raid::setVolumes(vector<RVolume> vols)
{
	volumes = vols;
}

void
Raid::setReserveDisks(vector<RDisk> dsks)
{
	disks = dsks;
}

vector<RVolume>
Raid::getVolumes(void)
{
	return (volumes);
}

vector<RDisk>
Raid::getReserveDisks(void)
{
	return (disks);
}

const char *
Raid::getRaidControllerName(void)
{
	return (controller.c_str());
}
