/*	$Kolabore: sensor.cpp,v 1.26 2006/10/08 19:46:54 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * sensor.cpp:
 * Provides a way for xsi to set and read sensor data, without having to care
 * for where the data originally originated from. It is close to what the
 * OpenBSD sensor framework looks like.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/time.h>
#include <sys/types.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
}

#include <climits>
#include <cstdio>
#include <string>

#include "sensor.h"
#include "compat/xsicompat.h"

#define SMALLBUFSIZ	64

using namespace std;

Sensor::Sensor(void)
{
	init();
}

Sensor::Sensor(enum SensorType t, int id, const char *dev, int64_t val,
	       enum SensorStatus s, const char *desc, bool value_unknown,
	       bool valid_sensor)
{
	type = t;
	sensorID = id;
	device = dev;
	value = val;
	status = s;
	description = desc;
	unknown_value = value_unknown;
	is_valid = valid_sensor;
	lastChange = NULL;
}

/* virtual */
Sensor::~Sensor(void)
{
	if (lastChange != NULL)
		delete lastChange;
}

void
Sensor::init(void)
{
	type = XMLSYS_SENSOR_UNKNOWN;
	sensorID = -1;
	status = XMLSYS_SENSOR_STATUS_UNSPEC;
	lastChange = NULL;
	value = 0;
	unknown_value = true;
	is_valid = false;
}

void
Sensor::setIsValid(bool valid)
{
	is_valid = valid;
}

void
Sensor::setHasUnknownValue(bool unknown)
{
	unknown_value = unknown;
}

void
Sensor::setID(int id)
{
	sensorID = id;
}

void
Sensor::setIntValue(int64_t val)
{
	value = val;
}

void
Sensor::setStatus(enum SensorStatus new_status)
{
	status = new_status;
}

void
Sensor::setDescription(const char *desc)
{
	description = desc;
}

void
Sensor::setSensorDevice(const char *dev)
{
	device = dev;
}

void
Sensor::setValueLastChanged(struct timeval *tv)
{
	if (tv == NULL) {
		lastChange = NULL;
		return;
	}

	if (lastChange == NULL)
		lastChange = new struct timeval;
	if (lastChange == NULL)
		errx(1, "%s: Out of memory", __func__);

	lastChange->tv_sec = tv->tv_sec;
	lastChange->tv_usec = tv->tv_usec;
}

bool
Sensor::isValid(void)
{
	return (is_valid);
}

bool
Sensor::hasUnknownValue(void)
{
	return (unknown_value);
}

enum SensorType
Sensor::getSensorType(void)
{
	return (type);
}

const char *
Sensor::getSensorTypeString(void)
{
	switch (type) {
	case XMLSYS_SENSOR_TEMP:
		return ("temperature");
	case XMLSYS_SENSOR_FANRPM:
		return ("fan_rpm");
	case XMLSYS_SENSOR_VOLTS_DC:
		return ("volts_dc");
	case XMLSYS_SENSOR_VOLTS_AC:
		return ("volts_ac");
	case XMLSYS_SENSOR_OHMS:
		return ("resistance");
	case XMLSYS_SENSOR_WATTS:
		return ("watts");
	case XMLSYS_SENSOR_AMPS:
		return ("amps");
	case XMLSYS_SENSOR_WATTHOUR:
		return ("watthour");
	case XMLSYS_SENSOR_AMPHOUR:
		return ("amphour");
	case XMLSYS_SENSOR_INDICATOR:
		return ("indicator");
	case XMLSYS_SENSOR_RAW_INT:
		return ("generic");
	case XMLSYS_SENSOR_PERCENT:
		return ("percentage");
	case XMLSYS_SENSOR_LUX:
		return ("luminance");
	case XMLSYS_SENSOR_DRIVE:
		return ("drive");
	case XMLSYS_SENSOR_TIMEDELTA:
		return ("timedelta");
	case XMLSYS_SENSOR_UNKNOWN:
	default:
		return ("unknown");
	}
}

int
Sensor::getID(void)
{
	return (sensorID);
}

const char *
Sensor::getIDString(void)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	snprintf(buf, sizeof(buf), "%d", sensorID);
	ret = buf;
	return (ret.c_str());
}

enum SensorStatus
Sensor::getStatus(void)
{
	return (status);
}

const char *
Sensor::getStatusString(void)
{
	switch (status) {
	case XMLSYS_SENSOR_STATUS_UNSPEC:
		return ("");
	case XMLSYS_SENSOR_STATUS_OK:
		return ("OK");
	case XMLSYS_SENSOR_STATUS_WARN:
		return ("WARNING");
	case XMLSYS_SENSOR_STATUS_CRIT:
		return ("CRITICAL");
	case XMLSYS_SENSOR_STATUS_UNKNOWN:
		return ("unknown");
	default:
		errx(1, "%s: unknown sensor status", __func__);
	}
}

const char *
Sensor::getDescription(void)
{
	if (description.empty())
		return (NULL);
	else
		return (description.c_str());
}

const char *
Sensor::getSensorDevice(void)
{
	return (device.c_str());
}

struct timeval *
Sensor::getValueLastChanged(void)
{
	return (lastChange);
}

const char *
Sensor::getSensorUnitString(void)
{
	switch (type) {
	case XMLSYS_SENSOR_TEMP:
		return ("C");
	case XMLSYS_SENSOR_FANRPM:
		return ("rpm");
	case XMLSYS_SENSOR_VOLTS_DC:
	case XMLSYS_SENSOR_VOLTS_AC:
		return ("V");
	case XMLSYS_SENSOR_OHMS:
		return ("Ohm");
	case XMLSYS_SENSOR_WATTS:
		return ("W");
	case XMLSYS_SENSOR_AMPS:
		return ("A");
	case XMLSYS_SENSOR_WATTHOUR:
		return ("Wh");
	case XMLSYS_SENSOR_AMPHOUR:
		return ("Ah");
	case XMLSYS_SENSOR_PERCENT:
		return ("pct");
	case XMLSYS_SENSOR_LUX:
		return ("lx");
	case XMLSYS_SENSOR_TIMEDELTA:
		return ("s");
	case XMLSYS_SENSOR_INDICATOR:
	case XMLSYS_SENSOR_RAW_INT:
	case XMLSYS_SENSOR_DRIVE:
	case XMLSYS_SENSOR_UNKNOWN:
	default:
		return (NULL);
	}
}

/* virtual */
int64_t
Sensor::getValue(void)
{
	return (value);
}

/* virtual */
const char *
Sensor::getValueString(void)
{
	return (makeFloatStr(value, 1000000, 6));
}

/***************************************************************************/

TemperatureSensor::TemperatureSensor(void)
{
	init();
}

TemperatureSensor::TemperatureSensor(int id, const char *dev, int64_t val,
				     enum SensorStatus s,
				     const char *desc, bool value_unknown,
				     bool valid_sensor) :
Sensor(XMLSYS_SENSOR_TEMP, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

int64_t
TemperatureSensor::getValue(void)
{
	return (value - 273150000);
}

const char *
TemperatureSensor::getValueString(void)
{
	return (makeFloatStr(value - 273150000, 1000000, 6));
}

/***************************************************************************/

FanRPMSensor::FanRPMSensor(void)
{
	init();
}

FanRPMSensor::FanRPMSensor(int id, const char *dev, int64_t val,
			   enum SensorStatus s,
			   const char *desc, bool value_unknown,
			   bool valid_sensor) :
Sensor(XMLSYS_SENSOR_FANRPM, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

const char *
FanRPMSensor::getValueString(void)
{
	return (makeIntStr(value));
}

/***************************************************************************/

VoltsDCSensor::VoltsDCSensor(void)
{
	init();
}

VoltsDCSensor::VoltsDCSensor(int id, const char *dev, int64_t val,
			     enum SensorStatus s,
			     const char *desc, bool value_unknown,
			     bool valid_sensor) :
Sensor(XMLSYS_SENSOR_VOLTS_DC, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

VoltsACSensor::VoltsACSensor(void)
{
	init();
}

VoltsACSensor::VoltsACSensor(int id, const char *dev, int64_t val,
			     enum SensorStatus s,
			     const char *desc, bool value_unknown,
			     bool valid_sensor) :
Sensor(XMLSYS_SENSOR_VOLTS_AC, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

AmpsSensor::AmpsSensor(void)
{
	init();
}

AmpsSensor::AmpsSensor(int id, const char *dev, int64_t val,
		       enum SensorStatus s,
		       const char *desc, bool value_unknown,
		       bool valid_sensor) :
Sensor(XMLSYS_SENSOR_AMPS, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

WattsSensor::WattsSensor(void)
{
	init();
}

WattsSensor::WattsSensor(int id, const char *dev, int64_t val,
			 enum SensorStatus s,
			 const char *desc, bool value_unknown,
			 bool valid_sensor) :
Sensor(XMLSYS_SENSOR_WATTS, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

ResistanceSensor::ResistanceSensor(void)
{
	init();
}

ResistanceSensor::ResistanceSensor(int id, const char *dev, int64_t val,
				   enum SensorStatus s,
				   const char *desc, bool value_unknown,
				   bool valid_sensor) :
Sensor(XMLSYS_SENSOR_OHMS, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

WattHourSensor::WattHourSensor(void)
{
	init();
}

WattHourSensor::WattHourSensor(int id, const char *dev, int64_t val,
			       enum SensorStatus s,
			       const char *desc, bool value_unknown,
			       bool valid_sensor) :
Sensor(XMLSYS_SENSOR_WATTHOUR, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

AmpHourSensor::AmpHourSensor(void)
{
	init();
}

AmpHourSensor::AmpHourSensor(int id, const char *dev, int64_t val,
			     enum SensorStatus s,
			     const char *desc, bool value_unknown,
			     bool valid_sensor) :
Sensor(XMLSYS_SENSOR_AMPHOUR, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

IndicatorSensor::IndicatorSensor(void)
{
	init();
}

IndicatorSensor::IndicatorSensor(int id, const char *dev, int64_t val,
				 enum SensorStatus s,
				 const char *desc, bool value_unknown,
				 bool valid_sensor) :
Sensor(XMLSYS_SENSOR_INDICATOR, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

const char *
IndicatorSensor::getValueString(void)
{
	if (value == 1)
		return ("ON");
	else if (value == 0)
		return ("OFF");
	else
		return ("invalid");
}

/***************************************************************************/

GenericSensor::GenericSensor(void)
{
	init();
}

GenericSensor::GenericSensor(int id, const char *dev, int64_t val,
			     enum SensorStatus s,
			     const char *desc, bool value_unknown,
			     bool valid_sensor) :
Sensor(XMLSYS_SENSOR_RAW_INT, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

const char *
GenericSensor::getValueString(void)
{
	return (makeIntStr(value));
}

/***************************************************************************/

PercentageSensor::PercentageSensor(void)
{
	init();
}

PercentageSensor::PercentageSensor(int id, const char *dev, int64_t val,
				   enum SensorStatus s,
				   const char *desc, bool value_unknown,
				   bool valid_sensor) :
Sensor(XMLSYS_SENSOR_PERCENT, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

const char *
PercentageSensor::getValueString(void)
{
	return (makeFloatStr(value, 1000, 3));
}

/***************************************************************************/

LuminanceSensor::LuminanceSensor(void)
{
	init();
}

LuminanceSensor::LuminanceSensor(int id, const char *dev, int64_t val,
				 enum SensorStatus s,
				 const char *desc, bool value_unknown,
				 bool valid_sensor) :
Sensor(XMLSYS_SENSOR_LUX, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}

/***************************************************************************/

DriveSensor::DriveSensor(void)
{
	init();
}

DriveSensor::DriveSensor(int id, const char *dev, enum SensorDriveStatus val,
			 enum SensorStatus s, const char *desc,
			 bool value_unknown, bool valid_sensor) :
Sensor(XMLSYS_SENSOR_DRIVE, id, dev, (int64_t)val, s, desc, value_unknown, valid_sensor)
{
}

const char *
DriveSensor::getValueString(void)
{
	switch (value) {
	case XMLSYS_SENSOR_DRIVE_EMPTY:
		return ("empty");
	case XMLSYS_SENSOR_DRIVE_READY:
		return ("ready");
	case XMLSYS_SENSOR_DRIVE_POWERUP:
		return ("powering_up");
	case XMLSYS_SENSOR_DRIVE_ONLINE:
		return ("online");
	case XMLSYS_SENSOR_DRIVE_IDLE:
		return ("idle");
	case XMLSYS_SENSOR_DRIVE_ACTIVE:
		return ("active");
	case XMLSYS_SENSOR_DRIVE_REBUILD:
		return ("rebuilding");
	case XMLSYS_SENSOR_DRIVE_POWERDOWN:
		return ("powering_down");
	case XMLSYS_SENSOR_DRIVE_FAIL:
		return ("failed");
	case XMLSYS_SENSOR_DRIVE_PFAIL:
		return ("degraded");
	case XMLSYS_SENSOR_DRIVE_UNKNOWN:
	default:
		return ("unknown");
	}
}

/***************************************************************************/

TimeDeltaSensor::TimeDeltaSensor(void)
{
	init();
}

TimeDeltaSensor::TimeDeltaSensor(int id, const char *dev, int64_t val,
				 enum SensorStatus s,
				 const char *desc, bool value_unknown,
				 bool valid_sensor) :
Sensor(XMLSYS_SENSOR_TIMEDELTA, id, dev, val, s, desc, value_unknown, valid_sensor)
{
}
