/*	$Kolabore: sensor.h,v 1.18 2006/10/08 19:40:56 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * sensor.h:
 * Interface to xsi sensor data.
 * Sensors in xsi look a lot like those on OpenBSD, but can be populated with
 * data in a platform-independent way. Sensors can be of various types, meaning
 * that their raw data usually has different meanings per type. It is important
 * to supply data in the unit that these classes expect, or the displayed
 * results will be wrong.
 */

#ifndef __SENSOR_H__
#define __SENSOR_H__

#include "util.h"

using namespace std;

/*
 * Supported sensor types. This is set automatically by the various
 * constructors. If your sensor type isn't listed here, please use
 * a GenericSensor in your port and send in a bug report.
 */
enum SensorType {
	XMLSYS_SENSOR_TEMP,
	XMLSYS_SENSOR_FANRPM,
	XMLSYS_SENSOR_VOLTS_DC,
	XMLSYS_SENSOR_VOLTS_AC,
	XMLSYS_SENSOR_OHMS,
	XMLSYS_SENSOR_WATTS,
	XMLSYS_SENSOR_AMPS,
	XMLSYS_SENSOR_WATTHOUR,
	XMLSYS_SENSOR_AMPHOUR,
	XMLSYS_SENSOR_INDICATOR,
	XMLSYS_SENSOR_RAW_INT,
	XMLSYS_SENSOR_PERCENT,
	XMLSYS_SENSOR_LUX,
	XMLSYS_SENSOR_DRIVE,
	XMLSYS_SENSOR_TIMEDELTA,
	XMLSYS_SENSOR_UNKNOWN
};

/*
 * Disk enclosures often have their own sensors, and their value's meaning
 * represents a certain disk state.
 */
enum SensorDriveStatus {
	XMLSYS_SENSOR_DRIVE_EMPTY,
	XMLSYS_SENSOR_DRIVE_READY,
	XMLSYS_SENSOR_DRIVE_POWERUP,
	XMLSYS_SENSOR_DRIVE_ONLINE,
	XMLSYS_SENSOR_DRIVE_IDLE,
	XMLSYS_SENSOR_DRIVE_ACTIVE,
	XMLSYS_SENSOR_DRIVE_REBUILD,
	XMLSYS_SENSOR_DRIVE_POWERDOWN,
	XMLSYS_SENSOR_DRIVE_FAIL,
	XMLSYS_SENSOR_DRIVE_PFAIL,
	XMLSYS_SENSOR_DRIVE_UNKNOWN
};

/*
 * Some sensors can also report whether their value is OK, not-so-nice or
 * "Something's-about-to-go-horribly-wrong!!". Sensors without this feature
 * should supply XMLSYS_SENSOR_STATUS_UNSPEC.
 */
enum SensorStatus {
	XMLSYS_SENSOR_STATUS_UNSPEC,
	XMLSYS_SENSOR_STATUS_OK,
	XMLSYS_SENSOR_STATUS_WARN,
	XMLSYS_SENSOR_STATUS_CRIT,
	XMLSYS_SENSOR_STATUS_UNKNOWN
};

/*
 * Sensor base class. Should not be used directly, use one of the specific
 * sensor type classes below.
 * If possible, use constructors directly instead of the various set*()
 * methods.
 */
class Sensor : protected Util {
public:
	Sensor(void);
	Sensor(enum SensorType, int, const char *, int64_t, enum SensorStatus,
	       const char *, bool value_unknown = false,
	       bool valid_sensor = true);
	virtual ~Sensor(void);

	void			init(void);

	void			setIsValid(bool);
	void			setHasUnknownValue(bool);
	void			setID(int);	// Sensor ID on system
	void			setIntValue(int64_t);
	void			setStatus(enum SensorStatus);
	void			setDescription(const char *);
	void			setSensorDevice(const char *);
	void			setValueLastChanged(struct timeval *);

	bool			isValid(void);
	bool			hasUnknownValue(void);
	enum SensorType 	getSensorType(void);
	const char *		getSensorTypeString(void);
	int			getID(void);
	const char *		getIDString(void);
	enum SensorStatus	getStatus(void);
	const char *		getStatusString(void);
	const char *		getDescription(void);
	const char *		getSensorDevice(void);
	struct timeval *	getValueLastChanged(void);
	const char *		getSensorUnitString(void);

	virtual int64_t 	getValue(void);
	virtual const char *	getValueString(void);

protected:
	enum SensorType 	type;
	bool			is_valid, unknown_value;
	int			sensorID;
	enum SensorStatus	status;
	string			description, device;
	struct timeval *	lastChange;
	int64_t 		value;
};

/*
 * Sensors:
 * Constructors always take
 *  - int ID (sensor ID as found on the system)
 *  - const char *devicename (the device/driver name of the sensor)
 *  - int64_t value (has a different meaning per type, see comments below)
 *  - enum SensorStatus status (sensor value status, as reported by the sensor)
 *  - const char *description (descr. as reported by the sensor device
 *			       (set by vendor, cannot be relied upon))
 *  - bool value_unknown = false (does the sensor know its own value?)
 *  - bool valid_sensor = true (is the sensor valid / does it even exist?)
 */

/* Temperature, value is uK ('u' for micro / "mu") */
class TemperatureSensor : public Sensor {
public:
	TemperatureSensor(void);
	TemperatureSensor(int, const char *, int64_t, enum SensorStatus,
			  const char *, bool value_unknown = false,
			  bool valid_sensor = true);

	int64_t 	getValue(void);
	const char *	getValueString(void);
};

/* Fan RPM, value is revolutions per minute */
class FanRPMSensor : public Sensor {
public:
	FanRPMSensor(void);
	FanRPMSensor(int, const char *, int64_t, enum SensorStatus,
		     const char *, bool value_unknown = false,
		     bool valid_sensor = true);

	const char *	getValueString(void);
};

/* Voltage, value is uV DC */
class VoltsDCSensor : public Sensor {
public:
	VoltsDCSensor(void);
	VoltsDCSensor(int, const char *, int64_t, enum SensorStatus,
		      const char *, bool value_unknown = false,
		      bool valid_sensor = true);
};

/* Voltage, value is uV AC */
class VoltsACSensor : public Sensor {
public:
	VoltsACSensor(void);
	VoltsACSensor(int, const char *, int64_t, enum SensorStatus,
		      const char *, bool value_unknown = false,
		      bool valid_sensor = true);
};

/* Amperes, value is uA */
class AmpsSensor : public Sensor {
public:
	AmpsSensor(void);
	AmpsSensor(int, const char *, int64_t, enum SensorStatus,
		   const char *, bool value_unknown = false,
		   bool valid_sensor = true);
};

/* Watts, value is uW */
class WattsSensor : public Sensor {
public:
	WattsSensor(void);
	WattsSensor(int, const char *, int64_t, enum SensorStatus,
		    const char *, bool value_unknown = false,
		    bool valid_sensor = true);
};

/* Resistance, value is uOhm */
class ResistanceSensor : public Sensor {
public:
	ResistanceSensor(void);
	ResistanceSensor(int, const char *, int64_t, enum SensorStatus,
			 const char *, bool value_unknown = false,
			 bool valid_sensor = true);
};

/* Watthours, value is uWh */
class WattHourSensor : public Sensor {
public:
	WattHourSensor(void);
	WattHourSensor(int, const char *, int64_t, enum SensorStatus,
		       const char *, bool value_unknown = false,
		       bool valid_sensor = true);
};

/* Amperehours, value is uAh */
class AmpHourSensor : public Sensor {
public:
	AmpHourSensor(void);
	AmpHourSensor(int, const char *, int64_t, enum SensorStatus,
		      const char *, bool value_unknown = false,
		      bool valid_sensor = true);
};

/* On/Off-sensor, 1 for 'on', 0 for 'off' */
class IndicatorSensor : public Sensor {
public:
	IndicatorSensor(void);
	IndicatorSensor(int, const char *, int64_t, enum SensorStatus,
			const char *, bool value_unknown = false,
			bool valid_sensor = true);

	const char *	getValueString(void);
};

/* Raw data, value is integer data directly from the sensor */
class GenericSensor : public Sensor {
public:
	GenericSensor(void);
	GenericSensor(int, const char *, int64_t, enum SensorStatus,
		      const char *, bool value_unknown = false,
		      bool valid_sensor = true);

	const char *	getValueString(void);
};

/*
 * Percentage, value is one tenth of a basis point, i.e. one-thousandth of a
 * percent (0.001%) -- one could call this unit "dbp", the "deci-basis-point".
 * The output of getValueString() is still percent, though, with a precision
 * of 3 decimal digits.
 */
class PercentageSensor : public Sensor {
public:
	PercentageSensor(void);
	PercentageSensor(int, const char *, int64_t, enum SensorStatus,
			 const char *, bool value_unknown = false,
			 bool valid_sensor = true);

	const char *	getValueString(void);
};

/* Luminance, value is ulx */
class LuminanceSensor : public Sensor {
public:
	LuminanceSensor(void);
	LuminanceSensor(int, const char *, int64_t, enum SensorStatus,
			const char *, bool value_unknown = false,
			bool valid_sensor = true);
};

/* Drive status, value is enum SensorDriveStatus */
class DriveSensor : public Sensor {
public:
	DriveSensor(void);
	DriveSensor(int, const char *, enum SensorDriveStatus,
		    enum SensorStatus, const char *,
		    bool value_unknown = false, bool valid_sensor = true);

	const char *	getValueString(void);
};

/* System time error, value is nanoseconds */
class TimeDeltaSensor : public Sensor {
public:
	TimeDeltaSensor(void);
	TimeDeltaSensor(int, const char *, int64_t, enum SensorStatus,
			const char *, bool value_unknown = false,
			bool valid_sensor = true);
};

#endif /* __SENSOR_H__ */
