/*	$Kolabore: server.cpp,v 1.30 2006/10/24 22:10:27 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * server.cpp:
 * as-simple-as-possible server class.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#include <netdb.h>
#include <poll.h>
#include <unistd.h>
}

#include <list>
#include <string>
#include <vector>

#include <cerrno>
#include <cstring>

#include "server.h"
#include "compat/xsicompat.h"

extern int errno;

using namespace std;

Server::Server(Config *c)
{
	int i;

	if (c == NULL)
		errx(1, "%s: configuration cannot be NULL", __func__);
	cfg = c;
	may_talk = cfg->dontDetach();
	memset(pfd, 0, MAX_POLLFDS * sizeof(struct pollfd));
	for (i = 0; i < MAX_POLLFDS; i++)
		pfd[i].fd = -1;
	listen();
}

Server::~Server(void)
{
	int i;

	for (i = 0; i < MAX_POLLFDS; i++) {
		if (pfd[i].fd != -1)
		    close(pfd[i].fd);
	}
}

void
Server::listen(void)
{
	struct addrinfo        *res, *res0 = cfg->getListenAddr();
	int			error, i;
	const int		yes = 1;
	char			sbuf[NI_MAXSERV];

	addr_count = 0;
	for (res = res0, i = 0; res != NULL && i < MAX_POLLFDS;
	     res = res->ai_next) {
		char	hbuf[MAXHOSTNAMELEN];

		if ((pfd[i].fd = socket(res->ai_family, res->ai_socktype,
					res->ai_protocol)) == -1) {
			if (errno == EAFNOSUPPORT) {
				/*
				 * This happens on Frankenstein'ed systems
				 * where the kernel doesn't support IPv6
				 * and the resolver returns IPv6 addresses
				 * anyways.
				 */
				warn("%s: socket", __func__);
				continue;
			} else
				err(1, "%s: socket", __func__);
		}
		if (setsockopt(pfd[i].fd, SOL_SOCKET, SO_REUSEADDR, &yes,
			       sizeof(yes)) == -1)
			err(1, "%s: setsockopt (reuseaddr)", __func__);
		if (res->ai_family == AF_INET6
		    && setsockopt(pfd[i].fd, IPPROTO_IPV6, IPV6_V6ONLY, &yes,
				  sizeof(yes)) == -1)
			err(1, "%s: setsockopt (v6)", __func__);
		if (bind(pfd[i].fd, res->ai_addr, res->ai_addrlen) == -1)
			err(1, "%s: bind", __func__);
		if (::listen(pfd[i].fd, 10) == -1)
			err (1, "%s: listen", __func__);

		pfd[i].events = POLLIN;
		i++;
		addr_count++;

		error = getnameinfo(res->ai_addr, res->ai_addrlen,
				    hbuf, sizeof(hbuf), NULL, 0,
				    NI_NUMERICHOST);
		if (error)
			errx(1, "%s: getnameinfo: %s", __func__,
			     gai_strerror(error));

		listenAddrs.push_back(hbuf);
	}

	error = getnameinfo(res0->ai_addr, res0->ai_addrlen, NULL, 0,
			    sbuf, sizeof(sbuf), NI_NUMERICSERV);
	if (error)
		errx(1, "%s: getnameinfo: %s", __func__, gai_strerror(error));

	listenService = sbuf;
}

vector<struct client>
Server::poll(void)
{
	vector<struct client> clients;
	int pret, i;

	if ((pret = ::poll(pfd, addr_count, -1)) == -1) {
		if (errno != EINTR)
			err(1, "%s: poll", __func__);
		// a signal interrupted us, probably SIGALRM;
		// return empty clients vector to tell the caller
		// that it should do some maintenance
		return (clients);
	}

	/*
	if (pret == 0)
		return (clients);
	*/

	for (i = 0; i < addr_count; i++) {
		struct sockaddr_storage c_addr;
		socklen_t c_addr_len = sizeof(c_addr);

		if (pfd[i].revents & (POLLERR | POLLNVAL))
			errx(1, "%s: poll error", __func__);
		if (pfd[i].revents & POLLIN) {
			struct client	client;

			if ((client.fd = accept(pfd[i].fd,
						(struct sockaddr *)&c_addr,
						&c_addr_len)) == -1)
				err(1, "%s: accept", __func__);
			memcpy(&client.c_addr, &c_addr, c_addr_len);
			client.c_addr_len = c_addr_len;

			clients.push_back(client);
		} else
			continue;
	}

	if (pret != (int)clients.size())
		errx(1, "%s: Developer brain-death error. :'-(", __func__);

	return (clients);
}

vector<string>
Server::getListenAddrStrings(void)
{
	return (listenAddrs);
}

string
Server::getListenServiceString(void)
{
	return (listenService);
}

#if 0
int
Server::parseClientRequest(string request, string &sReply, string &sFile, string &cProto, string &cAgent)
{
	vector<string> lines, reqCommand;
	string command, uri, protocol;
	vector<string>::iterator it;

	lines = split(request);
	if (lines.empty())
		return (-1);
	reqCommand = split(lines[0], ' ');

	if (reqCommand.size() < 3)
		return (-1);

	command = reqCommand[0];
	uri = reqCommand[1];
	protocol = reqCommand[2];

	if (command != "GET" || uri.empty() || protocol.empty())
		return (-1); // received non-HTTP garbage

	sReply = "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Type: application/xml; charset=utf-8\r\n\r\n";
	cProto = protocol;
	sFile = uri;

	cAgent = "";
	for (it = lines.begin(); it != lines.end(); it++)
		if ((*it).find("User-Agent: ", 0) == 0) {
			cAgent = it->substr(sizeof("User-Agent:"));
			break;
		}

	return (0);
}
#endif
