/*	$Kolabore: system.cpp,v 1.10 2006/10/03 09:34:31 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
}

#include <cstdio>

#include <string>
#include <vector>

#include "system.h"
#include "timekeeper.h"
#include "xmlsysinfo.h"
#include "compat/xsicompat.h"

#define SMALLBUFSIZ	64

using namespace std;

XSI_Interrupt::XSI_Interrupt(void)
{
	count = last_count = 0;
}

XSI_Interrupt::XSI_Interrupt(const char *src)
{
	name = src;
	count = last_count = 0;
}

void
XSI_Interrupt::setSourceName(const char *src)
{
	name = src;
}

void
XSI_Interrupt::setIntsCount(uint64_t n)
{
	last_count = count;
	count = n;
}

bool
XSI_Interrupt::hasInterrupts(void)
{
	return (count > 0);
}

const char *
XSI_Interrupt::getSourceName(void)
{
	return (name.c_str());
}

double
XSI_Interrupt::getIntsPerSec(void)
{
	uint64_t value;

	if (!count || XSI_TimeKeeper::get()->timeDelta() < 0.01 || !last_count)
		return (0);

	value = count - last_count;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}

const char *
XSI_Interrupt::getIntsPerSecString(void)
{
	return (makeFloatStr(getIntsPerSec(), 1));
}


XSI_System::XSI_System(void)
{
	vm_info.ram_total = vm_info.ram_free = 0;
	vm_info.swap_total = vm_info.swap_free = 0;
	uptime_seconds = 0;
	load_averages[0] = load_averages[1] = load_averages[2] = 0;
	num_cpus = cpu_freq = -1;
}

void
XSI_System::setArch(const char *str)
{
	arch = str;
}

void
XSI_System::setModel(const char *str)
{
	model = str;
}

void
XSI_System::setVendor(const char *str)
{
	vendor = str;
}

void
XSI_System::setProduct(const char *str)
{
	product = str;
}

void
XSI_System::setVersion(const char *str)
{
	version = str;
}

void
XSI_System::setSerialNo(const char *str)
{
	serial_no = str;
}

void
XSI_System::setUUID(const char *str)
{
	uuid = str;
}

void
XSI_System::setNumCPUs(int32_t ncpu)
{
	num_cpus = ncpu;
}

void
XSI_System::setCPUFreq(int32_t speed)
{
	cpu_freq = speed;
}

void
XSI_System::setCPUs(vector<struct cpu> ncpus)
{
	cpus = ncpus;
}

void
XSI_System::setInterrupts(vector<XSI_Interrupt> ints)
{
	interrupts = ints;
}

void
XSI_System::setVMInfo(struct vm_info mem)
{
	vm_info = mem;
}

void
XSI_System::setLoadAvg(double load[3])
{
	load_averages[0] = load[0];
	load_averages[1] = load[1];
	load_averages[2] = load[2];
}

void
XSI_System::setOSInfo(const char *type, const char *release,
			const char *status, const char *kernel,
			const char *version_string)
{
	os_type = type;
	os_release = release;
	if (status)
		os_status = status;
	os_kernel = kernel;
	os_version_string = version_string;
}

void
XSI_System::setOSUptime(time_t uptime)
{
	uptime_seconds = uptime;
}

const char *
XSI_System::getArch(void)
{
	return (arch.c_str());
}

const char *
XSI_System::getModel(void)
{
	if (model.empty())
		return (NULL);
	else
		return (model.c_str());
}

const char *
XSI_System::getCPUFreqString(void)
{
	if (cpu_freq == -1)
		return (NULL);
	else
		return (makeIntStr(cpu_freq));
}

const char *
XSI_System::getVendor(void)
{
	if (vendor.empty())
		return (NULL);
	else
		return (vendor.c_str());
}

const char *
XSI_System::getProduct(void)
{
	if (product.empty())
		return (NULL);
	else
		return (product.c_str());
}

const char *
XSI_System::getVersion(void)
{
	if (version.empty())
		return (NULL);
	else
		return (version.c_str());
}

const char *
XSI_System::getSerialNo(void)
{
	if (serial_no.empty())
		return (NULL);
	else
		return (serial_no.c_str());
}

const char *
XSI_System::getUUID(void)
{
	if (uuid.empty())
		return (NULL);
	else
		return (uuid.c_str());
}

const char *
XSI_System::getNumCPUString(void)
{
	return (makeIntStr(num_cpus));
}

vector<struct cpu>
XSI_System::getCPUs(void)
{
	return (cpus);
}

vector<XSI_Interrupt>
XSI_System::getInterrupts(void)
{
	return (interrupts);
}

struct vm_info
XSI_System::getVMInfo(void)
{
	return (vm_info);
}

const char *
XSI_System::getLoadAvg(int sample)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	if (sample < 0 || sample > 2)
		errx(1, "%s: Internal error: Bad sample number requested", __func__);
	snprintf(buf, sizeof(buf), "%.2f", load_averages[sample]);
	ret = buf;
	return (ret.c_str());
}

const char *
XSI_System::getOSType(void)
{
	return (os_type.c_str());
}

const char *
XSI_System::getOSRelease(void)
{
	return (os_release.c_str());
}

const char *
XSI_System::getOSStatus(void)
{
	if (os_status.empty())
		return (NULL);
	else
		return (os_status.c_str());
}

const char *
XSI_System::getOSKernel(void)
{
	return (os_kernel.c_str());
}

const char *
XSI_System::getOSVersionString(void)
{
	return (os_version_string.c_str());
}

const char *
XSI_System::getOSUptime(void)
{
	return (makeIntStr((int64_t)uptime_seconds));
}
