/*	$Kolabore: system.h,v 1.11 2006/10/08 14:23:32 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * system.h
 * All about the generic hardware and OS of the entire system, sans those
 * parts deemed special.
 */

#ifndef __SYSTEM_H__
#define __SYSTEM_H__

#include "util.h"

using namespace std;

/*
 * NOTE: If a comment explaining a method is prefixed with either MANDATORY
 *       or OPTIONAL, this means that MANDATORY information is required
 *       for XSI to generate valid (according to its Schema) XML. It has no
 *       meaning concerning the importance of the particular information or
 *       statistic. Libxml2's xmllint can be used as a regression test.
 */

/*
 * struct cpu:
 * Contains the current distribution of load on the various CPU states of
 * CPU <num> (count starts at 0.) The floats are percentages from 0.0 to 100.0.
 * Should be filled using all CPUSTATE counters at the beginning, to show
 * load distribution since the last reboot in -o (one-shot) mode.
 *
 * If an SMP system cannot fill one struct cpu per CPU for the one or other
 * reason, it fills only _one_ struct and sets int num to -1.
 */
struct cpu {
	int	num;
	float	system;
	float	user;
	float	nice;
	float	interrupt;
	float	idle;
};

/*
 * struct vm_info:
 * contains the usual suspects for RAM and swap usage, in bytes.
 */
struct vm_info {
	uint64_t	ram_total;
	uint64_t	ram_free;
	uint64_t	swap_total;
	uint64_t	swap_free;
};

class XSI_Interrupt : private Util {
public:
	XSI_Interrupt(void);
	XSI_Interrupt(const char *);

	void		setSourceName(const char *);
	void		setIntsCount(uint64_t);

	bool		hasInterrupts(void);
	const char *	getSourceName(void);
	double		getIntsPerSec(void);
	const char *	getIntsPerSecString(void);

private:
	string		name;
	uint64_t	count, last_count;
};

/*
 * XSI_System class
 * Gets and outputs generic information about the computer, including several
 * vendor strings. Such strings are set by OEMs, mainboard vendors, CPU vendors
 * et cetera and are purely informational. They should not be parsed, as they
 * follow no rules.
 * Keeps the following up-to-date information:
 *  - CPU state load distribution per CPU
 *  - interrupt sources and interrupts-per-device-per-second
 *  - memory and swap usage
 *  - uptime
 *  - "load average", that is, numbers describing the process queue length
 *    (which is OS-dependent)
 */
class XSI_System : private Util {
public:
	XSI_System(void);

	/* MANDATORY: Set system arch, e.g. i386, amd64, sparc64, ... */
	void			setArch(const char *);
	/* OPTIONAL: Set (CPU) model, a vendor string. E.g. "AMD Athlon(tm)
	 * ("AuthenticAMD" 686-class, 256KB L2 cache)" */
	void			setModel(const char *);
	/* OPTIONAL: Set CPU frequency, in MHz */
	void			setCPUFreq(int32_t);
	/* OPTIONAL: System/mainboard vendor; vendor string */
	void			setVendor(const char *);
	/* OPTIONAL: Product (e.g. mainboard) or system (e.g. laptop); vendor
	 * string */
	void			setProduct(const char *);
	/* OPTIONAL: Product version/revision; vendor string */
	void			setVersion(const char *);
	/* OPTIONAL: Product serial number; vendor string */
	void			setSerialNo(const char *);
	/* OPTIONAL: Set the system's UUID */
	void			setUUID(const char *);
	/* MANDATORY: Set the number of CPUs in the system */
	void			setNumCPUs(int32_t);
	/* MANDATORY: Set vector<struct cpu>, NumCPUs large, containing up-to
	 * date information wrt CPU state load distribution */
	void			setCPUs(vector<struct cpu>);
	/* OPTIONAL: Set vector of interrupt sources and their stats */
	void			setInterrupts(vector<XSI_Interrupt>);
	/* OPTIONAL: Set up-to-date memory/swap usage information */
	void			setVMInfo(struct vm_info);
	/*
	 * MANDATORY: Operating system information.
	 * status: May be NULL on systems with no userland-visible notion of
	 *         things like "release", "stable", "current", "beta", etc.
	 * kernel: The ``normalized'' kernel version in the format
	 *         <kernel_name>#<kernel_build_number> <kernel_build_datestamp>
	 *         e.g.:
	 *         SMP#1 200609131634 (on a Linux system with SMP kernel)
	 *         GENERIC#933 200610081150 (*BSD GENERIC kernel)
	 * version_string: Something that the OS would put as the first line
	 *         in /etc/motd; verbose and system-dependent.
	 */
	void			setOSInfo(const char *type, /* uname -s */
					  const char *release, /* uname -r */
					  const char *status, /* see above */
					  const char *kernel, /* see above */
					  const char *version_string); /* see above */
	/* OPTIONAL: Seconds since the last reboot */
	void			setOSUptime(time_t);
	/* OPTIONAL: Array of three doubles containing the number
	 * representing the load average of the last one, five and
	 * 15 minutes respectively. */
	void			setLoadAvg(double[3]);

	const char *		getArch(void);
	const char *		getModel(void);
	const char *		getCPUFreqString(void);
	const char *		getVendor(void);
	const char *		getProduct(void);
	const char *		getVersion(void);
	const char *		getSerialNo(void);
	const char *		getUUID(void);
	const char *		getNumCPUString(void);
	vector<struct cpu>	getCPUs(void);
	vector<XSI_Interrupt>	getInterrupts(void);
	struct vm_info		getVMInfo(void);
	const char *		getOSType(void);
	const char *		getOSRelease(void);
	const char *		getOSStatus(void);
	const char *		getOSKernel(void);
	const char *		getOSVersionString(void);
	const char *		getOSUptime(void);
	const char *		getLoadAvg(int);

private:
	string			arch, model, vendor, product, version,
				serial_no, uuid;
	int32_t 		num_cpus, cpu_freq;
	vector<struct cpu>	cpus;
	vector<XSI_Interrupt>	interrupts;
	struct vm_info		vm_info;
	double			load_averages[3];
	string			os_type, os_release, os_status, os_kernel,
				os_version_string;
	time_t			uptime_seconds;
};

#endif /* __SYSTEM_H__ */
