/*	$Kolabore: util.cpp,v 1.12 2006/10/29 01:23:45 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
}

#include <fstream>
#include <string>
#include <vector>

#include <cstdio>

#include "util.h"
#include "compat/xsicompat.h"

#define SMALLBUFSIZ	64

using namespace std;

const char *
Util::makeFloatStr(int64_t num, double divisor, unsigned short precision)
{
	char buf[SMALLBUFSIZ];
	char fmt[10];
	static string ret;

	snprintf(fmt, sizeof(fmt), "%%.%huf", precision);

	if (divisor < 0.0001)
		snprintf(buf, sizeof(buf), fmt, 0.0);
        else
		snprintf(buf, sizeof(buf), fmt, (double)num / divisor);

	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeFloatStr(uint64_t num, double divisor, unsigned short precision)
{
	char buf[SMALLBUFSIZ];
	char fmt[10];
	static string ret;

	snprintf(fmt, sizeof(fmt), "%%.%huf", precision);

	if (divisor < 0.0001)
		snprintf(buf, sizeof(buf), fmt, 0.0);
        else
		snprintf(buf, sizeof(buf), fmt, (double)num / divisor);

	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeFloatStr(double num, double divisor, unsigned short precision)
{
	char buf[SMALLBUFSIZ];
	char fmt[10];
	static string ret;

	snprintf(fmt, sizeof(fmt), "%%.%huf", precision);

	if (divisor < 0.0001)
		snprintf(buf, sizeof(buf), fmt, 0.0);
        else
		snprintf(buf, sizeof(buf), fmt, num / divisor);

	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeIntStr(int64_t num)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	snprintf(buf, sizeof(buf), "%" PRIi64, num);
	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeIntStr(uint64_t num)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	snprintf(buf, sizeof(buf), "%" PRIu64, num);
	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeIntStr(int32_t num)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	snprintf(buf, sizeof(buf), "%d", num);
	ret = buf;
	return (ret.c_str());
}

const char *
Util::makeIntStr(uint32_t num)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	snprintf(buf, sizeof(buf), "%u", num);
	ret = buf;
	return (ret.c_str());
}

vector<string>
Util::split(string str, char separator)
{
	string		tmp;
	string::iterator c;
	vector<string>	ret;
	bool		is_separator;

	is_separator = true;
	for (c = str.begin(); c != str.end(); c++) {
		/* be paranoid; control chars are separators, too */
		if (*c == separator || *c < 0x20) {
			if (!is_separator && !tmp.empty())
				ret.push_back(tmp);
			is_separator = true;
			tmp = "";
		} else {
			is_separator = false;
			tmp += *c;
		}
	}
	if (!tmp.empty())
		ret.push_back(tmp);

	return (ret);
}

vector<string>
Util::getFileVector(const char *file, bool *must_exist)
{
	vector<string> ret;
	string line;
	ifstream f;

	f.open(file, ifstream::in);
	if (!f.is_open()) {
		if (must_exist != NULL) {
			if (*must_exist)
				err(1, "%s: Error opening %s", __func__, file);
			else {
				*must_exist = true;
				return (ret);
			}
		} else
			err(1, "%s: Error opening %s", __func__, file);
	}


	while ((int)f.peek() != EOF) {
		line.push_back(f.get());
		if (line[line.size() - 1] == '\n') {
			line.resize(line.size() - 1);	// trim trailing newline
			ret.push_back(line);
			line = "";
		}
	}
	if (!line.empty() && !line[0] == '\n')
		ret.push_back(line);

	f.close();

	return (ret);
}

string
Util::getFileString(const char *file, bool *must_exist)
{
	string ret;
	ifstream f;

	f.open(file, ifstream::in);
	if (!f.is_open()) {
		if (must_exist != NULL) {
			if (*must_exist)
				err(1, "%s: Error opening %s", __func__, file);
			else {
				*must_exist = true;
				return (ret);
			}
		} else
			err(1, "%s: Error opening %s", __func__, file);
	}

	while (f.peek() != '\n' && (int)f.peek() != EOF) {
		ret.push_back(f.get());
	}

	return (ret);
}

int64_t
Util::getValueFromFile(const char *file, int64_t min, int64_t max)
{
	string s = getFileString(file);
	const char *errstr;
	int64_t ret;

	ret = strtonum(s.c_str(), min, max, &errstr);
	if (errstr)
		err(1, "%s: %s is %s", __func__,
		    s.c_str(), errstr);
	return (ret);
}

int64_t
Util::getValueFromFile(const char *file, bool *error, int64_t min, int64_t max)
{
	string s;
	const char *errstr;
	int64_t ret;
	bool exist = false;

	s = getFileString(file, &exist);
	if (exist) {
		*error = true;
		return (0);
	}

	ret = strtonum(s.c_str(), min, max, &errstr);
	if (errstr)
		*error = true;
	else
		*error = false;
	return (ret);
}
