/*	$Kolabore: xmlsysinfo.cpp,v 1.87 2006/10/28 09:49:04 maxx Exp $	*/
/*
 * Copyright (c) 2006 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "typesetup.h"

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>

#ifdef HAVE_ERR_H
#  include <err.h>
#endif
#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif
#include <unistd.h>

#include <libxml/tree.h>
}

#include <cstring>

#include <list>
#include <string>
#include <vector>

#include "xmlsysinfo.h"
#include "timekeeper.h"
#include "compat/xsicompat.h"

#define SMALLBUFSIZ	64

using namespace std;


/***************************************************************************/

XSIBase::XSIBase(void)
{
	sysInfoDocument = NULL;
	configuration = NULL;
}

XSIBase::~XSIBase(void)
{
	vector<Sensor *>::iterator it;
	for (it = sensors.begin(); it != sensors.end(); it++)
		delete *it;
	if (sysInfoDocument != NULL)
		xmlFreeDoc(sysInfoDocument);
	xmlCleanupParser();
}

void
XSIBase::XSIBase_init(Config *cfg)
{
	configuration = cfg;
}

void
XSIBase::setListenAddrs(vector<string> laddrs)
{
	listenAddrs = laddrs;
}

void
XSIBase::setListenService(string service)
{
	listenService = service;
}

void
XSIBase::readSysInfo(void)
{
	collectSysInfo();
	updateSysInfo();
}

void
XSIBase::updateSysInfo(void)
{
	updateTimeKeeper();
	updateStats();
}

void
XSIBase::buildSysInfoDoc(void)
{
	xmlNodePtr	root_node = NULL, node1 = NULL, node2 = NULL,
			node3 = NULL, node4 = NULL, node5 = NULL;
	xmlNsPtr	ns = NULL;
	char		buffer[BUFSIZ];

	if (sysInfoDocument != NULL)
		xmlFreeDoc(sysInfoDocument);

	/*
	 * XXX:
	 * The whole namespace nonsense that follows is a stupid hack, simply
	 * to coerce libxml2 into putting out what I want. I am certain that
	 * this is not how it's supposed to be done, but it serves its purpose
	 * until either the secret, proper way of doing this is discovered,
	 * libxml actually supports XML Schemas, ...
	 */
	sysInfoDocument = xmlNewDoc(BAD_CAST "1.0");
	root_node = xmlNewNode(NULL, BAD_CAST "xmlsysinfo");
	ns = xmlNewNs(root_node, BAD_CAST "http://xsi.kolabore.ath.cx", NULL);
	xmlNewNsProp(root_node, ns, BAD_CAST "xmlns:xsi",
		     BAD_CAST "http://www.w3.org/2001/XMLSchema-instance");
	xmlNewNsProp(root_node, ns, BAD_CAST "xsi:schemaLocation",
		     BAD_CAST XSI_XSD_URI);
	xmlDocSetRootElement(sysInfoDocument, root_node);

	/* XSI CONFIGURATION */

	node1 = xmlNewNode(NULL, BAD_CAST "xsi_config");
	xmlAddChild(root_node, node1);
	xmlNewProp(node1, BAD_CAST "schema_version", BAD_CAST XSI_XSD_VERSION);
	xmlNewProp(node1, BAD_CAST "xsi_version", BAD_CAST PACKAGE_VERSION);

	xmlNewChild(node1, NULL, BAD_CAST "update_timestamp", BAD_CAST XSI_TimeKeeper::get()->updateTimestampString());
	{
		list<string> opts = configuration->getXSIOptionsList();
		list<string>::iterator it;

		for (it = opts.begin(); it != opts.end(); it++) {
			xmlNewChild(node1, NULL, BAD_CAST (*it).c_str(), NULL);
		}
	}
	if (!configuration->oneShot()) {
		vector<string>::iterator it;

		if (XSI_TimeKeeper::get()->haveTimeDelta())
			xmlNewChild(node1, NULL, BAD_CAST "update_interval", BAD_CAST makeFloatStr(XSI_TimeKeeper::get()->timeDelta(), 1, 4));

		for (it = listenAddrs.begin(); it != listenAddrs.end(); it++) {
			xmlNewChild(node1, NULL, BAD_CAST "listen_addr", BAD_CAST (*it).c_str());
		}
		xmlNewChild(node1, NULL, BAD_CAST "listen_port", BAD_CAST listenService.c_str());
	}

	/* SYSTEM INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "system");
	xmlAddChild(root_node, node1);

	node2 = xmlNewNode(NULL, BAD_CAST "machine");
	xmlAddChild(node1, node2);
	xmlNewChild(node2, NULL, BAD_CAST "arch",
		    BAD_CAST system.getArch());
	if (system.getModel())
		xmlNewChild(node2, NULL, BAD_CAST "model",
			    BAD_CAST system.getModel());
	if (system.getCPUFreqString())
		xmlNewChild(node2, NULL, BAD_CAST "cpu_freq",
			    BAD_CAST system.getCPUFreqString());
	if (system.getVendor())
		xmlNewChild(node2, NULL, BAD_CAST "vendor",
			    BAD_CAST system.getVendor());
	if (system.getProduct())
		xmlNewChild(node2, NULL, BAD_CAST "product",
			    BAD_CAST system.getProduct());
	if (system.getVersion())
		xmlNewChild(node2, NULL, BAD_CAST "version",
			    BAD_CAST system.getVersion());
	if (system.getSerialNo())
		xmlNewChild(node2, NULL, BAD_CAST "serial_no",
			    BAD_CAST system.getSerialNo());
	if (system.getUUID())
		xmlNewChild(node2, NULL, BAD_CAST "uuid",
			    BAD_CAST system.getUUID());
	if (!system.getCPUs().empty()) {
		node2 = xmlNewNode(NULL, BAD_CAST "cpus");
		xmlAddChild(node1, node2);
		xmlNewProp(node2, BAD_CAST "count",
			   BAD_CAST system.getNumCPUString());
		{
			vector<struct cpu> cpuvect = system.getCPUs();
			vector<struct cpu>::iterator it;

			for (it = cpuvect.begin(); it != cpuvect.end(); it++) {
				node3 = xmlNewNode(NULL, BAD_CAST (configuration->oneShot()
								   ? "cpu_load_total"
								   : "cpu_load"));
				/*
				 * OSs that can only report overall CPU loads, even on SMP systems,
				 * set only one struct cpu and make num -1.
				 */
				if (cpuvect.size() != 1 || it->num != -1) {
					snprintf(buffer, sizeof(buffer), "%d", it->num);
					xmlNewProp(node3, BAD_CAST "num", BAD_CAST buffer);
				}
				xmlAddChild(node2, node3);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->system);
				xmlNewChild(node3, NULL, BAD_CAST "system",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->user);
				xmlNewChild(node3, NULL, BAD_CAST "user",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->nice);
				xmlNewChild(node3, NULL, BAD_CAST "nice",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->interrupt);
				xmlNewChild(node3, NULL, BAD_CAST "interrupt",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->idle);
				xmlNewChild(node3, NULL, BAD_CAST "idle",
					    BAD_CAST buffer);
			}
		}
	}
	if (!configuration->oneShot() && !system.getInterrupts().empty()) {
		node2 = xmlNewNode(NULL, BAD_CAST "interrupts");
		xmlAddChild(node1, node2);
		{
			vector<XSI_Interrupt> intrvect = system.getInterrupts();
			vector<XSI_Interrupt>::iterator it;
			double	totalIntrs = 0;

			for (it = intrvect.begin(); it != intrvect.end(); it++) {
				if (!it->hasInterrupts())
					continue;
				node3 = xmlNewNode(NULL, BAD_CAST "interrupt_source");
				xmlAddChild(node2, node3);
				xmlNewProp(node3, BAD_CAST "name", BAD_CAST it->getSourceName());
				xmlNewProp(node3, BAD_CAST "intspersec", BAD_CAST it->getIntsPerSecString());
				totalIntrs += it->getIntsPerSec();
			}

			xmlNewProp(node2, BAD_CAST "total_intspersec", BAD_CAST makeFloatStr(totalIntrs, 1));
		}
	}
	node2 = xmlNewNode(NULL, BAD_CAST "memory");
	xmlAddChild(node1, node2);
	snprintf(buffer, sizeof(buffer), "%" PRIu64,
		 system.getVMInfo().ram_total);
	xmlNewChild(node2, NULL, BAD_CAST "mem_total", BAD_CAST buffer);
	snprintf(buffer, sizeof(buffer), "%" PRIu64,
		 system.getVMInfo().ram_free);
	xmlNewChild(node2, NULL, BAD_CAST "mem_free", BAD_CAST buffer);
	snprintf(buffer, sizeof(buffer), "%" PRIu64,
		 system.getVMInfo().swap_total);
	xmlNewChild(node2, NULL, BAD_CAST "swap_total", BAD_CAST buffer);
	snprintf(buffer, sizeof(buffer), "%" PRIu64,
		 system.getVMInfo().swap_free);
	xmlNewChild(node2, NULL, BAD_CAST "swap_free", BAD_CAST buffer);
	node2 = xmlNewNode(NULL, BAD_CAST "os");
	xmlAddChild(node1, node2);
	xmlNewChild(node2, NULL, BAD_CAST "type",
		    BAD_CAST system.getOSType());
	xmlNewChild(node2, NULL, BAD_CAST "release",
		    BAD_CAST system.getOSRelease());
	if (system.getOSStatus() != NULL)
		xmlNewChild(node2, NULL, BAD_CAST "os_status",
			    BAD_CAST system.getOSStatus());
	xmlNewChild(node2, NULL, BAD_CAST "kernel",
		    BAD_CAST system.getOSKernel());
	xmlNewChild(node2, NULL, BAD_CAST "version_string",
		    BAD_CAST system.getOSVersionString());
	node3 = xmlNewNode(NULL, BAD_CAST "uptime");
	xmlAddChild(node2, node3);
	xmlNewProp(node3, BAD_CAST "seconds", BAD_CAST system.getOSUptime());
	node3 = xmlNewNode(NULL, BAD_CAST "load_averages");
	xmlAddChild(node2, node3);
	xmlNewProp(node3, BAD_CAST "avg1min", BAD_CAST system.getLoadAvg(0));
	xmlNewProp(node3, BAD_CAST "avg5min", BAD_CAST system.getLoadAvg(1));
	xmlNewProp(node3, BAD_CAST "avg15min", BAD_CAST system.getLoadAvg(2));


	/* SENSORS INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "sensors");
	xmlAddChild(root_node, node1);
	snprintf(buffer, sizeof(buffer), "%ld", (long int)sensors.size());
	xmlNewProp(node1, BAD_CAST "count", BAD_CAST buffer);
	{
		vector<Sensor *>::iterator it;

		for (it = sensors.begin(); it != sensors.end(); it++) {
			if (!(*it)->isValid())
				continue;
			node2 = xmlNewNode(NULL, BAD_CAST "sensor");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "id", BAD_CAST (*it)->getIDString());
			xmlNewProp(node2, BAD_CAST "dev", BAD_CAST (*it)->getSensorDevice());
			xmlNewProp(node2, BAD_CAST "type", BAD_CAST (*it)->getSensorTypeString());
			if ((*it)->getStatus() != XMLSYS_SENSOR_STATUS_UNSPEC)
				xmlNewProp(node2, BAD_CAST "status", BAD_CAST (*it)->getStatusString());
			if ((*it)->getDescription())
				xmlNewChild(node2, NULL, BAD_CAST "sensor_descr", BAD_CAST (*it)->getDescription());
			if ((*it)->hasUnknownValue()) {
				xmlNewChild(node2, NULL, BAD_CAST "value_unknown", NULL);
				continue;
			}
			node3 = xmlNewChild(node2, NULL, BAD_CAST "value", BAD_CAST (*it)->getValueString());
			if ((*it)->getSensorUnitString())
				xmlNewProp(node3, BAD_CAST "unit", BAD_CAST (*it)->getSensorUnitString());
			switch ((*it)->getSensorType()) {
			case XMLSYS_SENSOR_INDICATOR:
				node3 = xmlNewChild(node2, NULL, BAD_CAST "indicator", NULL);
				xmlNewProp(node3, BAD_CAST "value", BAD_CAST (*it)->getValueString());
				break;
			case XMLSYS_SENSOR_DRIVE:
				node3 = xmlNewChild(node2, NULL, BAD_CAST "drive_status", NULL);
				xmlNewProp(node3, BAD_CAST "value", BAD_CAST (*it)->getValueString());
				break;
			default:
				break;
			}
		}
	}

	/* STORAGE INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "storage");
	xmlAddChild(root_node, node1);
	{
		vector<struct fs> fsvect = storage.getFilesystems();
		vector<struct fs>::iterator it;

		for (it = fsvect.begin(); it != fsvect.end(); it++) {
			node2 = xmlNewNode(NULL, BAD_CAST "filesystem");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "name", BAD_CAST it->fs_name.c_str());
			xmlNewProp(node2, BAD_CAST "type", BAD_CAST it->fs_type.c_str());
			xmlNewChild(node2, NULL, BAD_CAST "mountpoint", BAD_CAST it->mountpoint.c_str());
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->size_total);
			xmlNewChild(node2, NULL, BAD_CAST "size_total", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->size_free);
			xmlNewChild(node2, NULL, BAD_CAST "size_free", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIi64, it->size_avail);
			xmlNewChild(node2, NULL, BAD_CAST "size_avail", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->nodes_total);
			xmlNewChild(node2, NULL, BAD_CAST "nodes_total", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->nodes_free);
			xmlNewChild(node2, NULL, BAD_CAST "nodes_free", BAD_CAST buffer);
		}
	}
	if (!configuration->oneShot() && storage.getDisks().size() > 0) {
		node2 = xmlNewNode(NULL, BAD_CAST "disks");
		xmlAddChild(node1, node2);
		snprintf(buffer, sizeof(buffer), "%" PRIi64, (int64_t)storage.getDisks().size());
		xmlNewProp(node2, BAD_CAST "count", BAD_CAST buffer);
		{
			vector<XSI_DiskIO> dIO = storage.getDisks();
			vector<XSI_DiskIO>::iterator it;

			for (it = dIO.begin(); it != dIO.end(); it++) {
				node3 = xmlNewNode(NULL, BAD_CAST "disk");
				xmlAddChild(node2, node3);
				xmlNewProp(node3, BAD_CAST "name", BAD_CAST it->getDiskName());
				if (it->isIdle())
					xmlNewProp(node3, BAD_CAST "idle", BAD_CAST "true");
				else {
					xmlNewChild(node3, NULL, BAD_CAST "read_bytes_persec", BAD_CAST it->getBytesReadBPSString());
					xmlNewChild(node3, NULL, BAD_CAST "write_bytes_persec", BAD_CAST it->getBytesWriteBPSString());
				}
			}
		}
		node3 = xmlNewChild(node2, NULL, BAD_CAST "disk_summary", NULL);
		xmlNewChild(node3, NULL, BAD_CAST "read_bytes_persec", BAD_CAST storage.getreadSpeedTotalString());
		xmlNewChild(node3, NULL, BAD_CAST "write_bytes_persec", BAD_CAST storage.getwriteSpeedTotalString());
	}

	if (configuration->showRaid() && raids.size() > 0) {
		vector<Raid>::iterator it;

		for (it = raids.begin(); it != raids.end(); it++) {
			vector<RVolume> rvol = it->getVolumes();
			vector<RDisk> rdisk = it->getReserveDisks();
			vector<RVolume>::iterator vit;
			vector<RDisk>::iterator rit;

			node2 = xmlNewNode(NULL, BAD_CAST "raid");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "controller", BAD_CAST it->getRaidControllerName());
			if (rvol.size() > 0) {
				for (vit = rvol.begin(); vit != rvol.end(); vit++) {
					vector<RDisk> vdisk = vit->getVolumeDisks();
					vector<RDisk>::iterator dit;

					if (vdisk.size() == 0)
						errx(1, "%s: Internal error: Can't have RAID volumes without disks", __func__);

					node3 = xmlNewNode(NULL, BAD_CAST "volume");
					xmlAddChild(node2, node3);
					xmlNewProp(node3, BAD_CAST "dev", BAD_CAST vit->getVolumeDevName());
					xmlNewProp(node3, BAD_CAST "id", BAD_CAST vit->getVolumeIDString());
					xmlNewProp(node3, BAD_CAST "level", BAD_CAST vit->getRaidLevelString());
					xmlNewProp(node3, BAD_CAST "status", BAD_CAST vit->getVolumeStatusString());
					if (vit->getVolumeOpProgressString() != NULL)
						xmlNewProp(node3, BAD_CAST "progress_pct", BAD_CAST vit->getVolumeOpProgressString());
					if (vit->getVolumeOpElapsedString() != NULL)
						xmlNewProp(node3, BAD_CAST "elapsed_secs", BAD_CAST vit->getVolumeOpElapsedString());
					xmlNewChild(node3, NULL, BAD_CAST "volume_size", BAD_CAST vit->getVolumeSizeString());
					for (dit = vdisk.begin(); dit != vdisk.end(); dit++) {
						node4 = xmlNewNode(NULL, BAD_CAST "volume_disk");
						xmlAddChild(node3, node4);
						xmlNewProp(node4, BAD_CAST "id", BAD_CAST dit->getDiskIDString());
						xmlNewProp(node4, BAD_CAST "bus", BAD_CAST dit->getDiskBusString());
						xmlNewProp(node4, BAD_CAST "target", BAD_CAST dit->getDiskTargetString());
						if (dit->getDiskLUNString() != NULL)
							xmlNewProp(node4, BAD_CAST "lun", BAD_CAST dit->getDiskLUNString());
						xmlNewProp(node4, BAD_CAST "status", BAD_CAST dit->getDiskStatusString());
						xmlNewChild(node4, NULL, BAD_CAST "size", BAD_CAST dit->getDiskSizeString());
						if (dit->getDiskEnclosureDevName() != NULL)
							xmlNewChild(node4, NULL, BAD_CAST "enclosure", BAD_CAST dit->getDiskEnclosureDevName());
						if (dit->getDiskVendorName() != NULL)
							xmlNewChild(node4, NULL, BAD_CAST "vendor", BAD_CAST dit->getDiskVendorName());
						if (dit->getDiskSerial() != NULL)
							xmlNewChild(node4, NULL, BAD_CAST "serial", BAD_CAST dit->getDiskSerial());
					}
				}
			}
			if (rdisk.size() > 0) {
				for (rit = rdisk.begin(); rit != rdisk.end(); rit++) {
					node3 = xmlNewNode(NULL, BAD_CAST "reserve_disk");
					xmlAddChild(node2, node3);
					xmlNewProp(node3, BAD_CAST "id", BAD_CAST rit->getDiskIDString());
					xmlNewProp(node3, BAD_CAST "bus", BAD_CAST rit->getDiskBusString());
					xmlNewProp(node3, BAD_CAST "target", BAD_CAST rit->getDiskTargetString());
					if (rit->getDiskLUNString() != NULL)
						xmlNewProp(node3, BAD_CAST "lun", BAD_CAST rit->getDiskLUNString());
					xmlNewProp(node3, BAD_CAST "status", BAD_CAST rit->getDiskStatusString());
					xmlNewChild(node3, NULL, BAD_CAST "size", BAD_CAST rit->getDiskSizeString());
					if (rit->getDiskEnclosureDevName() != NULL)
						xmlNewChild(node3, NULL, BAD_CAST "enclosure", BAD_CAST rit->getDiskEnclosureDevName());
					if (rit->getDiskVendorName() != NULL)
						xmlNewChild(node3, NULL, BAD_CAST "vendor", BAD_CAST rit->getDiskVendorName());
					if (rit->getDiskSerial() != NULL)
						xmlNewChild(node3, NULL, BAD_CAST "serial", BAD_CAST rit->getDiskSerial());
				}
			}
		}
	}

	/* NETWORK INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "networking");
	xmlAddChild(root_node, node1);
	xmlNewChild(node1, NULL, BAD_CAST "hostname", BAD_CAST network.getHostName());
	if (network.getDefaultRoutes().size() > 0) {
		vector<struct default_route> routes = network.getDefaultRoutes();
		vector<struct default_route>::iterator it;

		node2 = xmlNewNode(NULL, BAD_CAST "ip_routes");
		xmlAddChild(node1, node2);
		for (it = routes.begin(); it != routes.end(); it++) {
			node3 = xmlNewChild(node2, NULL, BAD_CAST "default", BAD_CAST it->gateway.c_str());
			if (it->family == XSI_RTAF_IPV4)
				xmlNewProp(node3, BAD_CAST "family", BAD_CAST "INET");
			else
				xmlNewProp(node3, BAD_CAST "family", BAD_CAST "INET6");
		}
	}
	{
		vector<XSI_NetIfStat> nsvect = network.getNetIfStats();
		vector<XSI_NetIfStat>::iterator it;

		for (it = nsvect.begin(); it != nsvect.end(); it++) {
			node2 = xmlNewNode(NULL, BAD_CAST "interface");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "name", BAD_CAST it->getIfName());
			if (it->hasMetric()) {
				xmlNewProp(node2, BAD_CAST "metric", BAD_CAST it->getMetricString());
			}
			if (it->hasMTU()) {
				xmlNewProp(node2, BAD_CAST "mtu", BAD_CAST it->getMTUString());
			}
			if (!it->isUp()) {
				xmlNewProp(node2, BAD_CAST "up", BAD_CAST "false");
				continue;
			}
			xmlNewProp(node2, BAD_CAST "up", BAD_CAST "true");
			if (it->isRunning() || it->isLoopback()
			    || it->isPointToPoint() || it->isPromiscuous()) {
				node3 = xmlNewNode(NULL, BAD_CAST "properties");
				xmlAddChild(node2, node3);
				if (it->isRunning())
					xmlNewChild(node3, NULL, BAD_CAST "running", NULL);
				if (it->isPromiscuous())
					xmlNewChild(node3, NULL, BAD_CAST "promiscuous", NULL);
				if (it->isLoopback())
					xmlNewChild(node3, NULL, BAD_CAST "loopback", NULL);
				if (it->isPointToPoint()) {
					node4 = xmlNewNode(NULL, BAD_CAST "point-to-point");
					xmlAddChild(node3, node4);
					xmlNewProp(node4, BAD_CAST "dest_addr", BAD_CAST it->getPTPDestAddressString());
				}
			}
			if (it->getDescriptionString())
				xmlNewChild(node2, NULL, BAD_CAST "if_descr", BAD_CAST it->getDescriptionString());
			if (it->getMediaString())
				xmlNewChild(node2, NULL, BAD_CAST "media", BAD_CAST it->getMediaString());
			if (it->getStatusString()) {
				node3 = xmlNewChild(node2, NULL, BAD_CAST "link_status", BAD_CAST it->getStatusString());
				xmlNewProp(node3, BAD_CAST "state", BAD_CAST it->getLinkStatusString());
			}
			if (it->isCARPDev()) {
				node3 = xmlNewNode(NULL, BAD_CAST "carp");
				xmlAddChild(node2, node3);
				xmlNewProp(node3, BAD_CAST "status", BAD_CAST it->getCARPStatusString());
				if (it->getCARPDeviceString())
					xmlNewProp(node3, BAD_CAST "carpdev", BAD_CAST it->getCARPDeviceString());
				xmlNewProp(node3, BAD_CAST "vhid", BAD_CAST it->getCARPVHIDString());
				xmlNewProp(node3, BAD_CAST "advskew", BAD_CAST it->getCARPAdvskewString());
				xmlNewProp(node3, BAD_CAST "advbase", BAD_CAST it->getCARPAdvbaseString());
				if (it->getCARPPreemptString())
					xmlNewProp(node3, BAD_CAST "preempt", BAD_CAST it->getCARPPreemptString());
				if (it->getCARPArpbalanceString())
					xmlNewProp(node3, BAD_CAST "arpbalance", BAD_CAST it->getCARPArpbalanceString());
			}
			if (it->hasAddresses()) {
				vector<XSI_IfAddress> addrs = it->getIfAddresses();
				vector<XSI_IfAddress>::iterator addrIt;

				node3 = xmlNewNode(NULL, BAD_CAST "addresses");
				xmlAddChild(node2, node3);
				if (it->getLladdrString())
					xmlNewChild(node3, NULL, BAD_CAST "lladdr", BAD_CAST it->getLladdrString());
				for (addrIt = addrs.begin(); addrIt != addrs.end(); addrIt++) {
					node4 = xmlNewChild(node3, NULL, BAD_CAST "if_address", BAD_CAST addrIt->getAddressString());
					xmlNewProp(node4, BAD_CAST "family", BAD_CAST addrIt->getAddressFamilyString());
					if (addrIt->hasPrefixLen()) {
						xmlNewProp(node4, BAD_CAST "prefixlen", BAD_CAST addrIt->getPrefixLenString());
					}
				}
			}
			if (!it->isLoopback()) {
				node3 = xmlNewNode(NULL, BAD_CAST "statistics");
				xmlAddChild(node2, node3);

				if (!it->hasTraffic())
					xmlNewProp(node3, BAD_CAST "has_traffic", BAD_CAST "false");
				else {
					node4 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node3, node4);
					xmlNewChild(node4, NULL, BAD_CAST "in", BAD_CAST it->getPacketsInString());
					if (it->getPacketsInPerSecString() && !configuration->oneShot())
						xmlNewChild(node4, NULL, BAD_CAST "in_persec", BAD_CAST it->getPacketsInPerSecString());
					xmlNewChild(node4, NULL, BAD_CAST "out", BAD_CAST it->getPacketsOutString());
					if (it->getPacketsOutPerSecString() && !configuration->oneShot())
						xmlNewChild(node4, NULL, BAD_CAST "out_persec", BAD_CAST it->getPacketsOutPerSecString());
					if (it->hasDroppedPkgs()) {
						xmlNewChild(node4, NULL, BAD_CAST "dropped", BAD_CAST it->getPacketsDroppedString());
						if (it->getPacketsDroppedPerSecString() && !configuration->oneShot())
							xmlNewChild(node4, NULL, BAD_CAST "dropped_persec", BAD_CAST it->getPacketsDroppedPerSecString());
					}

					node4 = xmlNewNode(NULL, BAD_CAST "bytes");
					xmlAddChild(node3, node4);
					xmlNewChild(node4, NULL, BAD_CAST "in", BAD_CAST it->getBytesInString());
					if (it->getBytesInPerSecString() && !configuration->oneShot())
						xmlNewChild(node4, NULL, BAD_CAST "in_persec", BAD_CAST it->getBytesInPerSecString());
					xmlNewChild(node4, NULL, BAD_CAST "out", BAD_CAST it->getBytesOutString());
					if (it->getBytesOutPerSecString() && !configuration->oneShot())
						xmlNewChild(node4, NULL, BAD_CAST "out_persec", BAD_CAST it->getBytesOutPerSecString());
				}

				if (it->hasErrors()) {
					xmlNewChild(node3, NULL, BAD_CAST "errors_in", BAD_CAST it->getErrorsInString());
					xmlNewChild(node3, NULL, BAD_CAST "errors_out", BAD_CAST it->getErrorsOutString());
				}
				if (it->hasCollisions()) {
					xmlNewChild(node3, NULL, BAD_CAST "collisions", BAD_CAST it->getCollisionsString());
				}
			}
		}
		if (configuration->showPacketFilter()) {
			node2 = xmlNewNode(NULL, BAD_CAST "packetfilter");
			xmlAddChild(node1, node2);
			if (!packetFilter.isEnabled())
				xmlNewProp(node2, BAD_CAST "enabled", BAD_CAST "false");
			else {
				vector<FilteredIfStat> fIfStats = packetFilter.getFilteredIfStats();
				vector<FilteredIfStat>::iterator it;

				xmlNewProp(node2, BAD_CAST "enabled", BAD_CAST "true");
				xmlNewProp(node2, BAD_CAST "runtimesecs", BAD_CAST packetFilter.getRuntimeSecString());
				xmlNewChild(node2, NULL, BAD_CAST "rule_matches", BAD_CAST packetFilter.getRuleMatchesString());
				xmlNewChild(node2, NULL, BAD_CAST "rule_matches_avg", BAD_CAST packetFilter.getRuleMatchesAverageString());
				if (packetFilter.getRuleMatchesPerSecString() && !configuration->oneShot())
					xmlNewChild(node2, NULL, BAD_CAST "rule_matches_persec", BAD_CAST packetFilter.getRuleMatchesPerSecString());
				if (packetFilter.hasStates()) {
					node3 = xmlNewNode(NULL, BAD_CAST "states");
					xmlAddChild(node2, node3);
					xmlNewProp(node3, BAD_CAST "count", BAD_CAST packetFilter.getStatesString());
					if (packetFilter.getStateSearchesString()) {
						xmlNewChild(node3, NULL, BAD_CAST "searches", BAD_CAST packetFilter.getStateSearchesString());
						xmlNewChild(node3, NULL, BAD_CAST "searches_avg", BAD_CAST packetFilter.getStateSearchesAverageString());
						if (packetFilter.getStateSearchesPerSecString() && !configuration->oneShot())
							xmlNewChild(node3, NULL, BAD_CAST "searches_persec", BAD_CAST packetFilter.getStateSearchesPerSecString());
					}
					if (packetFilter.getStateInsertsString()) {
						xmlNewChild(node3, NULL, BAD_CAST "inserts", BAD_CAST packetFilter.getStateInsertsString());
						xmlNewChild(node3, NULL, BAD_CAST "inserts_avg", BAD_CAST packetFilter.getStateInsertsAverageString());
						if (packetFilter.getStateInsertsPerSecString() && !configuration->oneShot())
							xmlNewChild(node3, NULL, BAD_CAST "inserts_persec", BAD_CAST packetFilter.getStateInsertsPerSecString());
					}
					if (packetFilter.getStateRemovalsString()) {
						xmlNewChild(node3, NULL, BAD_CAST "removals", BAD_CAST packetFilter.getStateRemovalsString());
						xmlNewChild(node3, NULL, BAD_CAST "removals_avg", BAD_CAST packetFilter.getStateRemovalsAverageString());
						if (packetFilter.getStateRemovalsPerSecString() && !configuration->oneShot())
							xmlNewChild(node3, NULL, BAD_CAST "removals_persec", BAD_CAST packetFilter.getStateRemovalsPerSecString());
					}
					if (packetFilter.getStateMismatchesString()) {
						xmlNewChild(node3, NULL, BAD_CAST "mismatches", BAD_CAST packetFilter.getStateMismatchesString());
						xmlNewChild(node3, NULL, BAD_CAST "mismatches_avg", BAD_CAST packetFilter.getStateMismatchesAverageString());
						if (packetFilter.getStateMismatchesPerSecString() && !configuration->oneShot())
							xmlNewChild(node3, NULL, BAD_CAST "mismatches_persec", BAD_CAST packetFilter.getStateMismatchesPerSecString());
					}
				}
				for (it = fIfStats.begin() ; it != fIfStats.end(); it++) {
					node3 = xmlNewNode(NULL, BAD_CAST "filtered_if");
					xmlAddChild(node2, node3);
					xmlNewProp(node3, BAD_CAST "name", BAD_CAST it->getIfName());

					if (!it->hasTraffic())
						xmlNewProp(node3, BAD_CAST "has_traffic", BAD_CAST "false");
					else {
						if (it->hasIPv4Traffic()) {
							node4 = xmlNewNode(NULL, BAD_CAST "ipv4");
							xmlAddChild(node3, node4);

							node5 = xmlNewNode(NULL, BAD_CAST "packets");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST it->getPktsV4InBlockedString());
							if (it->getPktsV4InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST it->getPktsV4InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST it->getPktsV4InPassedString());
							if (it->getPktsV4InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST it->getPktsV4InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST it->getPktsV4OutBlockedString());
							if (it->getPktsV4OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST it->getPktsV4OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST it->getPktsV4OutPassedString());
							if (it->getPktsV4OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST it->getPktsV4OutPassedPerSecString());

							node5 = xmlNewNode(NULL, BAD_CAST "bytes");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST it->getBytesV4InBlockedString());
							if (it->getBytesV4InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST it->getBytesV4InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST it->getBytesV4InPassedString());
							if (it->getBytesV4InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST it->getBytesV4InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST it->getBytesV4OutBlockedString());
							if (it->getBytesV4OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST it->getBytesV4OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST it->getBytesV4OutPassedString());
							if (it->getBytesV4OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST it->getBytesV4OutPassedPerSecString());
						}

						if (it->hasIPv6Traffic()) {
							node4 = xmlNewNode(NULL, BAD_CAST "ipv6");
							xmlAddChild(node3, node4);

							node5 = xmlNewNode(NULL, BAD_CAST "packets");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST it->getPktsV6InBlockedString());
							if (it->getPktsV6InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST it->getPktsV6InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST it->getPktsV6InPassedString());
							if (it->getPktsV6InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST it->getPktsV6InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST it->getPktsV6OutBlockedString());
							if (it->getPktsV6OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST it->getPktsV6OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST it->getPktsV6OutPassedString());
							if (it->getPktsV6OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST it->getPktsV6OutPassedPerSecString());

							node5 = xmlNewNode(NULL, BAD_CAST "bytes");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST it->getBytesV6InBlockedString());
							if (it->getBytesV6InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST it->getBytesV6InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST it->getBytesV6InPassedString());
							if (it->getBytesV6InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST it->getBytesV6InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST it->getBytesV6OutBlockedString());
							if (it->getBytesV6OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST it->getBytesV6OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST it->getBytesV6OutPassedString());
							if (it->getBytesV6OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST it->getBytesV6OutPassedPerSecString());
						}
					}
				}
			}
		}
	}
}

xmlDocPtr
XSIBase::getSysInfoDocPtr(void)
{
	return (sysInfoDocument);
}

void
XSIBase::updateTimeKeeper(void)
{
	XSI_TimeKeeper::get()->triggerUpdate();
}

/* virtual */
void
XSIBase::collectSysInfo(void)
{
	// empty
}

/* virtual */
void
XSIBase::updateStats(void)
{
	// empty
}

/* virtual */
void
XSIBase::openPrivilegedDevices(void)
{
	// empty
}
